package chessTrainingWheels;

/**
 * Command to be sent to the chess engine.
 *
 * @author Josh Weihnacht
 *
 */
public class EngineCommand {

	//==========================================================================
	//===   INSTANCE VARIABLES   ===============================================
	
	/**
	 * Command to be sent to the chess engine
	 */
	private String command;
	
	/**
	 * ID number of the command used for pinging the chess engine
	 */
	private Integer pingId;
	
	/**
	 * Has a ping for this command been sent?
	 * A ping needs to be sent for a command to make sure that the
	 * chess engine is done with everything and is ready to receive 
	 * a new command.
	 */
	private Boolean pingSent;
	
	
	//===   INSTANCE VARIABLES   ===============================================
	//==========================================================================
	
	
	
	
	
	//==========================================================================
	//===   CONSTRUCTORS   =====================================================

	/**
	 * Constructor for a command to be sent to the chess engine
	 */
	public EngineCommand(String command, Integer pingId) {
		this.command = command;
		this.pingId = pingId;
		this.pingSent = false;
	}
	
	//===   CONSTRUCTORS   =====================================================
	//==========================================================================
	
	
	
	
	
	//==========================================================================
	//===   ENGINE COMMAND METHODS   ===========================================

	/**
	 * Get the command string for this command
	 * @return the command string for this command
	 */
	public String getCommand() {
		return this.command;
	}
	
	/**
	 * Get the ID of this command
	 * @return the ID of this command
	 */
	public Integer getPingId() {
		return this.pingId;
	}
	
	/**
	 * 
	 * @return
	 */
	public Integer sendPingId() {
		this.setPingSent();
		return this.getPingId();
	}
	
	/**
	 * Has a ping for this command been sent?
	 * @return whether a ping for this command has been sent.
	 */
	public Boolean hasPingBeenSent() {
		return this.pingSent;
	}
	
	/**
	 * Mark that a ping has been sent.
	 */
	public void setPingSent() {
		this.pingSent = true;
	}
	
	/**
	 * Reset the ping sent flag.
	 */
	public void unsetPingSent() {
		this.pingSent = false;
	}
	
	
	//===   ENGINE COMMAND METHODS   ===========================================
	//==========================================================================
}
