/* This file is part of chessbox_gnu4j, a Java port of GNU Chess (5.0.7).

   Copyright (c) 2005 Xan Gregg (Java port)
   Copyright (c) 1999-2002 Free Software Foundation, Inc.

   chessbox_gnu4j is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 or (at your option)
   any later version.

   chessbox_gnu4j is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GNU Chess; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.

   Contact Info:
     bug-gnu-chess@gnu.org
     xan.gregg@pobox.com
*/
package org.chessbox.gnu4j;

/**
 * Board represents an entire board's state and is structured to
 * simplify analysis.
 * <p/>
 * Created by Xan Gregg.
 * Date: Apr 22, 2005
 */
public class Board {
    public /*BitBoard*/ long[][] b = new /*BitBoard*/long[2][7];
    /* piece/pawn positions by side (0=white, 1=black)
       and then by piece (1=pawn..6=king). For example,
       b[white][knight] has bits set for every board
       position occupied by a White Knight. */

    public /*BitBoard*/ long[] friends = new /*BitBoard*/long[2];   /* Friendly (this side's) pieces */
    public /*BitBoard*/ long blocker;
    public /*BitBoard*/ long blockerr90;   /* rotated 90 degrees */
    public /*BitBoard*/ long blockerr45;   /* rotated 45 degrees */
    public /*BitBoard*/ long blockerr315;  /* rotated 315 degrees */
    public short ep;              /* Location of en passant square */
    public short flag;            /* Flags related to castle privileges */
    public short side;            /* Color of side on move: 0=white, 1=black */
    public short[] material = new short[2];     /* Total material by side not inc. king */
    public short[] pmaterial = new short[2];    /* Total pawn material by side not inc. king */
    public boolean[] castled = new boolean[2];      /* True (1) if side is castled */
    public short[] king = new short[2];         /* Location of king 0 - a1 .. 63 - h8 */

    public Board cloneBoard() {
        Board clone = new Board();
        for (int i = 0; i < b.length; i++) {
            long[] longs = b[i];
            for (int j = 0; j < longs.length; j++) {
                clone.b[i][j] = longs[j];
            }
        }
        clone.friends[0] = friends[0];
        clone.friends[1] = friends[1];
        clone.blocker = blocker;
        clone.blockerr90 = blockerr90;
        clone.blockerr45 = blockerr45;
        clone.blockerr315 = blockerr315;
        clone.ep = ep;
        clone.flag = flag;
        clone.side = side;
        clone.material[0] = material[0];
        clone.material[1] = material[1];
        clone.pmaterial[0] = pmaterial[0];
        clone.pmaterial[1] = pmaterial[1];
        clone.castled[0] = castled[0];
        clone.castled[1] = castled[1];
        clone.king[0] = king[0];
        clone.king[1] = king[1];

        return clone;
    }

    public boolean equals(Board other) {
        for (int i = 0; i < b.length; i++) {
            long[] longs = b[i];
            for (int j = 0; j < longs.length; j++) {
                if (other.b[i][j] != longs[j])
                    return false;
            }
        }
        return other.friends[0] == friends[0] &&
                other.friends[1] == friends[1] &&
                other.blocker == blocker &&
                other.blockerr90 == blockerr90 &&
                other.blockerr45 == blockerr45 &&
                other.blockerr315 == blockerr315 &&
                other.ep == ep &&
                other.flag == flag &&
                other.side == side &&
                other.material[0] == material[0] &&
                other.material[1] == material[1] &&
                other.pmaterial[0] == pmaterial[0] &&
                other.pmaterial[1] == pmaterial[1] &&
                other.castled[0] == castled[0] &&
                other.castled[1] == castled[1] &&
                other.king[0] == king[0] &&
                other.king[1] == king[1];
    }
}
