/* This file is part of chessbox_gnu4j, a Java port of GNU Chess (5.0.7).

   Copyright (c) 2005 Xan Gregg (Java port)
   Copyright (c) 1999-2002 Free Software Foundation, Inc.

   chessbox_gnu4j is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 or (at your option)
   any later version.

   chessbox_gnu4j is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GNU Chess; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.

   Contact Info:
     bug-gnu-chess@gnu.org
     xan.gregg@pobox.com
*/
package org.chessbox.gnu4j;

import java.util.StringTokenizer;
import java.util.Map;
import java.util.HashMap;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.File;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;
@SuppressWarnings("all")
/**
 * Created by Xan Gregg.
 * Date: Apr 24, 2005
 */
public class Cmd extends Common{
    static String logfile;
    static String gamefile;

/*
 * Splitting input is actually not neccessary, but we find
 * tokens separated by whitespace and put pointers on them.
 * How many do we need? We just take 3 for now. Check if the
 * fact that tokens are not 0-terminated but whitespace-terminated
 * generates bugs. (Already killed one bug in move.c)
 * We also kill trailing whitespace. (The trailing '\n' might
 * be really annoying otherwise.)
 */

    static final int TOKENS = 4;    // was 3 for ccc, but 4 needed for cmd_level

    static String[] token = new String[TOKENS];

    static void split_input() {
        StringTokenizer st = new StringTokenizer(Input.inputstr);
        for (int i = 0; i < TOKENS; i++)
            if (st.hasMoreTokens())
                token[i] = st.nextToken();
            else
                token[i] = "";
    }

/*
 * Compares two tokens, returns true on equality. Tokens
 * are separated by whitespace.
 */
    static boolean tokeneq(String s, String t) {
        if (s.indexOf(" ") >= 0 || t.indexOf(" ") >= 0)
            throw new IllegalArgumentException();
        return s.equals(t);
    }

    static void cmd_accepted() {}

    static void cmd_activate() {
        ClearFlag(TIMEOUT);
        ClearFlag(ENDED);
    }

    static void cmd_analyze() {
        /*
         * "analyze" mode is similar to force, hard and post together
         * in that it produces a text output like post, but must
         * think indefinitely like ponder.
         *
         * Some additional output is expected in command ".\n" but if ignored
         * this should not be sent any more
         */

/* TODO correct output, add fail high low */

        if ((flags & ANALYZE) == 0) {
            preanalyze_flags = flags; /* save these flags for exit */
            SetFlag(ANALYZE);
            cmd_post();
            cmd_force();
            cmd_hard();
        }
    }

    static void cmd_bk() {
        /* Print moves from Open Book for Xboard/WinBoard */
        /* Lines must start with " " and end with blank line */
        /* No need to test for xboard as it is generally useful */
        Book.BookQuery(1);
        System.out.println(""); /* Blank line */
        System.out.flush();
    }

    static void cmd_black() {
        /*
         * No longer used by Xboard but requested as a feature
         */

        Init.NewPosition(); /* Reset some flags and settings */
        ClearFlag(THINK);
        ClearFlag(MANUAL);
        ClearFlag(TIMEOUT);
        computer = white;
        if (board.side == white) {
            board.side = black;
            board.ep = -1;
        }
    }

    static void cmd_book() {
        if (tokeneq(token[1], "add")) {
            if (!new File(token[2]).exists()) {
                System.out.println("The syntax to add a new book is:\n\n\tbook add file.pgn");
            } else {
                PGN.BookPGNReadFromFile(token[2]);
            }
        } else if (tokeneq(token[1], "on") || tokeneq(token[1], "prefer")) {
            bookmode = BOOKPREFER;
            System.out.println("book now on.");
        } else if (tokeneq(token[1], "off")) {
            bookmode = BOOKOFF;
            System.out.println("book now off.");
        } else if (tokeneq(token[1], "best")) {
            bookmode = BOOKBEST;
            System.out.println("book now best.");
        } else if (tokeneq(token[1], "worst")) {
            bookmode = BOOKWORST;
            System.out.println("book now worst.");
        } else if (tokeneq(token[1], "random")) {
            bookmode = BOOKRAND;
            System.out.println("book now random.");
        }
    }

/* Our opponent is a computer */
    static void cmd_computer() {}

    static void cmd_depth() {
        SearchDepth = Integer.parseInt(token[1]);
        System.out.println("Search to a depth of " + SearchDepth);
    }

/* Ignore draw offers */
    static void cmd_draw() {}

    static void cmd_easy() { ClearFlag(HARD); }

/* Predecessor to setboard */
    static void cmd_edit() {
        if ((flags & XBOARD) != 0) {
            System.out.println("tellusererror command 'edit' not implemented\n");
            System.out.flush();
        }
    }

    static void cmd_epd() {
        EPD.ParseEPD(token[1]);
        Init.NewPosition();
        Output.ShowBoard();
        System.out.println("\n" + id + " : Best move = " + solution);
    }

    static void cmd_exit() {
        /*
         * "exit" is a synonym for quit except in engine mode
         * when it means leave analyze mode
         */

        if ((flags & ANALYZE) != 0) {
            flags = preanalyze_flags; /* this implicitly clears ANALYZE flag */
        } else {
            cmd_quit();
        }


    }

    static void cmd_force() { SetFlag(MANUAL); }

    static void cmd_go() {
        SetFlag(THINK);
        ClearFlag(MANUAL);
        ClearFlag(TIMEOUT);
        ClearFlag(ENDED);
        computer = board.side;
    }

    static void cmd_hard() { SetFlag(HARD); }

    static void cmd_hash() {
        if (tokeneq(token[1], "off"))
            ClearFlag(USEHASH);
        else if (tokeneq(token[1], "on"))
            SetFlag(USEHASH);
        System.out.println("Hashing " + ((flags & USEHASH) != 0 ? "on" : "off"));
    }

    static void cmd_hashsize() {
        if (token[1].length() == 0) {
            System.out.println("Current HashSize is " + HashSize + " slots");
        } else {
            try {
                int i = Integer.parseInt(token[1]);
                Hash.CalcHashSize(i);
                Init.InitHashTable();
            } catch (NumberFormatException e) {
                System.out.println("Hashsize out of Range or Invalid\n");
            }
        }
    }

/* Give a possible move for the player to play */
    static void cmd_hint() {
        if ((flags & ENDED) != 0) {
            System.out.println("The game is over.\n");
        } else {
            int HintMove;
            HintMove = Tree[TreePtr[1]].move; /* Pick first move in tree */
            if (Move.IsLegalMove(HintMove)) {
                GenMove.GenMoves(1);
                Move.SANMove(Tree[TreePtr[1]].move, 1);
                System.out.println("Hint: " + SANmv);
            } else {
                System.out.println("No hint available at this time\n");
            }
        }
        System.out.flush();
    }

    static void cmd_level() {
        SearchDepth = 0;
        //sscanf (token[1], "%d %f %d", &TCMove, &TCTime, &TCinc);
        try {
            TCMove = Integer.parseInt(token[1]);
        } catch (NumberFormatException e) {
            TCMove = 0;
        }
        try {
            TCTime = Float.parseFloat(token[2]);
        } catch (NumberFormatException e) {
            TCTime = 0;
        }
        try {
            TCinc = Integer.parseInt(token[3]);
        } catch (NumberFormatException e) {
            TCMove = 0;
        }
        if (TCinc == 0) {
            TCMove = 35 /* MIN((5*(GameCnt+1)/2)+1,60) */;
            System.out.println("TCMove =" + TCMove);
            suddendeath = true;
        } else
            suddendeath = false;
        if (TCTime == 0) {
            SetFlag(TIMECTL);
            SearchTime = TCinc / 2.0f;
            System.out.println("Fischer increment of " + TCinc + " seconds");
        } else {
            SetFlag(TIMECTL);
            MoveLimit[white] = MoveLimit[black] = TCMove - (GameCnt + 1) / 2;
            TimeLimit[white] = TimeLimit[black] = TCTime * 60;
            if ((flags & XBOARD) == 0) {
                System.out.println("Time Control: " + MoveLimit[white] +
                        " moves in " + TimeLimit[white] + " secs"); // CCC %.2f
                System.out.println("Fischer increment of " + TCinc + " seconds");
            }
        }
    }

    static void cmd_list() {
        if (token[1] == null || token[1].length() == 0)
            token[1] = "rscore";
        if (token[1].startsWith("?")) {
            System.out.println("name    - list known players alphabetically");
            System.out.println("score   - list by GNU best result first ");
            System.out.println("reverse - list by GNU worst result first");
        } else {
            Players.DBListPlayer(token[1]);
        }
    }

    static void cmd_load() {
        EPD.LoadEPD(token[1]);
        if (!Util.ValidateBoard()) {
            SetFlag(ENDED);
            System.out.println("Board is wrong!");
        }
    }

    static void cmd_manual() { SetFlag(MANUAL); }

/* Move now, not applicable */
    static void cmd_movenow() {}

/*
 * TODO: Add a logpath variable or macro, not always dump into current
 * dir. Again, how does one handle paths portably across Unix/Windows?
 *   -- Lukas
 */
    static void cmd_name() {
        name = token[1];
        for (int suffix = 0; suffix < 1000; suffix++) {
            String suffix3 = Integer.toString(1000 + suffix).substring(1);
            logfile = "log." + suffix3;
            gamefile = "game." + suffix3;
            /*
             * There is an obvious race condition here but who cares, we just
             * bail out in case of failure... --Lukas
             */
            try {
                ofp = new PrintWriter(new FileWriter(logfile));
                ofpIsStdout = false;
            } catch (IOException e) {
                ofp = new PrintWriter(System.out);
                ofpIsStdout = true;
                //fprintf(stderr, "Failed to open %s for writing: %s\n",
                //    logfile, strerror(errno));
            }
            return;
        }

        System.err.println("Could not create logfile, all slots occupied.");
        System.err.println("You may consider deleting or renaming your existing logfiles.");
    }

    static void cmd_new() {
        Init.InitVars();
        Init.NewPosition();
        /* Protocol specification for ANALYZE says "new" does not end analysis */
        if ((flags & ANALYZE) == 0)
            ClearFlag(MANUAL);
        ClearFlag(THINK);
        myrating = opprating = 0;
    }

    static void cmd_nopost() {	ClearFlag(POST); }

    static void cmd_null() {
        if (tokeneq(token[1], "off"))
            ClearFlag(USENULL);
        else if (tokeneq(token[1], "on"))
            SetFlag(USENULL);
        System.out.println("Null moves " + ((flags & USENULL) != 0 ? "on" : "off"));
    }

    static void cmd_otim() {}

    static void cmd_pgnload() { PGN.PGNReadFromFile(token[1]); }

/*
 * XXX - Filenames with spaces will break here,
 * do we want to support them? I vote for "no"
 *   - Lukas
 */
    static void cmd_pgnsave() {
        if (token[1] != null && token[1].length() > 0)
            PGN.PGNSaveToFile(token[1], "");
        else
            System.out.println("Invalid filename.\n");
    }

    static void cmd_ping() {
        /* If ping is received when we are on move, we are supposed to
           reply only after moving.  In this version of GNU Chess, we
           never read commands while we are on move, so we don't have to
           worry about that here. */
        System.out.println("pong " + token[1]);
        System.out.flush();
    }

    static void cmd_post() { SetFlag(POST); }

    static void cmd_protover() {
        if ((flags & XBOARD) != 0) {
            /* Note: change this if "draw" command is added, etc. */
            System.out.println("feature setboard=1 analyze=1 ping=1 draw=0 sigint=0" +
                    " variants=\"normal\" myname=\"" +
                    Config.PROGRAM + " " + Config.VERSION + "\" done=1");
            System.out.flush();
        }
    }

    static void cmd_quit() { SetFlag(QUIT); }

    static void cmd_random() {}

    static void cmd_rating() {
        myrating = Integer.parseInt(token[1]);
        opprating = Integer.parseInt(token[2]);
        ofp.println("my rating = " + myrating + " opponent rating = " + opprating);
        /* Change randomness of book based on opponent rating. */
        /* Basically we play narrower book the higher the opponent */
        if (opprating >= 1700)
            bookfirstlast = 2;
        else if (opprating >= 1700)
            bookfirstlast = 2;
        else
            bookfirstlast = 2;
    }

    static void cmd_rejected() {}

    static void cmd_remove() {
        if (GameCnt >= 0) {
            ClearFlag(ENDED);
            ClearFlag(TIMEOUT);
            Game[GameCnt].move = Move.UnmakeMove(board.side, Game[GameCnt].move);
            if (GameCnt >= 0) {
                Game[GameCnt].move = Move.UnmakeMove(board.side, Game[GameCnt].move);
                if ((flags & XBOARD) == 0)
                    Output.ShowBoard();
            }
            PGN.PGNSaveToFile("game.log", "");
        } else
            System.out.println("No moves to undo! \n");
    }

    static void cmd_result() {
        if (!ofpIsStdout) {
            ofp.println("result: " + token[1]);
            ofp.close();
            ofp = new PrintWriter(System.out);
            System.out.println("Save to " + gamefile);
            PGN.PGNSaveToFile(gamefile, token[1]);
            Players.DBUpdatePlayer(name, token[1]);
        }
    }

    static void cmd_save() {
        if (token[1] != null && token[1].length() > 0)
            EPD.SaveEPD(token[1]);
        else
            System.out.println("Invalid filename.\n");
    }

    static void cmd_setboard() {
        /* setboard uses FEN, not EPD, but ParseEPD will accept FEN too */
        EPD.ParseEPD(token[1]);
        Init.NewPosition();
    }

    static void cmd_solve() { Solve.Solve(token[1]); }

/* Set total time for move to be N seconds is "st N" */
    static void cmd_st() {
        /* Approximately level 1 0 N */
        TCinc = Integer.parseInt(token[1]);
        suddendeath = false;
        /* Allow a little fussiness for failing low etc */
        SearchTime = TCinc * 0.90f;
        ClearFlag(TIMECTL);
    }

    static void cmd_switch() {
        board.side = (short) (1 ^ board.side);
        board.ep = -1; /* Enpassant doesn't apply after switch */
        System.out.println((board.side == white ? "White" : "Black") + " to move");
    }

    static void cmd_time() {
        TimeLimit[1 ^ board.side] = Integer.parseInt(token[1]) / 100.0f;
    }

    static void cmd_undo() {
        if (GameCnt >= 0)
            Game[GameCnt].move = Move.UnmakeMove(board.side, Game[GameCnt].move);
        else
            System.out.println("No moves to undo! \n");
        MoveLimit[board.side]++;
        TimeLimit[board.side] += Game[GameCnt + 1].et;
        if ((flags & XBOARD) == 0) Output.ShowBoard();
    }

    static void cmd_usage() {
        System.out.println("\n" +
                " Usage: " + progname + " [OPTION]\n" +
                "\n" +
                " -h, --help         display this help and exit\n" +
                " -v, --version      display version information and exit\n" +
                "\n" +
                " -x, --xboard       start in engine mode\n" +
                " -p, --post   	   start up showing thinking\n" +
                " -e, --easy   	   disable thinking in opponents time\n" +
                " -m, --manual  	   enable manual mode\n" +
                " -s size, --hashsize=size   specify hashtable size in slots\n" +
                "\n" +
                " Options xboard and post are accepted without leading dashes\n" +
                " for backward compatibility\n" +
                "\n" +
                "Report bugs to <bugs@chessbox.org>.\n" +
                "\n");
    }


/* Play variant, we instruct interface in protover we play normal */
    static void cmd_variant() {}

    static void cmd_version() {
        if ((flags & XBOARD) == 0)
            System.out.println(Config.PROGRAM + " " + Config.VERSION);
        else
            System.out.println("Chess");
    }

    static void cmd_white() {
        /*
         * No longer used by Xboard but requested as a feature
         */

        Init.NewPosition(); /* Reset some flags and settings */
        ClearFlag(THINK);
        ClearFlag(MANUAL);
        ClearFlag(TIMEOUT);
        computer = black;
        if (board.side == black) {
            board.side = white;
            board.ep = -1; /* Hack to fixed appearing pawn bug */
        }
    }

    static void cmd_xboard() {
        if (tokeneq(token[1], "off"))
            ClearFlag(XBOARD);
        else if (tokeneq(token[1], "on"))
            SetFlag(XBOARD);
        else if ((flags & XBOARD) == 0) { /* set if unset and only xboard called */
            SetFlag(XBOARD);	    /* like in xboard/winboard usage */
        }
    }

/*
 * Command with subcommands, could write secondary method
 * tables here
 */

    static void cmd_show()
            /************************************************************************
             *
             *  The show command driver section.
             *
             ************************************************************************/ {
        if (tokeneq(token[1], "board"))
            Output.ShowBoard();
        else if (tokeneq(token[1], "rating")) {
            System.out.println("My rating = " + myrating);
            System.out.println("Opponent rating = " + opprating);
        } else if (tokeneq(token[1], "time"))
            Output.ShowTime();
        else if (tokeneq(token[1], "moves")) {
            GenCnt = 0;
            TreePtr[2] = TreePtr[1];
            GenMove.GenMoves(1);
            Output.ShowMoveList(1);
            System.out.println("No. of moves generated = " + GenCnt);
        } else if (tokeneq(token[1], "escape")) {
            GenCnt = 0;
            TreePtr[2] = TreePtr[1];
            GenMove.GenCheckEscapes(1);
            Output.ShowMoveList(1);
            System.out.println("No. of moves generated = " + GenCnt);
        } else if (tokeneq(token[1], "noncapture")) {
            GenCnt = 0;
            TreePtr[2] = TreePtr[1];
            GenMove.GenNonCaptures(1);
            GenMove.FilterIllegalMoves(1);
            Output.ShowMoveList(1);
            System.out.println("No. of moves generated = " + GenCnt);
        } else if (tokeneq(token[1], "capture")) {
            GenCnt = 0;
            TreePtr[2] = TreePtr[1];
            GenMove.GenCaptures(1);
            GenMove.FilterIllegalMoves(1);
            Output.ShowMoveList(1);
            System.out.println("No. of moves generated = " + GenCnt);
        } else if (tokeneq(token[1], "eval") || tokeneq(token[1], "score")) {
            int s, wp, bp, wk, bk;
            /*BitBoard*/long[] b;

            phase = PHASE();
            Atak.GenAtaks();
            pinned = Atak.FindPins();
            hunged[white] = Hung.EvalHung(white);
            hunged[black] = Hung.EvalHung(black);
            b = board.b[white];
            pieces[white] = b[knight] | b[bishop] | b[rook] | b[queen];
            b = board.b[black];
            pieces[black] = b[knight] | b[bishop] | b[rook] | b[queen];
            wp = Eval.ScoreP(white);
            bp = Eval.ScoreP(black);
            wk = Eval.ScoreK(white);
            bk = Eval.ScoreK(black);
            // System.out.println ("White:  Mat:%4d/%4d  P:%d  N:%d  B:%d  R:%d  Q:%d  K:%d  Dev:%d  h:%d x:%d\n",
            System.out.println("White:  Mat:"
                    + board.pmaterial[white] + "/" + board.material[white]
                    + "  P:" + wp + "  N:" + Eval.ScoreN(white)
                    + "  B:" + Eval.ScoreB(white) + "  R:" + Eval.ScoreR(white)
                    + "  Q:" + Eval.ScoreQ(white) + "  K:" + wk
                    + "  Dev:" + Eval.ScoreDev(white) + "  h:" + hunged[white]
                    + " x: " + ExchCnt[white]);
            // System.out.println ("Black:  Mat:%4d/%4d  P:%d  N:%d  B:%d  R:%d  Q:%d  K:%d  Dev:%d  h:%d x:%d\n",
            System.out.println("Black:  Mat:"
                    + board.pmaterial[black] + "/" + board.material[black]
                    + "  P:" + bp + "  N:" + Eval.ScoreN(black)
                    + "  B:" + Eval.ScoreB(black) + "  R:" + Eval.ScoreR(black)
                    + "  Q:" + Eval.ScoreQ(black) + "  K:" + bk
                    + "  Dev:" + Eval.ScoreDev(black) + "  h:" + hunged[black]
                    + " x: " + ExchCnt[black]);
            s = (Eval.EvaluateDraw() ? DRAWSCORE : Eval.Evaluate(-INFINITY, INFINITY));
            System.out.println("Phase: " + PHASE() + "\tscore = " + s);
            System.out.println();
        } else if (tokeneq(token[1], "game"))
            Output.ShowGame();
        else if (tokeneq(token[1], "pin")) {
            /*BitBoard*/long b;
            Atak.GenAtaks();
            b = Atak.FindPins();
            Output.ShowBitBoard(b);
        }
    }

    static void cmd_test()
            /*************************************************************************
             *
             *  The test command driver section.
             *
             *************************************************************************/ {
        if (tokeneq(token[1], "movelist"))
            Test.TestMoveList();
        else if (tokeneq(token[1], "capture"))
            Test.TestCaptureList();
        else if (tokeneq(token[1], "movegenspeed"))
            Test.TestMoveGenSpeed();
        else if (tokeneq(token[1], "capturespeed"))
            Test.TestCaptureGenSpeed();
        else if (tokeneq(token[1], "eval"))
            Test.TestEval();
        else if (tokeneq(token[1], "evalspeed"))
            Test.TestEvalSpeed();
    }

/*
 * This is more or less copied from the readme, and the
 * parser is not very clever, so the lines containing
 * command names should not be indented, the lines with
 * explanations following them should be indented. Do not
 * use tabs for indentation, only spaces. CAPITALS are
 * reserved for parameters in the command names. The
 * array must be terminated by two NULLs.
 *
 * This one should be integrated in the method table.
 * (Very much like docstrings in Lisp.)
 */

    static final String[] helpstr = {
        "^C",
        " Typically the interrupt key stops a search in progress,",
        " makes the move last considered best and returns to the",
        " command prompt",
        "quit",
        " quit the program.",
        "exit",
        " In analysis mode this stops analysis, otherwise it quits the program.",
        "help",
        " Produces a help blurb corresponding to this list of commands.",
        "book",
        " add - compiles book.dat from book.pgn",
        " on - enables use of book",
        " off - disables use of book",
        " worst - play worst move from book",
        " best - play best move from book",
        " prefer - default, same as 'book on'",
        " random - play any move from book",
        "version",
        " prints out the version of this program",
        "pgnsave FILENAME",
        " saves the game so far to the file from memory",
        "pgnload FILENAME",
        " loads the game in the file into memory",
        "force",
        "manual",
        " Makes the program stop moving. You may now enter moves",
        " to reach some position in the future.",
        " ",
        "white",
        " Program plays white",
        "black",
        " Program plays black",
        "go",
        " Computer takes whichever side is on move and begins its",
        " thinking immediately",
        "post",
        " Arranges for verbose thinking output showing variation, score,",
        " time, depth, etc.",
        "nopost",
        " Turns off verbose thinking output",
        "name NAME",
        " Lets you input your name. Also writes the log.nnn and a",
        " corresponding game.nnn file. For details please see",
        " auxillary file format sections.",
        "result",
        " Mostly used by Internet Chess server.",
        "activate",
        " This command reactivates a game that has been terminated automatically",
        " due to checkmate or no more time on the clock. However, it does not",
        " alter those conditions. You would have to undo a move or two or",
        " add time to the clock with level or time in that case.",
        "rating COMPUTERRATING OPPONENTRATING",
        " Inputs the estimated rating for computer and for its opponent",
        "new",
        " Sets up new game (i.e. positions in original positions)",
        "time",
        " Inputs time left in game for computer in hundredths of a second.",
        " Mostly used by Internet Chess server.",
        "hash",
        " on - enables using the memory hash table to speed search",
        " off - disables the memory hash table",
        "hashsize N",
        " Sets the hash table to permit storage of N positions",
        "null",
        " on - enables using the null move heuristic to speed search",
        " off - disables using the null move heuristic",
        "xboard",
        " on - enables use of xboard/winboard",
        " off - disables use of xboard/winboard",
        "depth N",
        " Sets the program to look N ply (half-moves) deep for every",
        " search it performs. If there is a checkmate or other condition",
        " that does not allow that depth, then it will not be ",
        "level MOVES MINUTES INCREMENT",
        " Sets time control to be MOVES in MINUTES with each move giving",
        " an INCREMENT (in seconds, i.e. Fischer-style clock).",
        "load",
        "epdload",
        " Loads a position in EPD format from disk into memory.",
        "save",
        "epdsave",
        " Saves game position into EPD format from memory to disk.",
        "switch",
        " Switches side to move",
        "solve FILENAME",
        "solveepd FILENAME",
        " Solves the positions in FILENAME",
        "remove",
        " Backs up two moves in game history",
        "undo",
        " Backs up one move in game history",
        "usage",
        " Display command line syntax",
        "show",
        " board - displays the current board",
        " time - displays the time settings",
        " moves - shows all moves using one call to routine",
        " escape - shows moves that escape from check using one call to routine",
        " noncapture - shows non-capture moves",
        " capture - shows capture moves",
        " eval [or score] - shows the evaluation per piece and overall",
        " game - shows moves in game history",
        " pin - shows pinned pieces",
        "test",
        " movelist - reads in an epd file and shows legal moves for its entries",
        " capture - reads in an epd file and shows legal captures for its entries",
        " movegenspeed - tests speed of move generator",
        " capturespeed - tests speed of capture move generator",
        " eval - reads in an epd file and shows evaluation for its entries",
        " evalspeed tests speed of the evaluator",
        "bk",
        " show moves from opening book."
    };

    static void cmd_help()
            /**************************************************************************
             *
             *  Display all the help commands.
             *
             **************************************************************************/ {
        if (token[1] != null && token[1].length() > 0) {
            for (int i = 0; i < helpstr.length; i++) {
                String s = helpstr[i];
                if (s.equals(token[1])) {
                    ofp.println(s);
                    for (i++; i < helpstr.length && helpstr[i].startsWith(" "); i++)
                        ofp.println(helpstr[i]);
                    return;
                }
            }
            System.out.println("Help for command " + token[1] + " not found");
        }
        System.out.println("List of commands: (help COMMAND to get more help)\n");
        int count = 0;
        for (int i = 0; i < helpstr.length; i++) {
            if (helpstr[i].startsWith(" "))
                continue;
            count += helpstr[i].length() + 4;    // approx
            System.out.print(helpstr[i] + " \t");
            if (count > 60) {
                count = 0;
                System.out.println();
            }
        }
        System.out.println();
    }

/*
 * Try a method table, one could also include the documentation
 * strings here
 */



/* List commands we don't implement to avoid illegal moving them */

    static Map commands = new HashMap();

    static {
        commands.put("?", "cmd_movenow");
        commands.put("accepted", "cmd_accepted");
        commands.put("activate", "cmd_activate");
        commands.put("analyze", "cmd_analyze");
        commands.put("bk", "cmd_bk");
        commands.put("black", "cmd_black");
        commands.put("book", "cmd_book");
        commands.put("computer", "cmd_computer");
        commands.put("depth", "cmd_depth");
        commands.put("draw", "cmd_draw");
        commands.put("easy", "cmd_easy");
        commands.put("edit", "cmd_edit");
        commands.put("epd", "cmd_epd");
        commands.put("epdload", "cmd_load");
        commands.put("epdsave", "cmd_save");
        commands.put("exit", "cmd_exit");
        commands.put("force", "cmd_force");
        commands.put("go", "cmd_go");
        commands.put("hard", "cmd_hard");
        commands.put("hash", "cmd_hash");
        commands.put("hashsize", "cmd_hashsize");
        commands.put("help", "cmd_help");
        commands.put("hint", "cmd_hint");
        commands.put("level", "cmd_level");
        commands.put("list", "cmd_list");
        commands.put("load", "cmd_load");
        commands.put("manual", "cmd_manual");
        commands.put("name", "cmd_name");
        commands.put("new", "cmd_new");
        commands.put("nopost", "cmd_nopost");
        commands.put("null", "cmd_null");
        commands.put("otim", "cmd_otim");
        commands.put("pgnload", "cmd_pgnload");
        commands.put("pgnsave", "cmd_pgnsave");
        commands.put("ping", "cmd_ping");
        commands.put("post", "cmd_post");
        commands.put("protover", "cmd_protover");
        commands.put("quit", "cmd_quit");
        commands.put("random", "cmd_random");
        commands.put("rating", "cmd_rating");
        commands.put("rejected", "cmd_rejected");
        commands.put("remove", "cmd_remove");
        commands.put("result", "cmd_result");
        commands.put("save", "cmd_save");
        commands.put("setboard", "cmd_setboard");
        commands.put("show", "cmd_show");
        commands.put("solve", "cmd_solve");
        commands.put("solveepd", "cmd_solve");
        commands.put("st", "cmd_st");
        commands.put("switch", "cmd_switch");
        commands.put("test", "cmd_test");
        commands.put("time", "cmd_time");
        commands.put("undo", "cmd_undo");
        commands.put("usage", "cmd_usage");
        commands.put("variant", "cmd_variant");
        commands.put("version", "cmd_version");
        commands.put("white", "cmd_white");
        commands.put("xboard", "cmd_xboard");
    }

    static void parse_input()
            /*************************************************************************
             *
             *  This is the main user command interface driver.
             *
             *************************************************************************/ {

        Debug.dbg_printf("parse_input() called, inputstr = *" + Input.inputstr + "*");

        split_input();

        String methodName = (String) commands.get(token[0]);
        if (methodName != null) {
            try {
                Method method = Cmd.class.getDeclaredMethod(methodName, null);
                method.invoke(null, null);
                return;
            } catch (NoSuchMethodException e) {
                // not a known command -- fall thorugh
            } catch (IllegalAccessException e) {
                e.printStackTrace();
            } catch (InvocationTargetException e) {
                e.printStackTrace();
            }
        }


        /* OK, no known command, this should be a move */
        int ptr = Move.ValidateMove(token[0]);
        if (ptr >= 0) {
            Move.SANMove(Tree[ptr].move, 1);
            Tree[ptr].move = Move.MakeMove(board.side, Tree[ptr].move);
            Game[GameCnt].SANmv = SANmv;
            System.out.println(GameCnt / 2 + 1 + "." + token[0]);
            if (!ofpIsStdout) {
                ofp.println(GameCnt / 2 + 1 + "." + token[0]);
                ofp.flush();
            }
            if ((flags & XBOARD) == 0)
                Output.ShowBoard();
            SetFlag(THINK);
        } else {
            /*
             * Must Output Illegal move to prevent Xboard accepting illegal
             * en passant captures and other subtle mistakes
             */
            System.out.println("Illegal move: " + token[0]);
            System.out.flush();
        }
    }
}
