/* This file is part of chessbox_gnu4j, a Java port of GNU Chess (5.0.7).

   Copyright (c) 2005 Xan Gregg (Java port)
   Copyright (c) 1999-2002 Free Software Foundation, Inc.

   chessbox_gnu4j is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 or (at your option)
   any later version.

   chessbox_gnu4j is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GNU Chess; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.

   Contact Info:
     bug-gnu-chess@gnu.org
     xan.gregg@pobox.com
*/
package org.chessbox.gnu4j;

import java.io.PrintWriter;
import java.util.Random;

/**
 * Constants from common.h.
 * <p/>
 * Created by Xan Gregg.
 * Date: Apr 22, 2005
 */
public class Common {

    // convert Java boolean to C boolean (0 or 1)
    public static int boolbit(boolean b) {
        return b ? 1 : 0;
    }

/* constants for move description */
    public static final int KNIGHTPRM = 0x00002000;
    public static final int BISHOPPRM = 0x00003000;
    public static final int ROOKPRM = 0x00004000;
    public static final int QUEENPRM = 0x00005000;
    public static final int PROMOTION = 0x00007000;
    public static final int PAWNCAP = 0x00008000;
    public static final int KNIGHTCAP = 0x00010000;
    public static final int BISHOPCAP = 0x00018000;
    public static final int ROOKCAP = 0x00020000;
    public static final int QUEENCAP = 0x00028000;
    public static final int CAPTURE = 0x00038000;
    public static final int NULLMOVE = 0x00100000;
    public static final int CASTLING = 0x00200000;
    public static final int ENPASSANT = 0x00400000;
    public static final int MOVEMASK = CASTLING | ENPASSANT | PROMOTION | 0x0FFF;

    public static final int white = 0;
    public static final int black = 1;
    public static final int falsebit = 0;
    public static final int truebit = 1;
    public static final int ks = 0;
    public static final int qs = 1;
    public static final int INFINITY = 32767;
    public static final int MATE = 32767;

    public static final int MATESCORE(int a) {
        return boolbit(a > MATE - 255 || a < -MATE + 255);
    }

/* Material values */
    public static final int ValueP = 100;
    public static final int ValueN = 350;
    public static final int ValueB = 350;
    public static final int ValueR = 550;
    public static final int ValueQ = 1100;
    public static final int ValueK = 2000;

    public static final int empty = 0;
    public static final int pawn = 1;
    public static final int knight = 2;
    public static final int bishop = 3;
    public static final int rook = 4;
    public static final int queen = 5;
    public static final int king = 6;
    public static final int bpawn = 7;

    // enum Square
    public static final int A1 = 0;
    public static final int B1 = 1;
    public static final int C1 = 2;
    public static final int D1 = 3;
    public static final int E1 = 4;
    public static final int F1 = 5;
    public static final int G1 = 6;
    public static final int H1 = 7;
    public static final int A2 = 8;
    public static final int B2 = 9;
    public static final int C2 = 10;
    public static final int D2 = 11;
    public static final int E2 = 12;
    public static final int F2 = 13;
    public static final int G2 = 14;
    public static final int H2 = 15;
    public static final int A3 = 16;
    public static final int B3 = 17;
    public static final int C3 = 18;
    public static final int D3 = 19;
    public static final int E3 = 20;
    public static final int F3 = 21;
    public static final int G3 = 22;
    public static final int H3 = 23;
    public static final int A4 = 24;
    public static final int B4 = 25;
    public static final int C4 = 26;
    public static final int D4 = 27;
    public static final int E4 = 28;
    public static final int F4 = 29;
    public static final int G4 = 30;
    public static final int H4 = 31;
    public static final int A5 = 32;
    public static final int B5 = 33;
    public static final int C5 = 34;
    public static final int D5 = 35;
    public static final int E5 = 36;
    public static final int F5 = 37;
    public static final int G5 = 38;
    public static final int H5 = 39;
    public static final int A6 = 40;
    public static final int B6 = 41;
    public static final int C6 = 42;
    public static final int D6 = 43;
    public static final int E6 = 44;
    public static final int F6 = 45;
    public static final int G6 = 46;
    public static final int H6 = 47;
    public static final int A7 = 48;
    public static final int B7 = 49;
    public static final int C7 = 50;
    public static final int D7 = 51;
    public static final int E7 = 52;
    public static final int F7 = 53;
    public static final int G7 = 54;
    public static final int H7 = 55;
    public static final int A8 = 56;
    public static final int B8 = 57;
    public static final int C8 = 58;
    public static final int D8 = 59;
    public static final int E8 = 60;
    public static final int F8 = 61;
    public static final int G8 = 62;
    public static final int H8 = 63;

    // enum File
    public static final int A_FILE = 0;
    public static final int B_FILE = 1;
    public static final int C_FILE = 2;
    public static final int D_FILE = 3;
    public static final int E_FILE = 4;
    public static final int F_FILE = 5;
    public static final int G_FILE = 6;
    public static final int H_FILE = 7;

/*
 * Return values (errorcodes) for the book routines,
 * maybe one should have a global enum of errorcodes
 */
    public static final int BOOK_SUCCESS = 0;
    public static final int BOOK_EFORMAT = 1;  /* Wrong format (e.g. produced by older version) */
    public static final int BOOK_EMIDGAME = 2; /* Move is past the opening book's move limit */
    public static final int BOOK_EIO = 3;      /* I/O error, e.g. caused by wrong permissions */
    public static final int BOOK_EFULL = 4;    /* Book hash is full, new position was not added. */
    public static final int BOOK_ENOBOOK = 5;  /* No book present */
    public static final int BOOK_ENOMOVES = 6; /* No moves found (in BookQuery() only) */
    public static final int BOOK_ENOMEM = 7;   /* Memory allocation failed */


/* Error codes for ParseEPD */
    public static final int EPD_SUCCESS = 0;
    public static final int EPD_ERROR = 1;


/*
 * Status variable used by the input thread to signal
 * pending input. Thought about using flags for this
 * but this can be refined and is conceptually different
 * from flags.
 */
    public static final int INPUT_NONE = 0;
    public static final int INPUT_AVAILABLE = 1;

/*  Game flags */;
    public static final int QUIT = 0x0001;
    public static final int TESTT = 0x0002;
    public static final int THINK = 0x0004;
    public static final int MANUAL = 0x0008;
    public static final int TIMEOUT = 0x0010;
    public static final int SPARE1 = 0x0020;
    public static final int ENDED = 0x0040;
    public static final int USEHASH = 0x0080;
    public static final int SOLVE = 0x0100;
    public static final int USENULL = 0x0200;
    public static final int XBOARD = 0x0400;
    public static final int TIMECTL = 0x0800;
    public static final int POST = 0x1000;
    public static final int PONDER = 0x2000; /* We are in pondering (during search) */
    public static final int HARD = 0x4000;/* Pondering is turned on */
    public static final int ANALYZE = 0x8000; /* In ANALYZE mode */

/*  Node types  */;
    public static final int PV = 0;
    public static final int ALL = 1;
    public static final int CUT = 2;

/*  Transposition table flags  */
    public static final int EXACTSCORE = 1;
    public static final int LOWERBOUND = 2;
    public static final int UPPERBOUND = 3;
    public static final int POORDRAFT = 4;
    public static final int QUIESCENT = 5;
    public static final int STICKY = 8;

/*  Book modes */
    public static final int BOOKOFF = 0;
    public static final int BOOKRAND = 1;
    public static final int BOOKBEST = 2;
    public static final int BOOKWORST = 3;
    public static final int BOOKPREFER = 4;

/*  The various phases during move selection  */
    public static final int PICKHASH = 1;
    public static final int PICKGEN1 = 2;
    public static final int PICKCAPT = 3;
    public static final int PICKKILL1 = 4;
    public static final int PICKKILL2 = 5;
    public static final int PICKGEN2 = 6;
    public static final int PICKHIST = 7;
    public static final int PICKREST = 8;
    public static final int PICKCOUNTER = 9;

    public static final int MAXTREEDEPTH = 2000;
    public static final int MAXPLYDEPTH = 65;
    public static final int MAXGAMEDEPTH = 600;

/*
   Smaller HASHSLOT defaults 20011017 to improve blitz play
   and make it easier to run on old machines
*/
    public static final int HASHSLOTS = 1024*32;
    public static final int PAWNSLOTS = 512*16;

/*  Some special BitBoards  */
    public static /*BitBoard*/long NULLBITBOARD = 0x0000000000000000L;
    public static /*BitBoard*/long WHITESQUARES = 0x55AA55AA55AA55AAL;
    public static /*BitBoard*/long BLACKSQUARES = 0xAA55AA55AA55AA55L;
    public static /*BitBoard*/long CENTRESQUARES = 0x0000001818000000L;
    public static /*BitBoard*/long COMPUTERHALF = 0xFFFFFFFF00000000L;
    public static /*BitBoard*/long OPPONENTHALF = 0x00000000FFFFFFFFL;

    public static /*BitBoard*/long[] BitPosArray = new /*BitBoard*/long[64];
    public static /*BitBoard*/long[] NotBitPosArray = new /*BitBoard*/long[64];

    // globals
    public static short[][] distance = new short[64][64];
    public static short[][] taxicab = new short[64][64];
    public static byte[] lzArray = new byte[65536];   // was unsigned
    public static /*BitBoard*/long[][] DistMap = new /*BitBoard*/long[64][8];
    public static /*BitBoard*/long[][] MoveArray = new /*BitBoard*/long[8][64];
    public static /*BitBoard*/long[][] Ray = new /*BitBoard*/long[64][8];
    public static /*BitBoard*/long[][] FromToRay = new /*BitBoard*/long[64][64];
    public static /*BitBoard*/long[] RankBit = new /*BitBoard*/long[8];
    public static /*BitBoard*/long[] FileBit = new /*BitBoard*/long[8];
    public static /*BitBoard*/long[][] Ataks = new /*BitBoard*/long[2][7];
    public static /*BitBoard*/long[][] PassedPawnMask = new /*BitBoard*/long[2][64];
    public static /*BitBoard*/long[] IsolaniMask = new /*BitBoard*/long[8];
    public static /*BitBoard*/long[][] SquarePawnMask = new /*BitBoard*/long[2][64];
    public static /*BitBoard*/long[][] Rook00Atak = new /*BitBoard*/long[64][256];
    public static /*BitBoard*/long[][] Rook90Atak = new /*BitBoard*/long[64][256];
    public static /*BitBoard*/long[][] Bishop45Atak = new /*BitBoard*/long[64][256];
    public static /*BitBoard*/long[][] Bishop315Atak = new /*BitBoard*/long[64][256];
    public static /*BitBoard*/long pinned;
    public static /*BitBoard*/long[] rings = new /*BitBoard*/long[4];
    public static /*BitBoard*/long[] boxes = new /*BitBoard*/long[2];
    public static /*BitBoard*/long[] stonewall = new /*BitBoard*/long[2];
    public static /*BitBoard*/long[] pieces = new /*BitBoard*/long[2];
    public static /*BitBoard*/long[] mask_kr_trapped_w = new /*BitBoard*/long[3];
    public static /*BitBoard*/long[] mask_kr_trapped_b = new /*BitBoard*/long[3];
    public static /*BitBoard*/long[] mask_qr_trapped_w = new /*BitBoard*/long[3];
    public static /*BitBoard*/long[] mask_qr_trapped_b = new /*BitBoard*/long[3];
    public static /*BitBoard*/long[] boardhalf = new /*BitBoard*/long[2];
    public static /*BitBoard*/long[] boardside = new /*BitBoard*/long[2];
    public static short[][] directions = new short[64][64];
    public static byte[] BitCount = new byte[65536];   // was unsigned char
    public static Leaf[] Tree = new Leaf[MAXTREEDEPTH];
    public static int[] TreePtr = new int[MAXPLYDEPTH];   // index into Tree; was Leaf*
    public static int RootPV;
    public static GameRec[] Game = new GameRec[MAXGAMEDEPTH];
    public static int GameCnt;
    public static int RealGameCnt;
    public static short RealSide;
    public static int computer;
    public static int flags;  // was unsigned
    public static int preanalyze_flags;  // was unsigned
    public static int[] cboard = new int[64];
    public static int[] Mvboard = new int[64];
    public static Board board = new Board();
    public static /*HashType*/long[][][] hashcode = new /*HashType*/long[2][7][64];
    public static /*HashType*/long[] ephash = new /*HashType*/long[64];
    public static /*HashType*/long WKCastlehash;
    public static /*HashType*/long WQCastlehash;
    public static /*HashType*/long BKCastlehash;
    public static /*HashType*/long BQCastlehash;
    public static /*HashType*/long Sidehash;
    public static /*HashType*/long HashKey;
    public static /*HashType*/long PawnHashKey;
    public static HashSlot[][] HashTab = new HashSlot[2][];   // was HashSlot *HashTab[2];
    public static PawnSlot[][] PawnTab = new PawnSlot[2][];   // was PawnSlot *PawnTab[2];
    public static int Idepth;
    public static int SxDec;
    public static int Game50;
    public static int[] lazyscore = new int[2];
    public static int[] maxposnscore = new int[2];
    public static int rootscore;
    public static int lastrootscore;
    public static long GenCnt;  // was unsigned
    public static long NodeCnt;  // was unsigned
    public static long QuiesCnt;  // was unsigned
    public static long EvalCnt;  // was unsigned
    public static long EvalCall;  // was unsigned
    public static long ChkExtCnt;  // was unsigned
    public static long OneRepCnt;  // was unsigned
    public static long RcpExtCnt;  // was unsigned
    public static long PawnExtCnt;  // was unsigned
    public static long HorzExtCnt;  // was unsigned
    public static long ThrtExtCnt;  // was unsigned
    public static long KingExtCnt;  // was unsigned
    public static long NullCutCnt;  // was unsigned
    public static long FutlCutCnt;  // was unsigned
    public static long RazrCutCnt;  // was unsigned
    public static long TotalGetHashCnt;  // was unsigned
    public static long GoodGetHashCnt;  // was unsigned
    public static long TotalPutHashCnt;  // was unsigned
    public static long CollHashCnt;  // was unsigned
    public static long TotalPawnHashCnt;  // was unsigned
    public static long GoodPawnHashCnt;  // was unsigned
    public static long RepeatCnt;  // was unsigned
    public static int HashSize;  // was unsigned
    public static long TTHashMask;  // was unsigned
    public static long PHashMask;  // was unsigned
    public static String SANmv;
    public static long[][] history = new long[2][4096];  // was unsigned
    public static int[] killer1 = new int[MAXPLYDEPTH];
    public static int[] killer2 = new int[MAXPLYDEPTH];
    public static int[] ChkCnt = new int[MAXPLYDEPTH];
    public static int[] ThrtCnt = new int[MAXPLYDEPTH];
    public static String id;    // was char[32]
    public static String solution;  // was char[64]
    public static double ElapsedTime;
    public static /*TimeVal*/long StartTime;
    public static double SearchTime;
    public static int SearchDepth;
    public static int[] MoveLimit = new int[2];
    public static double[] TimeLimit = new double[2];
    public static int TCMove;
    public static int TCinc;
    public static double TCTime;
    public static int[] castled = new int[2];
    public static int[] hunged = new int[2];
    public static int phase;
    public static int[] Hashmv = new int[MAXPLYDEPTH];
    public static int RootPieces; // short
    public static int RootPawns; // short
    public static int RootMaterial; // short
    public static int RootAlpha; // short
    public static int RootBeta; // short
    public static short[] pickphase = new short[MAXPLYDEPTH];
    public static boolean[] InChk = new boolean[MAXPLYDEPTH];
    public static boolean[][] KingThrt = new boolean[2][MAXPLYDEPTH];
    public static short threatmv;
    public static byte threatply;   // was uint8_t
    public static short[] KingSafety = new short[2];
    public static short[] pscore = new short[64];
    public static short bookmode;
    public static short bookfirstlast;

    public static String progname;
    public static PrintWriter ofp;
    public static boolean ofpIsStdout;
    public static int myrating, opprating, TCionc;
    public static boolean suddendeath;
    public static String name; // was char[50]
    public static int computerplays;		/* Side computer is playing */
    public static boolean wasbookmove;		/* True if last move was book move */
    public static int nmovesfrombook;		/* Number of moves since last book move */
    public static int newpos, existpos;		/* For book statistics */
    public static double maxtime;		/* Max time for the next searched move */
    public static int n;		/* Last mobility returned by CTL */
    public static int[] ExchCnt = new int[2];	/* How many exchanges? */
    public static boolean bookloaded = false;  	/* Is the book loaded already into memory? */

//    public static Random random = new Random(new GregorianCalendar().getTimeInMillis());
    public static Random random = new Random(213);

    public static int[] slider = new int[]{0, 0, 0, 1, 1, 1, 0, 0};
    public static int[] Value = new int[]{0, ValueP, ValueN, ValueB, ValueR, ValueQ, ValueK};
    public static int[] range = new int[]{0, 0, 0, 1, 1, 1, 0, 0};
    public static int[] ptype = new int[]{pawn, bpawn};

    public static String[] algbr = new String[]
    {"a1", "b1", "c1", "d1", "e1", "f1", "g1", "h1",
     "a2", "b2", "c2", "d2", "e2", "f2", "g2", "h2",
     "a3", "b3", "c3", "d3", "e3", "f3", "g3", "h3",
     "a4", "b4", "c4", "d4", "e4", "f4", "g4", "h4",
     "a5", "b5", "c5", "d5", "e5", "f5", "g5", "h5",
     "a6", "b6", "c6", "d6", "e6", "f6", "g6", "h6",
     "a7", "b7", "c7", "d7", "e7", "f7", "g7", "h7",
     "a8", "b8", "c8", "d8", "e8", "f8", "g8", "h8"
    };
    public static char[] algbrfile = {'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h'};
    public static char[] algbrrank = {'1', '2', '3', '4', '5', '6', '7', '8'};

    public static char[] notation = {' ', 'P', 'N', 'B', 'R', 'Q', 'K'};
    public static char[] lnotation = {' ', 'p', 'n', 'b', 'r', 'q', 'k'};

    public static short[] Shift00 = new short[]
    {56, 56, 56, 56, 56, 56, 56, 56,
     48, 48, 48, 48, 48, 48, 48, 48,
     40, 40, 40, 40, 40, 40, 40, 40,
     32, 32, 32, 32, 32, 32, 32, 32,
     24, 24, 24, 24, 24, 24, 24, 24,
     16, 16, 16, 16, 16, 16, 16, 16,
     8, 8, 8, 8, 8, 8, 8, 8,
     0, 0, 0, 0, 0, 0, 0, 0
    };

    public static int[] r90 = new int[]
    {A8, A7, A6, A5, A4, A3, A2, A1,
     B8, B7, B6, B5, B4, B3, B2, B1,
     C8, C7, C6, C5, C4, C3, C2, C1,
     D8, D7, D6, D5, D4, D3, D2, D1,
     E8, E7, E6, E5, E4, E3, E2, E1,
     F8, F7, F6, F5, F4, F3, F2, F1,
     G8, G7, G6, G5, G4, G3, G2, G1,
     H8, H7, H6, H5, H4, H3, H2, H1};

    public static short[] Shift90 = new short[]
    {0, 8, 16, 24, 32, 40, 48, 56,
     0, 8, 16, 24, 32, 40, 48, 56,
     0, 8, 16, 24, 32, 40, 48, 56,
     0, 8, 16, 24, 32, 40, 48, 56,
     0, 8, 16, 24, 32, 40, 48, 56,
     0, 8, 16, 24, 32, 40, 48, 56,
     0, 8, 16, 24, 32, 40, 48, 56,
     0, 8, 16, 24, 32, 40, 48, 56
    };

    public static int[] r45 = new int[]
    {E4, F3, H2, C2, G1, D1, B1, A1,
     E5, F4, G3, A3, D2, H1, E1, C1,
     D6, F5, G4, H3, B3, E2, A2, F1,
     B7, E6, G5, H4, A4, C3, F2, B2,
     G7, C7, F6, H5, A5, B4, D3, G2,
     C8, H7, D7, G6, A6, B5, C4, E3,
     F8, D8, A8, E7, H6, B6, C5, D4,
     H8, G8, E8, B8, F7, A7, C6, D5};

    public static short[] Shift45 = new short[]
    {28, 36, 43, 49, 54, 58, 61, 63,
     21, 28, 36, 43, 49, 54, 58, 61,
     15, 21, 28, 36, 43, 49, 54, 58,
     10, 15, 21, 28, 36, 43, 49, 54,
     6, 10, 15, 21, 28, 36, 43, 49,
     3, 6, 10, 15, 21, 28, 36, 43,
     1, 3, 6, 10, 15, 21, 28, 36,
     0, 1, 3, 6, 10, 15, 21, 28};

    public static int[] Mask45 = new int[]
    {0xFF, 0x7F, 0x3F, 0x1F, 0x0F, 0x07, 0x03, 0x01,
     0x7F, 0xFF, 0x7F, 0x3F, 0x1F, 0x0F, 0x07, 0x03,
     0x3F, 0x7F, 0xFF, 0x7F, 0x3F, 0x1F, 0x0F, 0x07,
     0x1F, 0x3F, 0x7F, 0xFF, 0x7F, 0x3F, 0x1F, 0x0F,
     0x0F, 0x1F, 0x3F, 0x7F, 0xFF, 0x7F, 0x3F, 0x1F,
     0x07, 0x0F, 0x1F, 0x3F, 0x7F, 0xFF, 0x7F, 0x3F,
     0x03, 0x07, 0x0F, 0x1F, 0x3F, 0x7F, 0xFF, 0x7F,
     0x01, 0x03, 0x07, 0x0F, 0x1F, 0x3F, 0x7F, 0xFF};

    public static int[] r315 = new int[]
    {A1, C1, F1, B2, G2, E3, D4, D5,
     B1, E1, A2, F2, D3, C4, C5, C6,
     D1, H1, E2, C3, B4, B5, B6, A7,
     G1, D2, B3, A4, A5, A6, H6, F7,
     C2, A3, H3, H4, H5, G6, E7, B8,
     H2, G3, G4, G5, F6, D7, A8, E8,
     F3, F4, F5, E6, C7, H7, D8, G8,
     E4, E5, D6, B7, G7, C8, F8, H8};

    public static short[] Shift315 = new short[]
    {63, 61, 58, 54, 49, 43, 36, 28,
     61, 58, 54, 49, 43, 36, 28, 21,
     58, 54, 49, 43, 36, 28, 21, 15,
     54, 49, 43, 36, 28, 21, 15, 10,
     49, 43, 36, 28, 21, 15, 10, 6,
     43, 36, 28, 21, 15, 10, 6, 3,
     36, 28, 21, 15, 10, 6, 3, 1,
     28, 21, 15, 10, 6, 3, 1, 0};

    public static int[] Mask315 = new int[]
    {0x01, 0x03, 0x07, 0x0F, 0x1F, 0x3F, 0x7F, 0xFF,
     0x03, 0x07, 0x0F, 0x1F, 0x3F, 0x7F, 0xFF, 0x7F,
     0x07, 0x0F, 0x1F, 0x3F, 0x7F, 0xFF, 0x7F, 0x3F,
     0x0F, 0x1F, 0x3F, 0x7F, 0xFF, 0x7F, 0x3F, 0x1F,
     0x1F, 0x3F, 0x7F, 0xFF, 0x7F, 0x3F, 0x1F, 0x0F,
     0x3F, 0x7F, 0xFF, 0x7F, 0x3F, 0x1F, 0x0F, 0x07,
     0x7F, 0xFF, 0x7F, 0x3F, 0x1F, 0x0F, 0x07, 0x03,
     0xFF, 0x7F, 0x3F, 0x1F, 0x0F, 0x07, 0x03, 0x01};

    public static int[] rank6 = new int[]{5, 2};
    public static int[] rank7 = new int[]{6, 1};
    public static int[] rank8 = new int[]{7, 0};

/* Draw score can be used to penalise draws if desired */
    public static final int DRAWSCORE = 0;

    public static int MATERIAL(int side) {
        return board.material[side] - board.material[1 ^ side];
    }

    public static int PHASE() {
        return (8 - (board.material[white] + board.material[black]) / 1150);
    }

    public static long KEY(long a) {
        return a >>> 32;
    }

/*  Attack MACROS */

    public static /*BitBoard*/long BishopAttack(int sq) {
        return Bishop45Atak[sq][((int) ((board.blockerr45 >>> Shift45[sq]) & Mask45[sq]))] |
                Bishop315Atak[sq][((int) ((board.blockerr315 >>> Shift315[sq]) & Mask315[sq]))];
    }

    public static /*BitBoard*/long RookAttack(int sq) {
        return Rook00Atak[sq][((int) ((board.blocker >>> Shift00[sq]) & 0xFF))] |
                Rook90Atak[sq][((int) ((board.blockerr90 >>> Shift90[sq]) & 0xFF))];
    }

    public static /*BitBoard*/long QueenAttack(int sq) {
        return BishopAttack(sq) | RookAttack(sq);
    }


/*
    public static int SET(int a, int b) {
        return a | b;
    }

    public static int CLEAR(int a, int b) {
        return a & ~b;
    }
*/

    public static void SetFlag(int flagBit) {
        flags |= flagBit;
    }

    public static void ClearFlag(int flagBit) {
        flags &= ~flagBit;
    }

    public static int RANK(int move) {
        return move >> 3;
    }

    public static int ROW(int move) {
        return move & 7;
    }

    public static int MIN(int a, int b) {
        return Math.min(a, b);
    }

    public static int MAX(int a, int b) {
        return Math.max(a, b);
    }


/*    public static int leadz(long x) {
        // from HD nlz1a()
        int n;

        // if (x == 0) return(32);
        if (x <= 0) return (int) ((~x >> 57) & 64);
        n = 1;
        if (x >>> 32 == 0) {
            n += 32;
            x <<= 32;
        }
        if (x >>> 48 == 0) {
            n += 16;
            x <<= 16;
        }
        if (x >>> 56 == 0) {
            n += 8;
            x <<= 8;
        }
        if (x >>> 60 == 0) {
            n += 4;
            x <<= 4;
        }
        if (x >>> 62 == 0) {
            n += 2;
            x <<= 2;
        }
        n -= (int) (x >> 63);
        return n;
    }

    public static int leadz(int x) {
        // from HD nlz1a()
        int n;

        // if (x == 0) return(32);
        if (x <= 0) return (~x >> 26) & 32;
        n = 1;
        if (x >>> 16 == 0) {
            n += 16;
            x <<= 16;
        }
        if (x >>> 24 == 0) {
            n += 8;
            x <<= 8;
        }
        if (x >>> 28 == 0) {
            n += 4;
            x <<= 4;
        }
        if (x >>> 30 == 0) {
            n += 2;
            x <<= 2;
        }
        n -= (x >> 31);
        return n;
    }

    public static int trailz(int b) {
        return leadz((b) & ((~b) + 1)); // xfixme looks suspicious
    }*/

/* Move Descriptions (moves) are represented internally as integers.
 * The lowest 6 bits define the destination ("TO") square, and
 * the next lowest 6 bits define the source ("FROM") square,
 * using the values defined by "enum Square" (0=a1, 63=h8).
 * Upper bits are used to identify other move information such as
 * a promotion (and to what), a capture (and of what),
 * CASTLING moves, and ENPASSANT moves; see the "constants for
 * move description" below for more information on these upper bits.
 */
    public static int PROMOTEPIECE(int a) {
        return ((a >> 12) & 0x0007);
    }

    public static int CAPTUREPIECE(int a) {
        return ((a >> 15) & 0x0007);
    }

    public static int TOSQ(int a) {
        return ((a) & 0x003F);
    }

    public static int FROMSQ(int a) {
        return ((a >> 6) & 0x003F);
    }

    public static int MOVE(int a, int b) {
//        if (a > 63 || b > 63 || a < 0 || b < 0) throw new NullPointerException();
        return a << 6 | b;
    }

/* constants for Board */;
    public static final int WKINGCASTLE = 0x0001;
    public static final int WQUEENCASTLE = 0x0002;
    public static final int BKINGCASTLE = 0x0004;
    public static final int BQUEENCASTLE = 0x0008;
    public static final int WCASTLE = WKINGCASTLE | WQUEENCASTLE;
    public static final int BCASTLE = BKINGCASTLE | BQUEENCASTLE;

    public static double round(double x, int digits) {
        double scale = Math.pow(10, digits);
        return Math.rint((x * scale + 0.5) / scale);
    }

    static String spaces = "                                      ";

    public static String pad(int value, int count) {
        String s = Integer.toString(value);
        if (count > s.length())
            return spaces.substring(0, count - s.length()) + s;
        else
            return s;
    }

    public static String pad(double value, int count) {
        String s = Double.toString(value);
        if (count > s.length())
            return spaces.substring(0, count - s.length()) + s;
        else
            return s;
    }

    public static String pad(String s, int count) {
        if (count > s.length())
            return spaces.substring(0, count - s.length()) + s;
        else
            return s;
    }
    public static String padRight(String s, int count) {
        if (count > s.length())
            return s + spaces.substring(0, count - s.length());
        else
            return s;
    }

    static String zeros = "00000000000000000000000";

    public static String pad0(int value, int count) {
        String s = Integer.toString(value);
        if (count > s.length())
            return zeros.substring(0, count - s.length()) + s;
        else
            return s;
    }

    public static String pad0(double value, int count) {
        String s = Double.toString(value);
        if (count > s.length())
            return zeros.substring(0, count - s.length()) + s;
        else
            return s;
    }

}
