/* This file is part of chessbox_gnu4j, a Java port of GNU Chess (5.0.7).

   Copyright (c) 2005 Xan Gregg (Java port)
   Copyright (c) 1999-2002 Free Software Foundation, Inc.

   chessbox_gnu4j is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 or (at your option)
   any later version.

   chessbox_gnu4j is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GNU Chess; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.

   Contact Info:
     bug-gnu-chess@gnu.org
     xan.gregg@pobox.com
*/
package org.chessbox.gnu4j;

/**
 * Created by Xan Gregg.
 * Date: May 7, 2005
 */
public class GenMove extends Common {
    static final short[] raybeg = {0, 0, 0, 0, 4, 0, 0};
    static final short[] rayend = {0, 0, 0, 4, 8, 8, 0};

    static int node;    //leaf *

    static void ADDMOVE(int a, int b,int c) {
        Tree[node].move = a << 6 | b | c;
        node++;
    }

    static void ADDPROMOTE(int a,int b) {
        ADDMOVE(a, b, QUEENPRM);
        ADDMOVE(a, b, KNIGHTPRM);
        ADDMOVE(a, b, ROOKPRM);
        ADDMOVE(a, b, BISHOPPRM);
    }

    /**
     * Convert a bitboard into a list of moves.  These are stored
     * into the tree.  f is the origin square.
     */
    static void BitToMove(int f, /*BitBoard*/long b) {
        int t;

        while (b != 0) {
            t = Util.leadz(b);
            b &= ~Util.longBit(t);  //CLEARBIT (b, t);
            ADDMOVE(f, t, 0);
        }
//        if (b != 0) {
//            int b16 = (int)(b >>> 48);
//            while( b16 != 0) {
//                t = lzArray[b16];
//                b16 &= ~Util.shortBit(t);  //CLEARBIT (b, t);
//                ADDMOVE(f, t, 0);
//            }
//
//            b16 = (int)(b >>> 32) & 0xffff;
//            while( b16 != 0) {
//                t = lzArray[b16];
//                b16 &= ~Util.shortBit(t);  //CLEARBIT (b, t);
//                ADDMOVE(f, t+16, 0);
//            }
//
//            b16 = (int)(b >>> 16) & 0xffff;
//            while( b16 != 0) {
//               t = lzArray[b16];
//               b16 &= ~Util.shortBit(t);  //CLEARBIT (b, t);
//               ADDMOVE(f, t+32, 0);
//           }
//
//            b16 = (int)(b) & 0xffff;
//            while( b16 != 0) {
//                t = lzArray[b16];
//                b16 &= ~Util.shortBit(t);  //CLEARBIT (b, t);
//                ADDMOVE(f, t+48, 0);
//            }
//        }
    }


    /**
     * My bitboard move generator.  Let's see how fast we can go!
     * This routine generates pseudo-legal moves.
     */
    static void GenMoves(int ply) {
        int side;
        int piece, sq, t, ep;
        /*BitBoard*/long b, c, d, e, friends, notfriends, blocker, notblocker;
        /*BitBoard*/long[] a;

        side = board.side;
        a = board.b[side];
        friends = board.friends[side];
        notfriends = ~friends;
        blocker = board.blocker;
        notblocker = ~blocker;
        node = TreePtr[ply + 1];
        ep = board.ep;

        /* Knight & King */
        for (piece = knight; piece <= king; piece += 4) {
            b = a[piece];
            while (b != 0) {
                sq = Util.leadz(b);
                b &= ~Util.longBit(sq);  //CLEARBIT (b, sq);
                BitToMove(sq, MoveArray[piece][sq] & notfriends);
            }
        }

        /* Bishops */
        b = a[bishop];
        while (b != 0) {
            sq = Util.leadz(b);
            b &= ~Util.longBit(sq);  //CLEARBIT (b, sq);
            d = BishopAttack(sq);
            BitToMove(sq, d & notfriends);
        }

        /* Rooks */
        b = a[rook];
        while (b != 0) {
            sq = Util.leadz(b);
            b &= ~Util.longBit(sq);  //CLEARBIT (b, sq);
            d = RookAttack(sq);
            BitToMove(sq, d & notfriends);
        }

        /* Queen */
        b = a[queen];
        while (b != 0) {
            sq = Util.leadz(b);
            b &= ~Util.longBit(sq);  //CLEARBIT (b, sq);
            d = QueenAttack(sq);
            BitToMove(sq, d & notfriends);
        }

        /*  White pawn moves  */
        e = (board.friends[1 ^ side] | (ep > -1 ? BitPosArray[ep] : NULLBITBOARD));
        if (side == white) {
            c = (a[pawn] >>> 8) & notblocker;        /* 1 square forward */
            while (c != 0) {
                t = Util.leadz(c);
                c &= ~Util.longBit(t);  //CLEARBIT (c, t);
                if (t >= 56)                /* promotion */ {
                    ADDPROMOTE(t - 8, t);
                } else
                    ADDMOVE(t - 8, t, 0);
            }

            b = a[pawn] & RankBit[1];            /* all pawns on 2nd rank */
            c = (b >>> 8) & notblocker;
            c = (c >>> 8) & notblocker;        /* 2 squares forward */
            while (c != 0) {
                t = Util.leadz(c);
                c &= ~Util.longBit(t);  //CLEARBIT (c, t);
                ADDMOVE(t - 16, t, 0);
            }

            b = a[pawn] & ~FileBit[0];        /* captures to the left */
            c = (b >>> 7) & e;
            while (c != 0) {
                t = Util.leadz(c);
                c &= ~Util.longBit(t);  //CLEARBIT (c, t);
                if (t >= 56)                /* promotion */ {
                    ADDPROMOTE(t - 7, t);
                } else if (ep == t) {
                    ADDMOVE(t - 7, t, ENPASSANT);
                } else {
                    ADDMOVE(t - 7, t, 0);
                }
            }

            b = a[pawn] & ~FileBit[7];        /* captures to the right */
            c = (b >>> 9) & e;
            while (c != 0) {
                t = Util.leadz(c);
                c &= ~Util.longBit(t);  //CLEARBIT (c, t);
                if (t >= 56)                /* promotion */ {
                    ADDPROMOTE(t - 9, t);
                } else if (ep == t) {
                    ADDMOVE(t - 9, t, ENPASSANT);
                } else {
                    ADDMOVE(t - 9, t, 0);
                }
            }
        }


        /*  Black pawn forward moves   */
        if (side == black) {
            c = (a[pawn] << 8) & notblocker;
            while (c != 0) {
                t = Util.leadz(c);
                c &= ~Util.longBit(t);  //CLEARBIT (c, t);
                if (t <= 7)                /* promotion */ {
                    ADDPROMOTE(t + 8, t);
                } else
                    ADDMOVE(t + 8, t, 0);
            }

            b = a[pawn] & RankBit[6];                 /* all pawns on 7th rank */
            c = (b << 8) & notblocker;
            c = (c << 8) & notblocker;
            while (c != 0) {
                t = Util.leadz(c);
                c &= ~Util.longBit(t);  //CLEARBIT (c, t);
                ADDMOVE(t + 16, t, 0);
            }

            b = a[pawn] & ~FileBit[7];        /* captures to the left */
            c = (b << 7) & e;
            while (c != 0) {
                t = Util.leadz(c);
                c &= ~Util.longBit(t);  //CLEARBIT (c, t);
                if (t <= 7)                /* promotion */ {
                    ADDPROMOTE(t + 7, t);
                } else if (ep == t) {
                    ADDMOVE(t + 7, t, ENPASSANT);
                } else {
                    ADDMOVE(t + 7, t, 0);
                }
            }

            b = a[pawn] & ~FileBit[0];        /* captures to the right */
            c = (b << 9) & e;
            while (c != 0) {
                t = Util.leadz(c);
                c &= ~Util.longBit(t);  //CLEARBIT (c, t);
                if (t <= 7)                /* promotion */ {
                    ADDPROMOTE(t + 9, t);
                } else if (ep == t) {
                    ADDMOVE(t + 9, t, ENPASSANT);
                } else {
                    ADDMOVE(t + 9, t, 0);
                }
            }
        }

        /* Castling code */
        b = board.b[side][rook];
        if (side == white && ((board.flag & WKINGCASTLE) != 0) && ((b & BitPosArray[H1]) != 0) &&
                (FromToRay[E1][G1] & blocker) == 0 &&
                !Atak.SqAtakd(E1, black) && !Atak.SqAtakd(F1, black) && !Atak.SqAtakd(G1, black)) {
            ADDMOVE(E1, G1, CASTLING);
        }
        if (side == white && ((board.flag & WQUEENCASTLE) != 0) && ((b & BitPosArray[A1]) != 0) &&
                (FromToRay[E1][B1] & blocker) == 0 &&
                !Atak.SqAtakd(E1, black) && !Atak.SqAtakd(D1, black) && !Atak.SqAtakd(C1, black)) {
            ADDMOVE(E1, C1, CASTLING);
        }
        if (side == black && ((board.flag & BKINGCASTLE) != 0) && ((b & BitPosArray[H8]) != 0) &&
                (FromToRay[E8][G8] & blocker) == 0 &&
                !Atak.SqAtakd(E8, white) && !Atak.SqAtakd(F8, white) && !Atak.SqAtakd(G8, white)) {
            ADDMOVE(E8, G8, CASTLING);
        }
        if (side == black && ((board.flag & BQUEENCASTLE) != 0) && ((b & BitPosArray[A8]) != 0) &&
                (FromToRay[E8][B8] & blocker) == 0 &&
                !Atak.SqAtakd(E8, white) && !Atak.SqAtakd(D8, white) && !Atak.SqAtakd(C8, white)) {
            ADDMOVE(E8, C8, CASTLING);
        }

        /* Update tree pointers and count */
        TreePtr[ply + 1] = node;
        GenCnt += TreePtr[ply + 1] - TreePtr[ply];
    }


    /**
     * Here I generate only non-captures.  Promotions are considered
     * as captures and are not generated.
     */
    static void GenNonCaptures(int ply) {
        int side;
        int piece, sq, t;
        /*BitBoard*/long b, c, d, blocker, notblocker;
        /*BitBoard*/long[] a;

        side = board.side;
        a = board.b[side];
        blocker = board.blocker;
        notblocker = ~blocker;
        node = TreePtr[ply + 1];

        /* Knight & King */
        for (piece = knight; piece <= king; piece += 4) {
            b = a[piece];
            while (b != 0) {
                sq = Util.leadz(b);
                b &= ~Util.longBit(sq);  //CLEARBIT (b, sq);
                BitToMove(sq, MoveArray[piece][sq] & notblocker);
            }
        }

        /* Bishops */
        b = a[bishop];
        while (b != 0) {
            sq = Util.leadz(b);
            b &= ~Util.longBit(sq);  //CLEARBIT (b, sq);
            d = BishopAttack(sq);
            BitToMove(sq, d & notblocker);
        }

        /* Rooks */
        b = a[rook];
        while (b != 0) {
            sq = Util.leadz(b);
            b &= ~Util.longBit(sq);  //CLEARBIT (b, sq);
            d = RookAttack(sq);
            BitToMove(sq, d & notblocker);
        }

        /* Queen */
        b = a[queen];
        while (b != 0) {
            sq = Util.leadz(b);
            b &= ~Util.longBit(sq);  //CLEARBIT (b, sq);
            d = QueenAttack(sq);
            BitToMove(sq, d & notblocker);
        }

        /*  White pawn moves  */
        if (side == white) {
            c = (a[pawn] >>> 8) & notblocker;        /* 1 square forward */
            while (c != 0) {
                t = Util.leadz(c);
                c &= ~Util.longBit(t);  //CLEARBIT (c, t);
                if (t < 56)
                    ADDMOVE(t - 8, t, 0);
            }

            b = a[pawn] & RankBit[1];            /* all pawns on 2nd rank */
            c = (b >>> 8) & notblocker;
            c = (c >>> 8) & notblocker;        /* 2 squares forward */
            while (c != 0) {
                t = Util.leadz(c);
                c &= ~Util.longBit(t);  //CLEARBIT (c, t);
                ADDMOVE(t - 16, t, 0);
            }
        }


        /*  Black pawn forward moves   */
        if (side == black) {
            c = (a[pawn] << 8) & notblocker;
            while (c != 0) {
                t = Util.leadz(c);
                c &= ~Util.longBit(t);  //CLEARBIT (c, t);
                if (t > 7)
                    ADDMOVE(t + 8, t, 0);
            }

            b = a[pawn] & RankBit[6];                 /* all pawns on 7th rank */
            c = (b << 8) & notblocker;
            c = (c << 8) & notblocker;
            while (c != 0) {
                t = Util.leadz(c);
                c &= ~Util.longBit(t);  //CLEARBIT (c, t);
                ADDMOVE(t + 16, t, 0);
            }
        }

        /* Castling code */
        b = board.b[side][rook];
        if (side == white && (board.flag & WKINGCASTLE) != 0 && (b & BitPosArray[7]) != 0 &&
                (FromToRay[4][6] & blocker) == 0 &&
                !Atak.SqAtakd(4, black) && !Atak.SqAtakd(5, black) && !Atak.SqAtakd(6, black)) {
            ADDMOVE(4, 6, CASTLING);
        }
        if (side == white && (board.flag & WQUEENCASTLE) != 0 && (b & BitPosArray[0]) != 0 &&
                (FromToRay[4][1] & blocker) == 0 &&
                !Atak.SqAtakd(4, black) && !Atak.SqAtakd(3, black) && !Atak.SqAtakd(2, black)) {
            ADDMOVE(4, 2, CASTLING);
        }
        if (side == black && (board.flag & BKINGCASTLE) != 0 && (b & BitPosArray[63]) != 0 &&
                (FromToRay[60][62] & blocker) == 0 &&
                !Atak.SqAtakd(60, white) && !Atak.SqAtakd(61, white) && !Atak.SqAtakd(62, white)) {
            ADDMOVE(60, 62, CASTLING);
        }
        if (side == black && (board.flag & BQUEENCASTLE) != 0 && (b & BitPosArray[56]) != 0 &&
                (FromToRay[60][57] & blocker) == 0 &&
                !Atak.SqAtakd(60, white) && !Atak.SqAtakd(59, white) && !Atak.SqAtakd(58, white)) {
            ADDMOVE(60, 58, CASTLING);
        }

        /* Update tree pointers and count */
        TreePtr[ply + 1] = node;
        GenCnt += TreePtr[ply + 1] - TreePtr[ply];
    }


    /**
     * This routine generates captures.  En passant and pawn promotions
     * are included.
     */
    static void GenCaptures(int ply) {
        int side;
        int piece, sq, t, ep;
        /*BitBoard*/long b, c, enemy, blocker;
        /*BitBoard*/long[] a;

        side = board.side;
        a = board.b[side];
        enemy = board.friends[1 ^ side];
        blocker = board.blocker;
        node = TreePtr[ply + 1];
        ep = board.ep;

        /* Knight  & King */
        for (piece = knight; piece <= king; piece += 4) {
            b = a[piece];
            while (b != 0) {
                sq = Util.leadz(b);
                b &= ~Util.longBit(sq);  //CLEARBIT (b, sq);
                BitToMove(sq, MoveArray[piece][sq] & enemy);
            }
        }

        /* Bishop */
        b = a[bishop];
        while (b != 0) {
            sq = Util.leadz(b);
            b &= ~Util.longBit(sq);  //CLEARBIT (b, sq);
            c = BishopAttack(sq);
            BitToMove(sq, c & enemy);
        }

        /* Rook */
        b = a[rook];
        while (b != 0) {
            sq = Util.leadz(b);
            b &= ~Util.longBit(sq);  //CLEARBIT (b, sq);
            c = RookAttack(sq);
            BitToMove(sq, c & enemy);
        }

        /* Queen */
        b = a[queen];
        while (b != 0) {
            sq = Util.leadz(b);
            b &= ~Util.longBit(sq);  //CLEARBIT (b, sq);
            c = QueenAttack(sq);
            BitToMove(sq, c & enemy);
        }

        /*  White pawn moves  */
        if (side == white) {
            b = a[pawn] & RankBit[6];            /* all pawns on 7 rank */
            c = (b >>> 8) & ~blocker;            /* 1 square forward */
            while (c != 0) {
                t = Util.leadz(c);
                c &= ~Util.longBit(t);  //CLEARBIT (c, t);
                ADDPROMOTE(t - 8, t);
            }

            b = a[pawn] & ~FileBit[0];        /* captures to the left */
            c = (b >>> 7) & (board.friends[1 ^ side] | (ep > -1 ? BitPosArray[ep] : 0));
            while (c != 0) {
                t = Util.leadz(c);
                c &= ~Util.longBit(t);  //CLEARBIT (c, t);
                if (t >= 56)                /* promotion */ {
                    ADDPROMOTE(t - 7, t);
                } else if (ep == t) {
                    ADDMOVE(t - 7, t, ENPASSANT);
                } else {
                    ADDMOVE(t - 7, t, 0);
                }
            }

            b = a[pawn] & ~FileBit[7];        /* captures to the right */
            c = (b >>> 9) & (board.friends[1 ^ side] | (ep > -1 ? BitPosArray[ep] : 0));
            while (c != 0) {
                t = Util.leadz(c);
                c &= ~Util.longBit(t);  //CLEARBIT (c, t);
                if (t >= 56)                /* promotion */ {
                    ADDPROMOTE(t - 9, t);
                } else if (ep == t) {
                    ADDMOVE(t - 9, t, ENPASSANT);
                } else {
                    ADDMOVE(t - 9, t, 0);
                }
            }
        }

        /*  Black pawn forward moves   */
        if (side == black) {
            b = a[pawn] & RankBit[1];            /* all pawns on 2nd rank */
            c = (b << 8) & ~blocker;
            while (c != 0) {
                t = Util.leadz(c);
                c &= ~Util.longBit(t);  //CLEARBIT (c, t);
                ADDPROMOTE(t + 8, t);
            }

            b = a[pawn] & ~FileBit[7];        /* captures to the left */
            c = (b << 7) & (board.friends[1 ^ side] | (ep > -1 ? BitPosArray[ep] : 0));
            while (c != 0) {
                t = Util.leadz(c);
                c &= ~Util.longBit(t);  //CLEARBIT (c, t);
                if (t <= 7)                /* promotion */ {
                    ADDPROMOTE(t + 7, t);
                } else if (ep == t) {
                    ADDMOVE(t + 7, t, ENPASSANT);
                } else {
                    ADDMOVE(t + 7, t, 0);
                }
            }

            b = a[pawn] & ~FileBit[0];        /* captures to the right */
            c = (b << 9) & (board.friends[1 ^ side] | (ep > -1 ? BitPosArray[ep] : 0));
            while (c != 0) {
                t = Util.leadz(c);
                c &= ~Util.longBit(t);  //CLEARBIT (c, t);
                if (t <= 7)                /* promotion */ {
                    ADDPROMOTE(t + 9, t);
                } else if (ep == t) {
                    ADDMOVE(t + 9, t, ENPASSANT);
                } else {
                    ADDMOVE(t + 9, t, 0);
                }
            }
        }


        /* Update tree pointers and count */
        TreePtr[ply + 1] = node;
        GenCnt += TreePtr[ply + 1] - TreePtr[ply];
    }


    /**
     * The king is in check, so generate only moves which get the king out
     * of check.
     * Caveat:  The special case of an enpassant capture must be taken into
     * account too as the captured pawn could be the checking piece.
     */
    static void GenCheckEscapes(int ply) {
        int side, xside;
        int kingsq, chksq, sq, sq1, epsq, dir;
        /*BitBoard*/long checkers, b, c, p, escapes;
        escapes = NULLBITBOARD;
        side = board.side;
        xside = 1 ^ side;
/*   TreePtr[ply + 1] = TreePtr[ply];  */
        node = TreePtr[ply + 1];
        kingsq = board.king[side];
        checkers = Atak.AttackTo(kingsq, xside);
        p = board.b[side][pawn];

        if (Util.nbits(checkers) == 1) {
            /*  Captures of checking pieces (except by king) */
            chksq = Util.leadz(checkers);
            b = Atak.AttackTo(chksq, side);
            b &= ~board.b[side][king];
            while (b != 0) {
                sq = Util.leadz(b);
                b &= ~Util.longBit(sq);  //CLEARBIT (b, sq);
                if (!Atak.PinnedOnKing(sq, side)) {
                    if (cboard[sq] == pawn && (chksq <= H1 || chksq >= A8)) {
                        ADDPROMOTE(sq, chksq);
                    } else
                        ADDMOVE(sq, chksq, 0);
                }
            }

            /*  Maybe enpassant can help  */
            if (board.ep > -1) {
                epsq = board.ep;
                if (epsq + (side == white ? -8 : 8) == chksq) {
                    b = MoveArray[ptype[1 ^ side]][epsq] & p;
                    while (b != 0) {
                        sq = Util.leadz(b);
                        b &= ~Util.longBit(sq);  //CLEARBIT (b, sq);
                        if (!Atak.PinnedOnKing(sq, side))
                            ADDMOVE(sq, epsq, ENPASSANT);
                    }
                }
            }

            /* Lets block/capture the checking piece */
            if (slider[cboard[chksq]] != 0) {
                c = FromToRay[kingsq][chksq] & NotBitPosArray[chksq];
                while (c != 0) {
                    sq = Util.leadz(c);
                    c &= ~Util.longBit(sq); //CLEARBIT (c, sq);
                    b = Atak.AttackTo(sq, side);
                    b &= ~(board.b[side][king] | p);

                    /* Add in pawn advances */
                    if (side == white && sq > H2) {
                        if ((BitPosArray[sq - 8] & p) != 0)
                            b |= BitPosArray[sq - 8];
                        if (RANK(sq) == 3 && cboard[sq - 8] == empty &&
                                (BitPosArray[sq - 16] & p) != 0)
                            b |= BitPosArray[sq - 16];
                    }
                    if (side == black && sq < H7) {
                        if ((BitPosArray[sq + 8] & p) != 0)
                            b |= BitPosArray[sq + 8];
                        if (RANK(sq) == 4 && cboard[sq + 8] == empty &&
                                (BitPosArray[sq + 16] & p) != 0)
                            b |= BitPosArray[sq + 16];
                    }
                    while (b != 0) {
                        sq1 = Util.leadz(b);
                        b &= ~Util.longBit(sq1); //CLEARBIT (b, sq1);
                        if (!Atak.PinnedOnKing(sq1, side)) {
                            if (cboard[sq1] == pawn && (sq > H7 || sq < A2)) {
                                ADDPROMOTE(sq1, sq);
                            } else
                                ADDMOVE(sq1, sq, 0);
                        }
                    }
                }
            }
        }

        /* If more than one checkers, move king to get out of check */
        if (checkers != 0)
            escapes = MoveArray[king][kingsq] & ~board.friends[side];
        while (checkers != 0) {
            chksq = Util.leadz(checkers);
            checkers &= ~Util.longBit(chksq); //CLEARBIT (checkers, chksq);
            dir = directions[chksq][kingsq];
            if (slider[cboard[chksq]] != 0)
                escapes &= ~Ray[chksq][dir];
        }
        while (escapes != 0) {
            sq = Util.leadz(escapes);
            escapes &= ~Util.longBit(sq); //CLEARBIT (escapes, sq);
            if (!Atak.SqAtakd(sq, xside))
                ADDMOVE(kingsq, sq, 0);
        }

        /* Update tree pointers and count */
        TreePtr[ply + 1] = node;
        GenCnt += TreePtr[ply + 1] - TreePtr[ply];
        return;
    }


    /**
     * All the illegal moves in the current ply is removed.
     */
    static void FilterIllegalMoves(int ply) {
        int p;   // leaf *
        int side, xside;
        boolean check;
        int sq;

        side = board.side;
        xside = 1 ^ side;
        sq = board.king[side];
        for (p = TreePtr[ply]; p < TreePtr[ply + 1]; p++) {
            Tree[p].move = Move.MakeMove(side, Tree[p].move);
            if (cboard[TOSQ(Tree[p].move)] != king)
                check = Atak.SqAtakd(sq, xside);
            else
                check = Atak.SqAtakd(TOSQ(Tree[p].move), xside);
            Tree[p].move = Move.UnmakeMove(xside, Tree[p].move);

            if (check)    /* its an illegal move */ {
                --TreePtr[ply + 1];
                Tree[p].copy(Tree[TreePtr[ply + 1]]);
                --p;
                GenCnt--;
            }
        }
    }

}
