/* This file is part of chessbox_gnu4j, a Java port of GNU Chess (5.0.7).

   Copyright (c) 2005 Xan Gregg (Java port)
   Copyright (c) 1999-2002 Free Software Foundation, Inc.

   chessbox_gnu4j is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 or (at your option)
   any later version.

   chessbox_gnu4j is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GNU Chess; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.

   Contact Info:
     bug-gnu-chess@gnu.org
     xan.gregg@pobox.com
*/
package org.chessbox.gnu4j;

/**
 * Created by Xan Gregg.
 * Date: May 5, 2005
 */
public class Hash extends Common {
    /**
     * Calculates the hashkey for the current board position.  We sometimes
     * need to do this especially when loading an EPD  position.
     * Note:  The hashkey is a 64 bit unsigned integer number.
     * Added in pawnhashkey calculation.
     */
    static void CalcHashKey() {
        int sq, piece, color;
        /*BitBoard*/long b;

        PawnHashKey = HashKey = 0;
        for (color = white; color <= black; color++) {
            for (piece = pawn; piece <= king; piece++) {
                b = board.b[color][piece];
                while (b != 0) {
                    sq = Util.leadz(b);
                    b &= ~Util.longBit(sq); //  CLEARBIT (b, sq);
                    HashKey ^= hashcode[color][piece][sq];
                    if (piece == pawn)
                        PawnHashKey ^= hashcode[color][piece][sq];
                }
            }
        }

        /* Take into account castling status & en passant */
        if (board.ep > -1)
            HashKey ^= ephash[board.ep];
        if ((board.flag & WKINGCASTLE) != 0)
            HashKey ^= WKCastlehash;
        if ((board.flag & WQUEENCASTLE) != 0)
            HashKey ^= WQCastlehash;
        if ((board.flag & BKINGCASTLE) != 0)
            HashKey ^= BKCastlehash;
        if ((board.flag & BQUEENCASTLE) != 0)
            HashKey ^= BQCastlehash;

        /* Take into account side to move */
        if (board.side == black)
            HashKey ^= Sidehash;

    }


    /**
     * Just a small utility routine to print out the hashkey (which is 64 bits).
     */
    static void ShowHashKey(/*HashType*/long HashKey) {
        System.out.println(Long.toHexString(HashKey));
    }

    /**
     * Calculates the ttable hashtable size, ttable hashmask, and pawntable hashmask
     */
    static void CalcHashSize(int tablesize) {
        int i;

        i = (tablesize < HASHSLOTS ? HASHSLOTS : tablesize);

        if (i > 1073741823) {
            System.out.println("Max hashsize exceeded\n");
            i = 107374183;
        }

        TTHashMask = 0;
        while ((i >>= 1) > 0) {
            TTHashMask <<= 1;
            TTHashMask |= 1;
        }
        HashSize = (int) (TTHashMask + 1);
        System.out.println("Adjusting HashSize to " + HashSize + " slots");

        i = PAWNSLOTS;
        PHashMask = 0;
        while ((i >>= 1) > 0) {
            PHashMask <<= 1;
            PHashMask |= 1;
        }
    }

}
