/* This file is part of chessbox_gnu4j, a Java port of GNU Chess (5.0.7).

   Copyright (c) 2005 Xan Gregg (Java port)
   Copyright (c) 1999-2002 Free Software Foundation, Inc.

   chessbox_gnu4j is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 or (at your option)
   any later version.

   chessbox_gnu4j is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GNU Chess; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.

   Contact Info:
     bug-gnu-chess@gnu.org
     xan.gregg@pobox.com
*/
package org.chessbox.gnu4j;

import java.util.Arrays;
@SuppressWarnings("all")
/**
 * Created by Xan Gregg.
 * Date: May 7, 2005
 */
public class Iterate extends Common {
    static final int WINDOW = 75;

    /**
     * Begin the iterative deepening code.  The variable Idepth represents
     * the iteration depth.
     */
    static void Iterate() {
        int side;
        int score, RootAlpha, RootBeta;

        Debug.dbg_printf("Entered iterate().\n");
        side = board.side;

        /* Note computer is playing side we are making move for. */
        computerplays = board.side;

        lazyscore[white] = lazyscore[black] = 150; /* was 50 */
        maxposnscore[white] = maxposnscore[black] = 150;
        GenCnt = 0;
        NodeCnt = QuiesCnt = 0;
        EvalCnt = EvalCall = 0;
        OneRepCnt = ChkExtCnt = RcpExtCnt = PawnExtCnt = HorzExtCnt = ThrtExtCnt = 0;
        KingExtCnt = 0;
        NullCutCnt = FutlCutCnt = 0;
        TotalGetHashCnt = GoodGetHashCnt = 0;
        TotalPutHashCnt = CollHashCnt = 0;
        TotalPawnHashCnt = GoodPawnHashCnt = 0;
        RootPawns = Util.nbits(board.b[white][pawn] | board.b[black][pawn]);
        RootPieces = Util.nbits(board.friends[white] | board.friends[black]) -
                RootPawns;
        RootMaterial = MATERIAL(side);
        RepeatCnt = 0;
        ElapsedTime = 0.0;
        StartTime = StartTiming();
        Arrays.fill(ChkCnt, 0);
        Arrays.fill(ThrtCnt, 0);
        Arrays.fill(history[0], 0);
        Arrays.fill(history[1], 0);
        Arrays.fill(killer1, 0);
        Arrays.fill(killer2, 0);
        ClearFlag(TIMEOUT);
        if ((flags & TIMECTL) != 0) {
            /* 2 seconds for unknown latencies */

            SearchTime = (TimeLimit[side] - 2) / MoveLimit[side];

            /* Allocate 10% of total time as reserve */
            /* mcriley - was 9 * S /10 */

            SearchTime = 95 * SearchTime / 100;
            /* Reserve more if in sudden death */
            /* mcrikey was 8 * S / 10 */

            if (suddendeath)
                SearchTime = 92 * SearchTime / 100;

            /* From loss against Ryo Saeba where we ran out of
               time first move out of book and blundered pawn */
            /* Book lookups took too long */
            if (nmovesfrombook <= 3) {
                System.out.println("Search time bonus near book\n");
                SearchTime = 1.5 * SearchTime;
            }

            /* Always target at least the increment seconds for the move! */
            /* To prevent huge surplus build-up. */

            if (TCinc != 0)
            if (SearchTime < TCinc) {
                System.out.println("TimeLimit[" + (side == white ? "White" : "Black") + "] = " + pad(round(TimeLimit[side], 2), 9));
                if (TimeLimit[side] > 30) {   /* Only if > 15 seconds left */
                    SearchTime = TCinc;
                }
            }
            Output.ShowTime();
        }
        Idepth = 0;
        TreePtr[2] = TreePtr[1];
        GenMove.GenMoves(1);
        GenMove.FilterIllegalMoves(1);
        Sort.SortRoot();

        InChk[1] = Atak.SqAtakd(board.king[side], 1 ^ side);

        /*  Are there any legal moves? */
        if (GenCnt == 0) {
            if ((flags & ENDED) == 0) { /* Don't give result after mate as it will be muddled */
                if (InChk[1]) {
                    if (computerplays == black)
                        System.out.println("1-0 {computer loses as black}");
                    else if (computerplays == white)
                        System.out.println("0-1 {computer loses as white}");
                } else
                    System.out.println("1/2-1/2 {stalemate}");
                System.out.flush();
            }
            SetFlag(TIMEOUT | ENDED);
            return;
        } else if (GenCnt == 1) {
            RootPV = Tree[TreePtr[1]].move;
            SetFlag(TIMEOUT);
        }
        lastrootscore = score = Eval.Evaluate(-INFINITY, INFINITY);

        wasbookmove = false;

        /* Don't look up moves in book in opponents time, think instead */

        if (bookmode != BOOKOFF && (flags & SOLVE) == 0 && (flags & PONDER) == 0 && nmovesfrombook <= 3) {
            Debug.dbg_printf("Doing book lookup.\n");
            if (Book.BookQuery(0) == BOOK_SUCCESS) {
                nmovesfrombook = 0;
                wasbookmove = true;
                SetFlag(TIMEOUT);
            } else
                nmovesfrombook++;
        } else
            nmovesfrombook++;


/* mcriley - was 2 * searchtime */

        maxtime = 4 * SearchTime;

        if ((flags & POST) != 0) {
            System.out.println("Root = " + score + ", ");
            System.out.println("Phase = " + phase + " ");
        }
        if (!ofpIsStdout) {
            ofp.println("Entered Iterate() for " +
                    ((flags & PONDER) != 0 ? "pondering" : "analyzing") + ".");
            ofp.print("Root = " + score + "\t");
            ofp.print("Phase = " + phase + "\t");
        }
        if (SearchDepth == 0) {
            if (!ofpIsStdout)
                ofp.println("\nTime = " + round(SearchTime, 2) + ", Max = " + round(maxtime, 2) +
                        ", Left = " + round(TimeLimit[side], 2) + ", Moves = " + MoveLimit[side]);
            if ((flags & POST) != 0)
//            printf ("\nTime = %.2f, Max = %.2f, Left = %.2f, Moves = %d\n",
                System.out.println("\nTime = " + round(SearchTime, 2) + ", Max = " + round(maxtime, 2) +
                        ", Left = " + round(TimeLimit[side], 2) + ", Moves = " + MoveLimit[side]);
        } else {
            if (!ofpIsStdout)
                ofp.println("Depth = " + SearchDepth);
            if ((flags & POST) != 0)
                System.out.println("Depth = " + SearchDepth);
        }

        if ((flags & POST) != 0) {
            System.out.println("Ply   Time     Eval      Nodes   Principal-Variation");
            if (!ofpIsStdout)
                ofp.println("Ply   Time     Eval      Nodes   Principal-Variation");
        }
        while ((flags & TIMEOUT) == 0) {
            if (score > MATE - 255) {
                RootAlpha = score - 1;
                RootBeta = MATE;
            } else if (score < -MATE + 255) {
                RootAlpha = -MATE;
                RootBeta = score + 1;
            } else {
                RootAlpha = MAX(score - WINDOW, -MATE);
                RootBeta = MIN(score + WINDOW, MATE);
            }
            Idepth += 1; /* increase iteration depth */
            rootscore = -INFINITY - 1;
            score = Search.SearchRoot(Idepth, RootAlpha, RootBeta);
            if (score >= RootBeta && score < MATE && (flags & TIMEOUT) == 0) {
                Search.ShowLine(RootPV, score, '+');
                rootscore = -INFINITY - 1;
                RootAlpha = RootBeta;
                RootBeta = INFINITY;
                score = Search.SearchRoot(Idepth, RootAlpha, RootBeta);
            }
            /*  If we fail low, then research. */
                    else {
                if (score <= RootAlpha && (flags & TIMEOUT) == 0) {
                    Search.ShowLine(RootPV, score, '-');
                    rootscore = -INFINITY - 1;
                    RootBeta = RootAlpha;
                    RootAlpha = -INFINITY;
                    score = Search.SearchRoot(Idepth, RootAlpha, RootBeta);
                }
            }

            /*  Print the final PV line  */
            Search.ShowLine(RootPV, score, '.');
            lastrootscore = score;


            /* See if we have time to start another iteration */
            /* mcriley - was 2 * S / 3 */
            if (SearchDepth == 0 && (flags & TIMECTL) != 0 && ElapsedTime >= 2 * SearchTime / 3)
                SetFlag(TIMEOUT);

            if (Math.abs(score) + Idepth >= MATE + 1)
                SetFlag(TIMEOUT);

            if ((flags & PONDER) == 0 && Idepth == SearchDepth)
                break;
        }

        /*
         * Elapsed time is calculated in Search for timed games
         * work it out here for statistical purposes
         */
        ElapsedTime = GetElapsed(StartTime);

        /* For the moment, just bail out if pondering after search */
        if ((flags & PONDER) != 0) return;
        /***************************************************************************
         *
         *  We've finished the search.  Do things like update the game history,
         *  time control stuff and print the search result.
         *  mcriley - MoveLimit was = 60
         ***************************************************************************/
        Move.SANMove(RootPV, 1);
        Game[GameCnt + 1].SANmv = SANmv;
        Game[GameCnt + 1].et = ElapsedTime;
        RootPV = Move.MakeMove(side, RootPV);
        if ((flags & TIMECTL) != 0) {
            if (suddendeath) {
                if (TimeLimit[side] > 0 && TimeLimit[side] <= 60)
                    MoveLimit[side] = 60;
                else
                    MoveLimit[side] = 35;
                System.out.println("MoveLimit is " + MoveLimit[side]);
                System.out.println("TimeLimit is " + TimeLimit[side]);
            } else
                MoveLimit[side]--;
            TimeLimit[side] -= ElapsedTime;
            if (TCinc != 0)
                TimeLimit[side] += TCinc;
            if (MoveLimit[side] == 0) {
                MoveLimit[side] = TCMove;
            }
        }

        if ((flags & XBOARD) != 0) {
            System.out.println((GameCnt / 2 + 1) + ". ... " + Move.AlgbrMove(RootPV));
            System.out.println("My move is: " + Move.AlgbrMove(RootPV));
            System.out.flush();
            if (!ofpIsStdout) {
                ofp.println((GameCnt / 2 + 1) + ". ... " + Move.AlgbrMove(RootPV));
                ofp.println("My move is: " + Move.AlgbrMove(RootPV));
                ofp.flush();
            }
        } else {
            if (!wasbookmove) {
                ofp.println("\nDepth = " + Idepth);
                ofp.println("Time = " + round(ElapsedTime, 1)
                        + " Rate=" + (ElapsedTime > 0 ? (long) ((NodeCnt + QuiesCnt) / ElapsedTime) : 0)
                        + " Nodes=[" + NodeCnt + "/" + QuiesCnt + "/" + (NodeCnt + QuiesCnt) +
                        "] GenCnt=" + GenCnt);
                ofp.println("Eval=[" + EvalCnt + "/" + EvalCall + "] RptCnt=" + RepeatCnt
                        + " NullCut=" + NullCutCnt + " FutlCut=" + FutlCutCnt);
                ofp.println("Ext: Chk=" + ChkExtCnt + " Recap=" + RcpExtCnt + " Pawn=" + PawnExtCnt +
                        " OneRep=" + OneRepCnt + " Horz=" + HorzExtCnt + " Mate=" + ThrtExtCnt + " KThrt=" + KingExtCnt);
                ofp.print("Material=[" +
                        board.pmaterial[white] + "/" +
                        board.pmaterial[black] + " : " +
                        board.material[white] + "/" +
                        board.material[black] + "] ");
                ofp.print("Lazy=[" + lazyscore[white] + "/" + lazyscore[black] + "] ");
                ofp.println("MaxPosnScore=[" + maxposnscore[white] + "/" + maxposnscore[black] + "]");
                ofp.println("Hash: Success=" +
                        (GoodGetHashCnt * 100 / (TotalGetHashCnt + 1)) + "% Collision=" +
                        (CollHashCnt * 100 / (TotalPutHashCnt + 1)) + "% Pawn=" +
                        (GoodPawnHashCnt * 100 / (TotalPawnHashCnt + 1)) + "%");
            }
            if ((flags & SOLVE) == 0) Output.ShowBoard();
            System.out.println("\nMy move is : " + SANmv);
            System.out.flush();
            if (!ofpIsStdout) {
                ofp.println("\nMy move is : " + SANmv);
                ofp.flush();
            }
        }

        /***************************************************************************
         *
         *  Before we leave, check to see if this is mate or stalemate.
         *
         ***************************************************************************/
        TreePtr[2] = TreePtr[1];
        GenMove.GenMoves(1);
        GenMove.FilterIllegalMoves(1);
        if (TreePtr[1] == TreePtr[2]) {
            if (Atak.SqAtakd(board.king[board.side], 1 ^ board.side)) {
                if (computerplays == black)
                    System.out.println("0-1 {computer wins as black}\n");
                else
                    System.out.println("1-0 {computer wins as white}\n");
            } else
                System.out.println("1/2-1/2 {stalemate}\n");
            System.out.flush();
            SetFlag(ENDED);
        }
        if (Eval.EvaluateDraw() || Repeat.Repeat() >= 2) {
            System.out.println("1/2-1/2 {draw}\n");
            System.out.flush();
            SetFlag(ENDED);
        }
    }


    static double GetElapsed(/*TimeVal*/long start) {
        long now = System.currentTimeMillis();
        return (now - start) / 1e3;
    }

    static /*TimeVal*/long StartTiming() {
        long now = System.currentTimeMillis();
        return now;
    }
}
