/* This file is part of chessbox_gnu4j, a Java port of GNU Chess (5.0.7).

   Copyright (c) 2005 Xan Gregg (Java port)
   Copyright (c) 1999-2002 Free Software Foundation, Inc.

   chessbox_gnu4j is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 or (at your option)
   any later version.

   chessbox_gnu4j is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GNU Chess; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.

   Contact Info:
     bug-gnu-chess@gnu.org
     xan.gregg@pobox.com
*/
package org.chessbox.gnu4j;

import java.util.Iterator;
import java.io.PrintWriter;
@SuppressWarnings("all")
/**
 * Created by Xan Gregg.
 * Date: Apr 22, 2005
 */
public class Main extends Common{

    public static void main(String[] args) {
      int i;

      /*
       * Parse command line arguments conforming with getopt_long syntax
       * Note: we have to support "xboard" and "post" as bare strings
       * for backward compatibility.
       */

      int opt_help = 0, opt_version = 0, opt_post = 0, opt_xboard = 0, opt_hash = 0, opt_easy = 0, opt_manual = 0;

        progname = Config.PROGRAM;

        CommandLine.Descriptor[] longOptions = new CommandLine.Descriptor[]{
            new CommandLine.Descriptor("hashsize", CommandLine.REQ_PARAM, 's'),
            new CommandLine.Descriptor("version", CommandLine.NO_PARAM, 'v'),
            new CommandLine.Descriptor("help", CommandLine.NO_PARAM, 'h'),
            new CommandLine.Descriptor("xboard", CommandLine.NO_PARAM, 'x'),
            new CommandLine.Descriptor("post", CommandLine.NO_PARAM, 'p'),
            new CommandLine.Descriptor("easy", CommandLine.NO_PARAM, 'e'),
            new CommandLine.Descriptor("manual", CommandLine.NO_PARAM, 'm')
        };

        for (Iterator it = CommandLine.parse(longOptions, args, 0); it.hasNext();) {
            CommandLine.OptionInstance option = (CommandLine.OptionInstance) it.next();
            switch (option.shortName) {

         case 'v':
           opt_version = 1;
           break;
         case 'h':
           opt_help = 1;
           break;
         case 'x':
           opt_xboard = 1;
           break;
         case 'p':
           opt_post = 1;
           break;
         case 'e':
           opt_easy = 1;
           break;
         case 'm':
           opt_manual = 1;
           break;
         case 's':
           if  ( option.param == null ){ /* we have error such as two -s */
             opt_help = 1;
             break;
           }
                    try {
                        opt_hash = Integer.parseInt(option.param);
                    } catch (NumberFormatException e) {
                        System.out.println("Hashsize out of Range or Invalid");
                        System.exit(1);
                    }
           break;
         case '?': /* On error give help - getopt does a basic message. */
           opt_help = 1;
           break;
         default:
           System.out.println ("Option Processing Failed\n");
           System.exit(1);    //abort();
         }
      }

      /* initialize control flags */
      flags = 0;

      /* output for thinking */
        ofp = new PrintWriter(System.out);
        ofpIsStdout = true;

      /* Handle old style command line options */
      if (args.length > 1) {
        for (i = 0; i < args.length; i++) {
          if (args[i].equals("xboard")) {
        SetFlag(XBOARD);
          } else if (args[i].equals("post")) {
        SetFlag(POST);
          }
        }
      }
      if (opt_xboard == 1)
        SetFlag(XBOARD);
      if (opt_post == 1)
        SetFlag(POST);
      if (opt_manual ==1)
        SetFlag(MANUAL);
      Cmd.cmd_version();

      /* If the version option was specified we can exit here */
      if (opt_version == 1)
        return;

      /* If a usage statement is required output it here */
      if (opt_help == 1){
        Cmd.cmd_usage();
        System.exit(1); /* Maybe an error if due to bad arguments. */
      }

      Debug.dbg_open(null);

      HashSize = 0 ; /* Set HashSize zero */
      if ( opt_hash != 0)
        Hash.CalcHashSize(opt_hash);

      Init.Initialize ();

      if ( opt_easy == 0)
       SetFlag(HARD);

      if (args.length > 1) {
        for (i = 0; i < args.length; i++) {
          if (args[i].equals("xboard")) {
        SetFlag(XBOARD);
          } else if (args[i].equals("post")) {
        SetFlag(POST);
          }
        }
      }

      bookmode = BOOKPREFER;
      bookfirstlast = 3;

      while ((flags & QUIT) == 0) {
        Debug.dbg_printf("Waiting for input...");
        Input.wait_for_input();
        Debug.dbg_printf("Parsing input...");
        Cmd.parse_input();
        Debug.dbg_printf("input_status = " + Input.input_status);
        if ((flags & THINK)!=0 && (flags & MANUAL)==0
                && (flags & ENDED)==0) {
          if ((flags & XBOARD)==0) System.out.println("Thinking...\n");
          Iterate.Iterate ();
          ClearFlag(THINK);
        }
        RealGameCnt = GameCnt;
        RealSide = board.side;
        Debug.dbg_printf("Waking up input...\n");
        Debug.dbg_printf("input_status = " + Input.input_status);
        Input.input_wakeup();
        Debug.dbg_printf("input_status = " + Input.input_status);
        /* Ponder only after first move */
        /* Ponder or (if pondering disabled) just wait for input */
        if ((flags & HARD) != 0 && (flags & QUIT) ==0) {
          Ponder.ponder();
        }
      }

      Input.CleanupInput();

      Debug.dbg_close();
    }

}
