/* This file is part of chessbox_gnu4j, a Java port of GNU Chess (5.0.7).

   Copyright (c) 2005 Xan Gregg (Java port)
   Copyright (c) 1999-2002 Free Software Foundation, Inc.

   chessbox_gnu4j is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 or (at your option)
   any later version.

   chessbox_gnu4j is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GNU Chess; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.

   Contact Info:
     bug-gnu-chess@gnu.org
     xan.gregg@pobox.com
*/
package org.chessbox.gnu4j;
@SuppressWarnings("all")
/**
 * Created by Xan Gregg.
 * Date: May 5, 2005
 */
public class Move extends Common {
    /**
     * To make a move on the board and update the various game information.
     */
    static int MakeMove(int side, int move) {
        if (CAPTUREPIECE(move) >= 7)
            throw new NullPointerException();
        int f, t, fpiece, tpiece;
        int rookf, rookt, epsq, sq;
        int xside;

        xside = 1 ^ side;
        f = FROMSQ(move);
        t = TOSQ(move);
        fpiece = cboard[f];
        tpiece = cboard[t];
        board.b[side][fpiece] &= ~Util.longBit(f);  //CLEARBIT (*a, f);
        board.b[side][fpiece] |= Util.longBit(t);  //SETBIT (*a, t);
        board.blockerr90 &= ~Util.longBit(r90[f]);  //CLEARBIT (board.blockerr90, r90[f]);
        board.blockerr90 |= Util.longBit(r90[t]);  //SETBIT (board.blockerr90, r90[t]);
        board.blockerr45 &= ~Util.longBit(r45[f]);  //CLEARBIT (board.blockerr45, r45[f]);
        board.blockerr45 |= Util.longBit(r45[t]);  //SETBIT (board.blockerr45, r45[t]);
        board.blockerr315 &= ~Util.longBit(r315[f]);  //CLEARBIT (board.blockerr315, r315[f]);
        board.blockerr315 |= Util.longBit(r315[t]);  //SETBIT (board.blockerr315, r315[t]);
        cboard[f] = empty;
        cboard[t] = fpiece;
        GameCnt++;
        Game[GameCnt].epsq = board.ep;
        Game[GameCnt].bflag = board.flag;
        Game[GameCnt].Game50 = (short) Game50;
        Game[GameCnt].hashkey = HashKey;
        Game[GameCnt].phashkey = PawnHashKey;
        Game[GameCnt].mvboard = (short) Mvboard[t];
        Game[GameCnt].comments = null;
        Mvboard[t] = Mvboard[f] + 1;
        Mvboard[f] = 0;
        if (board.ep > -1)
            HashKey ^= ephash[board.ep];
        HashKey ^= hashcode[side][fpiece][f];
        HashKey ^= hashcode[side][fpiece][t];
        if (fpiece == king)
            board.king[side] = (short) t;
        if (fpiece == pawn) {
            PawnHashKey ^= hashcode[side][pawn][f];
            PawnHashKey ^= hashcode[side][pawn][t];
        }

        if (tpiece != 0)        /* Capture */ {
            ExchCnt[side]++;
            board.b[xside][tpiece] &= ~Util.longBit(t);  //CLEARBIT (board.b[xside][tpiece], t);
            move |= (tpiece << 15);
            if (CAPTUREPIECE(move) >= 7)
                throw new NullPointerException();
            HashKey ^= hashcode[xside][tpiece][t];
            if (tpiece == pawn)
                PawnHashKey ^= hashcode[xside][pawn][t];
            board.material[xside] -= Value[tpiece];
            if (tpiece != pawn)
                board.pmaterial[xside] -= Value[tpiece];
        }

        if ((move & PROMOTION) != 0)    /* Promotion */ {
            board.b[side][PROMOTEPIECE(move)] |= Util.longBit(t);  //SETBIT (board.b[side][PROMOTEPIECE (*move)], t);
            board.b[side][fpiece] &= ~Util.longBit(t);  //CLEARBIT (*a, t);
            cboard[t] = PROMOTEPIECE(move);
            HashKey ^= hashcode[side][pawn][t];
            HashKey ^= hashcode[side][cboard[t]][t];
            PawnHashKey ^= hashcode[side][pawn][t];
            board.material[side] += (Value[cboard[t]] - ValueP);
            board.pmaterial[side] += Value[cboard[t]];
        }

        if ((move & ENPASSANT) != 0)    /* En passant */ {
            ExchCnt[side]++;
            epsq = board.ep + (side == white ? -8 : 8);
            board.b[xside][pawn] &= ~Util.longBit(epsq);  //CLEARBIT (board.b[xside][pawn], epsq);
            board.blockerr90 &= ~Util.longBit(r90[epsq]);  //CLEARBIT (board.blockerr90, r90[epsq]);
            board.blockerr45 &= ~Util.longBit(r45[epsq]);  //CLEARBIT (board.blockerr45, r45[epsq]);
            board.blockerr315 &= ~Util.longBit(r315[epsq]);  //CLEARBIT (board.blockerr315, r315[epsq]);
            cboard[epsq] = empty;
            HashKey ^= hashcode[xside][pawn][epsq];
            PawnHashKey ^= hashcode[xside][pawn][epsq];
            board.material[xside] -= ValueP;
        }
        if ((move & (CAPTURE | CASTLING)) != 0 || fpiece == pawn)
            Game50 = GameCnt;

        if ((move & CASTLING) != 0)    /* Castling */ {
            if ((t & 0x04) != 0)        /* King side */ {
                rookf = t + 1;
                rookt = t - 1;
            } else            /* Queen side */ {
                rookf = t - 2;
                rookt = t + 1;
            }
            board.b[side][rook] &= ~Util.longBit(rookf);  //CLEARBIT (*a, rookf);
            board.b[side][rook] |= Util.longBit(rookt);  //SETBIT (*a, rookt);
            board.blockerr90 &= ~Util.longBit(r90[rookf]);  //CLEARBIT (board.blockerr90, r90[rookf]);
            board.blockerr90 |= Util.longBit(r90[rookt]);  //SETBIT (board.blockerr90, r90[rookt]);
            board.blockerr45 &= ~Util.longBit(r45[rookf]);  //CLEARBIT (board.blockerr45, r45[rookf]);
            board.blockerr45 |= Util.longBit(r45[rookt]);  //SETBIT (board.blockerr45, r45[rookt]);
            board.blockerr315 &= ~Util.longBit(r315[rookf]);  //CLEARBIT (board.blockerr315, r315[rookf]);
            board.blockerr315 |= Util.longBit(r315[rookt]);  //SETBIT (board.blockerr315, r315[rookt]);
            cboard[rookf] = empty;
            cboard[rookt] = rook;
            Mvboard[rookf] = 0;
            Mvboard[rookt] = 1;
            HashKey ^= hashcode[side][rook][rookf];
            HashKey ^= hashcode[side][rook][rookt];
            board.castled[side] = true;
        }

        /* If king or rook move, clear castle flag. */
        if (side == white) {
            if (fpiece == king && (board.flag & WCASTLE) != 0) {
                if ((board.flag & WKINGCASTLE) != 0)
                    HashKey ^= WKCastlehash;
                if ((board.flag & WQUEENCASTLE) != 0)
                    HashKey ^= WQCastlehash;
                board.flag &= ~WCASTLE;
            } else if (fpiece == rook) {
                if (f == H1) {
                    if ((board.flag & WKINGCASTLE) != 0)
                        HashKey ^= WKCastlehash;
                    board.flag &= ~WKINGCASTLE;
                } else if (f == A1) {
                    if ((board.flag & WQUEENCASTLE) != 0)
                        HashKey ^= WQCastlehash;
                    board.flag &= ~WQUEENCASTLE;
                }
            }
            if (tpiece == rook) {
                if (t == H8) {
                    if ((board.flag & BKINGCASTLE) != 0)
                        HashKey ^= BKCastlehash;
                    board.flag &= ~BKINGCASTLE;
                } else if (t == A8) {
                    if ((board.flag & BQUEENCASTLE) != 0)
                        HashKey ^= BQCastlehash;
                    board.flag &= ~BQUEENCASTLE;
                }
            }
        } else {
            if (fpiece == king && (board.flag & BCASTLE) != 0) {
                if ((board.flag & BKINGCASTLE) != 0)
                    HashKey ^= BKCastlehash;
                if ((board.flag & BQUEENCASTLE) != 0)
                    HashKey ^= BQCastlehash;
                board.flag &= ~BCASTLE;
            } else if (fpiece == rook) {
                if (f == H8) {
                    if ((board.flag & BKINGCASTLE) != 0)
                        HashKey ^= BKCastlehash;
                    board.flag &= ~BKINGCASTLE;
                } else if (f == A8) {
                    if ((board.flag & BQUEENCASTLE) != 0)
                        HashKey ^= BQCastlehash;
                    board.flag &= ~BQUEENCASTLE;
                }
            }
            if (tpiece == rook) {
                if (t == H1) {
                    if ((board.flag & WKINGCASTLE) != 0)
                        HashKey ^= WKCastlehash;
                    board.flag &= ~WKINGCASTLE;
                } else if (t == A1) {
                    if ((board.flag & WQUEENCASTLE) != 0)
                        HashKey ^= WQCastlehash;
                    board.flag &= ~WQUEENCASTLE;
                }
            }
        }


        /* If pawn move 2 squares, set ep passant square. */
        if (fpiece == pawn && Math.abs(f - t) == 16) {
            sq = (f + t) / 2;
            board.ep = (short) sq;
            HashKey ^= ephash[sq];
        } else
            board.ep = -1;

        board.side = (short) xside;
        HashKey ^= Sidehash;
        Util.UpdateFriends();

        /* Update game record */
        Game[GameCnt].move = move;
        return move;
    }


    /**
     * To unmake a move on the board and update the various game information.
     * Note that if side is black, then black is about to move, but we will be
     * undoing a move by white, not black.
     */
    static int UnmakeMove(int side, int move) {
        if (CAPTUREPIECE(move) >= 7)
            throw new NullPointerException();
        int f, t, fpiece, cpiece;
        int rookf, rookt, epsq;
        int xside;

        side = 1 ^ side;
        xside = 1 ^ side;
        f = FROMSQ(move);
        t = TOSQ(move);
        fpiece = cboard[t];
        cpiece = CAPTUREPIECE(move);
        board.b[side][fpiece] &= ~Util.longBit(t);  //CLEARBIT (*a, t);
        board.b[side][fpiece] |= Util.longBit(f);  //SETBIT (*a, f);
        board.blockerr90 &= ~Util.longBit(r90[t]);  //CLEARBIT (board.blockerr90, r90[t]);
        board.blockerr90 |= Util.longBit(r90[f]);  //SETBIT (board.blockerr90, r90[f]);
        board.blockerr45 &= ~Util.longBit(r45[t]);  //CLEARBIT (board.blockerr45, r45[t]);
        board.blockerr45 |= Util.longBit(r45[f]);  //SETBIT (board.blockerr45, r45[f]);
        board.blockerr315 &= ~Util.longBit(r315[t]);  //CLEARBIT (board.blockerr315, r315[t]);
        board.blockerr315 |= Util.longBit(r315[f]);  //SETBIT (board.blockerr315, r315[f]);
        cboard[f] = cboard[t];
        cboard[t] = empty;
        Mvboard[f] = Mvboard[t] - 1;
        Mvboard[t] = Game[GameCnt].mvboard;
        if (fpiece == king)
            board.king[side] = (short) f;

        /* if capture, put back the captured piece */
        if ((move & CAPTURE) != 0) {
            ExchCnt[side]--;
            board.b[xside][cpiece] |= Util.longBit(t);  //SETBIT (board.b[xside][cpiece], t);
            board.blockerr90 |= Util.longBit(r90[t]);  //SETBIT (board.blockerr90, r90[t]);
            board.blockerr45 |= Util.longBit(r45[t]);  //SETBIT (board.blockerr45, r45[t]);
            board.blockerr315 |= Util.longBit(r315[t]);  //SETBIT (board.blockerr315, r315[t]);
            cboard[t] = cpiece;
            board.material[xside] += Value[cpiece];
            if (cpiece != pawn)
                board.pmaterial[xside] += Value[cpiece];
        }

        /* Undo promotion */
        if ((move & PROMOTION) != 0) {
            board.b[side][fpiece] &= ~Util.longBit(f);  //CLEARBIT (*a, f);
            board.b[side][pawn] |= Util.longBit(f);  //SETBIT (board.b[side][pawn], f);
            cboard[f] = pawn;
            board.material[side] += (ValueP - Value[PROMOTEPIECE(move)]);
            board.pmaterial[side] -= Value[PROMOTEPIECE(move)];
        }

        /* Undo enpassant */
        if ((move & ENPASSANT) != 0) {
            ExchCnt[side]--;
            epsq = (side == white ? Game[GameCnt].epsq - 8 : Game[GameCnt].epsq + 8);
            board.b[xside][pawn] |= Util.longBit(epsq);  //SETBIT (board.b[xside][pawn], epsq);
            board.blockerr90 |= Util.longBit(r90[epsq]);  //SETBIT (board.blockerr90, r90[epsq]);
            board.blockerr45 |= Util.longBit(r45[epsq]);  //SETBIT (board.blockerr45, r45[epsq]);
            board.blockerr315 |= Util.longBit(r315[epsq]);  //SETBIT (board.blockerr315, r315[epsq]);
            cboard[epsq] = pawn;
            board.material[xside] += ValueP;
        }

        /* if castling, undo rook move */
        if ((move & CASTLING) != 0) {
            if ((t & 0x04) != 0)        /* King side */ {
                rookf = t + 1;
                rookt = t - 1;
            } else            /* Queen side */ {
                rookf = t - 2;
                rookt = t + 1;
            }
            board.b[side][rook] &= ~Util.longBit(rookt);  //CLEARBIT (*a, rookt);
            board.b[side][rook] |= Util.longBit(rookf);  //SETBIT (*a, rookf);
            board.blockerr90 &= ~Util.longBit(r90[rookt]);  //CLEARBIT (board.blockerr90, r90[rookt]);
            board.blockerr90 |= Util.longBit(r90[rookf]);  //SETBIT (board.blockerr90, r90[rookf]);
            board.blockerr45 &= ~Util.longBit(r45[rookt]);  //CLEARBIT (board.blockerr45, r45[rookt]);
            board.blockerr45 |= Util.longBit(r45[rookf]);  //SETBIT (board.blockerr45, r45[rookf]);
            board.blockerr315 &= ~Util.longBit(r315[rookt]);  //CLEARBIT (board.blockerr315, r315[rookt]);
            board.blockerr315 |= Util.longBit(r315[rookf]);  //SETBIT (board.blockerr315, r315[rookf]);
            cboard[rookf] = rook;
            cboard[rookt] = empty;
            Mvboard[rookf] = 0;
            Mvboard[rookt] = 0;
            board.castled[side] = false;
        }

        Util.UpdateFriends();
        board.side = (short) side;
        board.ep = Game[GameCnt].epsq;
        board.flag = Game[GameCnt].bflag;
        HashKey = Game[GameCnt].hashkey;
        PawnHashKey = Game[GameCnt].phashkey;
        Game50 = Game[GameCnt].Game50;
        GameCnt--;
        return move;
    }


    /**
     * Convert the move to a SAN format.  GenMoves (ply) needs to be called
     * by the calling routine for this to work.
     */
    static void SANMove(int move, int ply) {
        int side;
        int piece;
        int f, t;
        /*BitBoard*/long b;
        int node1;

        SANmv = "";
        side = board.side;
        f = FROMSQ(move);
        t = TOSQ(move);
        /* Check some special moves like castling */
        if ((move & CASTLING) != 0) {
            if (t == 6 || t == 62)
                SANmv = "O-O";
            else
                SANmv = "O-O-O";
            return;
        }

        /****************************************************************************
         *
         *  Here split the code into 2 parts for clarity sake.  First part deals
         *  with pawn moves only, 2nd part only piece moves.  However before doing
         *  that, see if the move is ambiguous.
         *
         ****************************************************************************/

        /*  AMBIGUITY CHECK  */
        piece = cboard[f];
        side = board.side;
        b = board.b[side][piece];
        boolean ambiguous = false;
        node1 = TreePtr[ply];
        if (Util.nbits(b) > 1) {
            /*
             *  Scan the movelist to see if another same-type piece is
             *  also moving to that particular to-square.
             */
            for (node1 = TreePtr[ply]; node1 < TreePtr[ply + 1]; node1++) {
                if (FROMSQ(Tree[node1].move) == f)
                    continue;                   /* original piece, skip */
                if (TOSQ(Tree[node1].move) != t)
                    continue;                   /* diff to-square, skip */
                if (cboard[FROMSQ(Tree[node1].move)] != piece)
                    continue;                   /* diff piece   */
                ambiguous = true;
                break;
            }
        }

        if (piece == pawn) {
            /* Capture or enpassant */
            if (cboard[t] != 0 || board.ep == t) {
                SANmv += algbrfile[ROW(f)];
                SANmv += 'x';
            }
            SANmv += algbr[t];

            /* Promotion */
            if ((move & PROMOTION) != 0) {
                SANmv += '=';
                SANmv += notation[PROMOTEPIECE(move)];
            }
        } else    /* its not pawn */ {
            SANmv += notation[piece];
            if (ambiguous) {
                if (ROW(f) == ROW(FROMSQ(Tree[node1].move)))
                    SANmv += algbrrank[RANK(f)];
                else
                    SANmv += algbrfile[ROW(f)];
            }
            if (cboard[t] != 0)        /* capture */
                SANmv += 'x';
            SANmv += algbr[t];
        }

        /* See if it is a checking or mating move */
        move = MakeMove(side, move);
        if (Atak.SqAtakd(board.king[1 ^ side], side)) {
            TreePtr[ply + 2] = TreePtr[ply + 1];
            GenMove.GenCheckEscapes(ply + 1);
            if (TreePtr[ply + 1] == TreePtr[ply + 2])
                SANmv += '#';
            else
                SANmv += '+';
            GenCnt -= TreePtr[ply + 2] - TreePtr[ply + 1];
        }
        move = UnmakeMove(1 ^ side, move);

        return;
    }


    static int ASCIITOFILE(char a) {return a - 'a';}

    static int ASCIITORANK(char a) {return a - '1';}

    static int ASCIITOSQ(char a, char b) {return ASCIITOFILE(a) + ASCIITORANK(b) * 8;}

    static boolean ATOH(char a) {return a >= 'a' && a <= 'h';}

    static boolean ITO8(char a) {return a >= '1' && a <= '8';}

    static int piece_id(char c) {
/* Given c, what is the piece id.  This only takes one char, which
 * isn't enough to handle two-character names (common in Russian text
 * and old English notation that used Kt), but we're not supposed to
 * see such text here anyway.  This will
 * accept "P" for pawn, and many lowercase chars (but not "b" for Bishop). */
        switch (c) {
            case 'n':
            case 'N':
                return knight;
            case 'B':
                return bishop;
            case 'r':
            case 'R':
                return rook;
            case 'q':
            case 'Q':
                return queen;
            case 'k':
            case 'K':
                return king;
            case 'p':
            case 'P':
                return pawn;
        }
        return empty;
    }


    /**
     * This routine takes a string and check to see if it is a legal move.
     * Note.  At the moment, we accept 2 types of moves notation.
     * 1.  e2e4 format.   2. SAN format. (e4)
     */
    static int ValidateMove(String s) //  returns -1 or index into Tree[]
    {
        int f, t, side, rank, file, fileto;
        char piece;
        int pieceid, pieceid2;
        char promote;
        int kount;
        String mvstring = "";
        char[] mvstr = new char[20];    // extra chars for simpler probing
        /*BitBoard*/long b, b2;
        int n1, n2;

        TreePtr[2] = TreePtr[1];
        GenMove.GenMoves(1);
        GenMove.FilterIllegalMoves(1);
        side = board.side;

        /************************************************************************
         * The thing to do now is to clean up the move string.  This
         * gets rid of things like 'x', '+', '=' and so forth.
         ************************************************************************/
        StringBuffer sb = new StringBuffer(20);
        int len = s.length();
        for (int i = 0; i < len; i++) {
            char c = s.charAt(i);
            // fixc added check for '#' char
            if (c != 'x' && c != '+' && c != '=' && c != ' ' && c != '#')
                sb.append(c);
        }
        mvstring = sb.toString();
        mvstring.getChars(0, mvstring.length(), mvstr, 0);

        /* Flush castles that check */
        // fixc this shouldn't be necessary -- we just removed these chars
//       if (mvstr[strlen(mvstr)-1] == '+' || mvstr[strlen(mvstr)-1] == '#' ||
//           mvstr[strlen(mvstr)-1] == '=') mvstr[strlen(mvstr)-1] = '\000';

        /* Check for castling */
        if (mvstring.equals("O-O") || mvstring.equals("o-o") ||
                mvstring.equals("0-0")) {
            if (side == white) {
                f = 4;
                t = 6;
            } else {
                f = 60;
                t = 62;
            }
            return IsInMoveList(1, f, t, ' ');
        }

        if (mvstring.equals("O-O-O") || mvstring.equals("o-o-o") ||
                mvstring.equals("0-0-0")) {
            if (side == white) {
                f = 4;
                t = 2;
            } else {
                f = 60;
                t = 58;
            }
            return IsInMoveList(1, f, t, ' ');
        }

        /*  Test to see if it is e2e4 type notation */
        if (ATOH(mvstr[0]) && ITO8(mvstr[1]) && ATOH(mvstr[2]) &&
                ITO8(mvstr[3])) {
            f = (short) ASCIITOSQ(mvstr[0], mvstr[1]);
            t = (short) ASCIITOSQ(mvstr[2], mvstr[3]);
            piece = (mvstr.length == 5 ? mvstr[4] : ' ');
            return (IsInMoveList(1, f, t, piece));
        }


        /***********************************************************************
         *  Its a SAN notation move.  More headache!
         *  We generate all the legal moves and start comparing them with
         *  the input move.
         ***********************************************************************/
        if (ATOH(mvstr[0]))    /* pawn move */ {
            if (ITO8(mvstr[1]))                    /* e4 type */ {
                t = ASCIITOSQ(mvstr[0], mvstr[1]);
                f = t + (side == white ? -8 : 8);
                /* Add Sanity Check */
                if (f > 0 && f < 64) {
                    if ((BitPosArray[f] & board.b[side][pawn]) != 0) {
                        if (mvstr[2] != '\0')
                            return (IsInMoveList(1, f, t, mvstr[2]));
                        else
                            return (IsInMoveList(1, f, t, ' '));
                    }
                    f = t + (side == white ? -16 : 16);
                    if (f > 0 && f < 64) {
                        if ((BitPosArray[f] & board.b[side][pawn]) != 0)
                            return (IsInMoveList(1, f, t, ' '));
                    } /* End bound check +/- 16 */
                } /* End bound check +/- 8 */
            } else if (ATOH(mvstr[1]) && ITO8(mvstr[2]))        /* ed4 type */ {
                t = ASCIITOSQ(mvstr[1], mvstr[2]);
                rank = ASCIITORANK(mvstr[2]) + (side == white ? -1 : 1);
                f = rank * 8 + ASCIITOFILE(mvstr[0]);
                piece = (mvstr.length == 4 ? mvstr[3] : ' ');
                return (IsInMoveList(1, f, t, piece));
            } else if (ATOH(mvstr[1]))                    /* ed type */ {
                file = ASCIITOFILE(mvstr[0]);
                fileto = ASCIITOFILE(mvstr[1]);
                b = board.b[side][pawn] & FileBit[file];
                if (side == white)
                    b = b >>> (fileto < file ? 7 : 9);
                else
                    b = b << (fileto < file ? 9 : 7);
                if (board.ep > -1)
                    b = b & (board.friends[1 ^ side] | BitPosArray[board.ep]);
                else
                    b = b & (board.friends[1 ^ side]);
                switch (Util.nbits(b)) {
                    case 0:
                        return -1;
                    case 1:
                        t = Util.leadz(b);
                        f = t - (side == white ? 8 : -8) + (file - fileto);
                        piece = (mvstr.length == 3 ? mvstr[2] : ' ');
                        return (IsInMoveList(1, f, t, piece));
                    default :
                        System.out.println("Ambiguous move: " + s + " " + mvstr);
                        Output.ShowBoard();
/*
		      getchar();
*/
                        return -1;
                }
            }

        } else if ((pieceid = piece_id(mvstr[0])) != empty &&
                (piece_id(mvstr[1]) == empty))    /* Is a piece move */ {
            /* Since piece_id accepts P as pawns, this will correctly
             * handle malformed commands like Pe4 */

            b = board.b[side][pieceid];
            t = -1;
            if (ITO8(mvstr[1]))                /* N1d2 type move */ {
                rank = ASCIITORANK(mvstr[1]);
                b &= RankBit[rank];
                t = ASCIITOSQ(mvstr[2], mvstr[3]);
            } else if (ATOH(mvstr[1]) && ATOH(mvstr[2]))    /* Nbd2 type move */ {
                file = ASCIITOFILE(mvstr[1]);
                b &= FileBit[file];
                t = ASCIITOSQ(mvstr[2], mvstr[3]);
            } else if (ATOH(mvstr[1]) && ITO8(mvstr[2]))    /* Nd2 type move */ {
                t = ASCIITOSQ(mvstr[1], mvstr[2]);
            }

            kount = 0;
            n1 = n2 = -1;
            while (b != 0) {
                f = Util.leadz(b);
                b &= ~Util.longBit(f);  //CLEARBIT (b, f);
                if ((n1 = IsInMoveList(1, f, t, ' ')) != -1) {
                    n2 = n1;
                    kount++;
                }
            }
            if (kount > 1) {
                System.out.println("Ambiguous move: " + s + " " + mvstr);
                Output.ShowBoard();
/*
	 getchar();
*/
                return -1;
            } else if (kount == 0)
                return -1;
            else
                return (n2);
        } else if (((pieceid = piece_id(mvstr[0])) != empty) &&
                ((pieceid2 = piece_id(mvstr[1])) != empty) &&
                ((mvstr[2] == '\0') ||
                        ((piece_id(mvstr[2]) != empty) && mvstr[3] == '\0'))) { /* KxP format */
            promote = ' ';
            // fixc:            if (piece_id(mvstr[2] != empty)) {
            if (piece_id(mvstr[2]) != empty) {
                promote = mvstr[2];
            }
            kount = 0;
            n1 = n2 = -1;
            b = board.b[side][pieceid];
            while (b != 0) {
                f = Util.leadz(b);
                b &= ~Util.longBit(f);  //CLEARBIT (b, f);
                b2 = board.b[1 ^ side][pieceid2];
                while (b2 != 0) {
                    t = Util.leadz(b2);
                    b2 &= ~Util.longBit(t);  //CLEARBIT (b2, t);
                    System.out.print("Trying " + AlgbrMove(MOVE(f, t)) + ": ");
                    if ((n1 = IsInMoveList(1, f, t, promote)) != -1) {
                        n2 = n1;
                        kount++;
                        System.out.print("Y  ");
                    } else
                        System.out.print("N  ");
                }
            }
            if (kount > 1) {
                System.out.println("Ambiguous move: " + s + " " + mvstr);
                Output.ShowBoard();
/*
	 getchar();
*/
                return -1;
            } else if (kount == 0)
                return -1;
            else
                return (n2);

        }

        /* Fall through.  Nothing worked, return that no move was performed. */
        return -1;
    }


    static int IsInMoveList(int ply, int f, int t, char piece)
    /**************************************************************************
     *
     *  Checks to see if from and to square can be found in the movelist
     *  and is legal.
     *
     **************************************************************************/ {
        int node;    // leaf*

        for (node = TreePtr[ply]; node < TreePtr[ply + 1]; node++) {
            if ((int) (Tree[node].move & 0x0FFF) == MOVE(f, t) &&
                    Character.toUpperCase(piece) == notation[PROMOTEPIECE(Tree[node].move)])
                return (node);
        }
        return -1;
    }


    /**
     * Check that a move is legal on the current board.
     * Perform some preliminary sanity checks.
     * 1.  If from square is emtpy, illegal.
     * 2.  Piece not of right color.
     * 3.  To square is friendly, illegal.
     * 4.  Promotion move or enpassant, so piece must be pawn.
     * 5.  Castling move, piece must be king.
     * 6.  Pawn moving to 8th rank must be a promotion. (xg)
     * Note that IsLegalMove() no longer cares about if a move will place the
     * king in check.  This will be caught by the Search().
     */
    static boolean IsLegalMove(int move) {
        int side;
        int f, t, piece;
        /*BitBoard*/long blocker, enemy;

        f = FROMSQ(move);
        t = TOSQ(move);

        /*  Empty from square  */
        if (cboard[f] == empty)
            return (false);

        /*  Piece is not right color  */
        side = board.side;
        if ((BitPosArray[f] & board.friends[side]) == 0)
            return (false);

        /*  TO square is a friendly piece, so illegal move  */
        if ((BitPosArray[t] & board.friends[side]) != 0)
            return (false);

        piece = cboard[f];
        /*  If promotion move, piece must be pawn */
        if ((move & (PROMOTION | ENPASSANT)) != 0 && piece != pawn)
            return (false);

        /*  If enpassant, then the enpassant square must be correct */
        if ((move & ENPASSANT) != 0 && t != board.ep)
            return (false);

        /*  If castling, then make sure its the king */
        if ((move & CASTLING) != 0 && piece != king)
            return (false);

        blocker = board.blocker;
        /*  Pawn moves need to be handle specially  */
        if (piece == pawn) {
            if ((move & ENPASSANT) != 0 && board.ep > -1)
                enemy = board.friends[1 ^ side] | BitPosArray[board.ep];
            else
                enemy = board.friends[1 ^ side];
            if (side == white) {
                if ((MoveArray[pawn][f] & BitPosArray[t] & enemy) == 0 &&
                        !(t - f == 8 && cboard[t] == empty) &&
                        !(t - f == 16 && RANK(f) == 1 && (FromToRay[f][t] & blocker) == 0))
                    return (false);
                // fixc test for pawn moving to last rank (happens in BT2630 test)
                if ((move & PROMOTION) == 0 && t >= 56)
                    return false;
            } else if (side == black) {
                if ((MoveArray[bpawn][f] & BitPosArray[t] & enemy) == 0 &&
                        !(t - f == -8 && cboard[t] == empty) &&
                        !(t - f == -16 && RANK(f) == 6 && (FromToRay[f][t] & blocker) == 0))
                    return (false);
                // fixc test for pawn moving to last rank (happens in BT2630 test)
                if ((move & PROMOTION) == 0 && t <= 7)
                    return false;
            }
        }
        /*  King moves are also special, especially castling  */
                else if (piece == king) {
            if (side == white) {
                if ((MoveArray[piece][f] & BitPosArray[t]) == 0 &&
                        !(f == E1 && t == G1 && (board.flag & WKINGCASTLE) != 0 &&
                                (FromToRay[E1][G1] & blocker) == 0 && !Atak.SqAtakd(E1, black) &&
                                !Atak.SqAtakd(F1, black)) &&
                        !(f == E1 && t == C1 && (board.flag & WQUEENCASTLE) != 0 &&
                                (FromToRay[E1][B1] & blocker) == 0 && !Atak.SqAtakd(E1, black) &&
                                !Atak.SqAtakd(D1, black)))
                    return (false);
            }
            if (side == black) {
                if ((MoveArray[piece][f] & BitPosArray[t]) == 0 &&
                        !(f == E8 && t == G8 && (board.flag & BKINGCASTLE) != 0 &&
                                (FromToRay[E8][G8] & blocker) == 0 && !Atak.SqAtakd(E8, white) &&
                                !Atak.SqAtakd(F8, white)) &&
                        !(f == E8 && t == C8 && (board.flag & BQUEENCASTLE) != 0 &&
                                (FromToRay[E8][B8] & blocker) == 0 && !Atak.SqAtakd(E8, white) &&
                                !Atak.SqAtakd(D8, white)))
                    return (false);
            }
        } else {
            if ((MoveArray[piece][f] & BitPosArray[t]) == 0)
                return (false);
        }

        /*  If there is a blocker on the path from f to t, illegal move  */
        if (slider[piece] != 0) {
            if ((FromToRay[f][t] & NotBitPosArray[t] & blocker) != 0)
                return (false);
        }

        return (true);
    }

    /**
     * Convert an int move format to algebraic format of g1f3.
     */
    static String AlgbrMove(int move) {
        int f, t;
        String s = "";

        f = FROMSQ(move);
        t = TOSQ(move);
        s += algbr[f];
        s += algbr[t];
        if ((move & PROMOTION) != 0) {
            if ((flags & XBOARD) != 0)
                s += lnotation[PROMOTEPIECE(move)];
            else
                s += notation[PROMOTEPIECE(move)];
        }
        return s;
    }

}
