/* This file is part of chessbox_gnu4j, a Java port of GNU Chess (5.0.7).

   Copyright (c) 2005 Xan Gregg (Java port)
   Copyright (c) 1999-2002 Free Software Foundation, Inc.

   chessbox_gnu4j is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 or (at your option)
   any later version.

   chessbox_gnu4j is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GNU Chess; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.

   Contact Info:
     bug-gnu-chess@gnu.org
     xan.gregg@pobox.com
*/
package org.chessbox.gnu4j;

/**
 * Created by Xan Gregg.
 * Date: May 7, 2005
 */
public class Output extends Common {
    /**
     * Print out the time settings.
     */
    static void ShowTime() {
    }

    /**
     * Print out the move list.
     */
    static void ShowMoveList(int ply) {
        int node;    // leaf *
        int i = 0;

        for (node = TreePtr[ply]; node < TreePtr[ply + 1]; node++) {
            Move.SANMove(Tree[node].move, ply);
            System.out.print(pad(SANmv, 5) + " " + pad(Swap.SwapOff(Tree[node].move), 3) + "\t");
            if (++i == 5) {
                System.out.println();
                i = 0;
            }
        }
        System.out.println();
    }

    /**
     * Display the board.  Not only that but display some useful information
     * like whether enpassant is legal and castling state.
     */

    static void ShowSmallBoard() {
        int r, c, sq;

        System.out.println();
        if (board.side == white)
            System.out.print("white  ");
        else
            System.out.print("black  ");

        if ((board.flag & WKINGCASTLE) != 0)
            System.out.print("K");
        if ((board.flag & WQUEENCASTLE) != 0)
            System.out.print("Q");
        if ((board.flag & BKINGCASTLE) != 0)
            System.out.print("k");
        if ((board.flag & BQUEENCASTLE) != 0)
            System.out.print("q");

        if (board.ep > -1)
            System.out.print("  " + algbr[board.ep]);

        System.out.println();
        for (r = 56; r >= 0; r -= 8) {
            for (c = 0; c < 8; c++) {
                sq = r + c;
                if ((board.b[white][pawn] & BitPosArray[sq]) != 0)
                    System.out.print("P ");
                else if ((board.b[white][knight] & BitPosArray[sq]) != 0)
                    System.out.print("N ");
                else if ((board.b[white][bishop] & BitPosArray[sq]) != 0)
                    System.out.print("B ");
                else if ((board.b[white][rook] & BitPosArray[sq]) != 0)
                    System.out.print("R ");
                else if ((board.b[white][queen] & BitPosArray[sq]) != 0)
                    System.out.print("Q ");
                else if ((board.b[white][king] & BitPosArray[sq]) != 0)
                    System.out.print("K ");
                else if ((board.b[black][pawn] & BitPosArray[sq]) != 0)
                    System.out.print("p ");
                else if ((board.b[black][knight] & BitPosArray[sq]) != 0)
                    System.out.print("n ");
                else if ((board.b[black][bishop] & BitPosArray[sq]) != 0)
                    System.out.print("b ");
                else if ((board.b[black][rook] & BitPosArray[sq]) != 0)
                    System.out.print("r ");
                else if ((board.b[black][queen] & BitPosArray[sq]) != 0)
                    System.out.print("q ");
                else if ((board.b[black][king] & BitPosArray[sq]) != 0)
                    System.out.print("k ");
                else
                    System.out.print(". ");
            }
            System.out.println();
        }
        System.out.println();
    }


    /**
     * Just to print a lousy ascii board
     */
    static void ShowBitBoard(/*BitBoard*/long b) {
        int r, c;

        System.out.println();
        for (r = 56; r >= 0; r -= 8) {
            for (c = 0; c < 8; c++) {
                if ((b & BitPosArray[r + c]) != 0)
                    System.out.print("1 ");
                else
                    System.out.print(". ");
            }
            System.out.println();
        }
        System.out.println();
    }


    /**
     * Display the board.  Not only that but display some useful information
     * like whether enpassant is legal and castling state.
     */
    static void ShowBoard() {
        int r, c, sq;

        ofp.println();
        if (board.side == white)
            ofp.print("white  ");
        else
            ofp.print("black  ");

        if ((board.flag & WKINGCASTLE) != 0)
            ofp.print("K");
        if ((board.flag & WQUEENCASTLE) != 0)
            ofp.print("Q");
        if ((board.flag & BKINGCASTLE) != 0)
            ofp.print("k");
        if ((board.flag & BQUEENCASTLE) != 0)
            ofp.print("q");

        if (board.ep > -1)
            ofp.print("  " + algbr[board.ep]);

        ofp.println();
        for (r = 56; r >= 0; r -= 8) {
            for (c = 0; c < 8; c++) {
                sq = r + c;
                if ((board.b[white][pawn] & BitPosArray[sq]) != 0)
                    ofp.print("P ");
                else if ((board.b[white][knight] & BitPosArray[sq]) != 0)
                    ofp.print("N ");
                else if ((board.b[white][bishop] & BitPosArray[sq]) != 0)
                    ofp.print("B ");
                else if ((board.b[white][rook] & BitPosArray[sq]) != 0)
                    ofp.print("R ");
                else if ((board.b[white][queen] & BitPosArray[sq]) != 0)
                    ofp.print("Q ");
                else if ((board.b[white][king] & BitPosArray[sq]) != 0)
                    ofp.print("K ");
                else if ((board.b[black][pawn] & BitPosArray[sq]) != 0)
                    ofp.print("p ");
                else if ((board.b[black][knight] & BitPosArray[sq]) != 0)
                    ofp.print("n ");
                else if ((board.b[black][bishop] & BitPosArray[sq]) != 0)
                    ofp.print("b ");
                else if ((board.b[black][rook] & BitPosArray[sq]) != 0)
                    ofp.print("r ");
                else if ((board.b[black][queen] & BitPosArray[sq]) != 0)
                    ofp.print("q ");
                else if ((board.b[black][king] & BitPosArray[sq]) != 0)
                    ofp.print("k ");
                else
                    ofp.print(". ");
            }
            ofp.println();
        }
        ofp.println();
        ofp.flush();
    }

    static void ShowCBoard() {
        int r, c;

        for (r = 56; r >= 0; r -= 8) {
            for (c = 0; c < 8; c++) {
                System.out.print(" " + (cboard[r + c] != 0 ? notation[cboard[r + c]] : '.') + " ");
            }
            System.out.println();
        }
        System.out.println();
    }


    /**
     * Print the Mvboard[] array.
     */
    static void ShowMvboard() {
        int r, c;

        for (r = 56; r >= 0; r -= 8) {
            for (c = 0; c < 8; c++) {

                int mv = Mvboard[r + c];
                System.out.print(pad(mv, 2));
            }
            System.out.println();
        }
        System.out.println();
    }

    static void ShowGame() {
        int i;

/* *********************************************
   * We must handle the special case of an EPD *
   * game where the first move is by black     *
   ********************************************* */

        if (GameCnt >= 0) {

            System.out.print("      White   Black\n");

            if ((board.side == white && GameCnt % 2 == 1) ||
                    (board.side == black && GameCnt % 2 == 0)) {

                for (i = 0; i <= GameCnt; i += 2) {
                    System.out.println(pad(i / 2 + 1, 3) + ".  " + padRight(Game[i].SANmv, 7) +
                            " " + padRight(Game[i + 1].SANmv, 7));
                }
            } else {

                System.out.println("  1.          " + padRight(Game[0].SANmv, 7));

                for (i = 1; i <= GameCnt; i += 2) {
                    System.out.print(pad(i / 2 + 1, 3) + ".  " + padRight(Game[i].SANmv, 7) +
                            " " + padRight(Game[i + 1].SANmv, 7));
                }
            }
            System.out.println();
        }
    }
}
