/* This file is part of chessbox_gnu4j, a Java port of GNU Chess (5.0.7).

   Copyright (c) 2005 Xan Gregg (Java port)
   Copyright (c) 1999-2002 Free Software Foundation, Inc.

   chessbox_gnu4j is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 or (at your option)
   any later version.

   chessbox_gnu4j is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GNU Chess; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.

   Contact Info:
     bug-gnu-chess@gnu.org
     xan.gregg@pobox.com
*/
package org.chessbox.gnu4j;

import java.io.*;
import java.util.Set;
import java.util.GregorianCalendar;
import java.util.Calendar;
@SuppressWarnings("all")
/**
 * Created by Xan Gregg.
 * Date: May 1, 2005
 */
public class PGN extends Common {
    static String NULL2EMPTY(String x) { return (x != null) ? (x) : "";}

    /**
     * To save a game into PGN format to a file.  If the file does not exist,
     * it will create it.  If the file exists, the game is appended to the file.
     */
    static void PGNSaveToFile(String file, String resultstr) {
        PrintWriter fp;
        String s;    // char[100]
        int len;
        int i;

        try {
            fp = new PrintWriter(new FileWriter (file, true));    /*  Can we append to it?  */
        } catch (IOException e) {
            e.printStackTrace();
            System.out.println("Cannot write to file " + file);
            return;
        }

        /* Write the seven tags */
        fp.println("[Event \"" + NULL2EMPTY(LexPGN.pgn_event) + "\"]");
        fp.println("[Site \"" + NULL2EMPTY(LexPGN.pgn_site) + "\"]");
        if (LexPGN.pgn_date != null)
            fp.println("[Date \"" + LexPGN.pgn_date + "\"]");
        else {
            GregorianCalendar now = new GregorianCalendar();
            fp.println("[Date \"" + pad(now.get(Calendar.YEAR), 4)
                    + "." + pad0(now.get(Calendar.MONTH), 2)
                    + "." + pad0(now.get(Calendar.DAY_OF_MONTH), 2) + "\"]");
        }
        fp.println("[Round \"" + NULL2EMPTY(LexPGN.pgn_round) + "\"]");

        if (LexPGN.pgn_white != null)
            fp.println("[White \"" + LexPGN.pgn_white + "\"]");
        else if (computer == white)
            fp.println("[White \"" + Config.PROGRAM + " " + Config.VERSION + "\"]");
        else
            fp.println("[White \"" + name + "\"]");

        if (LexPGN.pgn_black != null)
            fp.println("[Black \"" + LexPGN.pgn_black + "\"]");
        else if (computer == black)
            fp.println("[Black \"" + Config.PROGRAM + " " + Config.VERSION + "\"]");
        else
            fp.println("[Black \"" + name + "\"]");

        if (LexPGN.pgn_whiteELO != null)
            fp.println("[WhiteELO \"" + NULL2EMPTY(LexPGN.pgn_white) + "\"]");
        else
            fp.println("[WhiteELO \"" + (computer == white ? myrating : opprating) + "\"]");
        if (LexPGN.pgn_blackELO != null)
            fp.println("[BlackELO \"" + NULL2EMPTY(LexPGN.pgn_black) + "\"]");
        else
            fp.println("[BlackELO \"" + (computer == white ? opprating : myrating) + "\"]");

        if (LexPGN.pgn_result != null)
            fp.println("[Result \"" + LexPGN.pgn_result + "\"]");
        else {
            /* Revive the little-known standard functions! */
            len = Math.min(resultstr.indexOf(' '), resultstr.indexOf('{'));
            if (len >= 0)
                fp.println("[Result \"" + resultstr.substring(0, len) + "\"]");
        }

        if (LexPGN.pgn_othertags != null) {
            fp.println("" + LexPGN.pgn_othertags + "");
        }

        fp.println("\n");

        if (LexPGN.initial_comments != null) {
            fp.println("\n" + LexPGN.initial_comments + "");
            /* If it doesn't end in \n, add it */
            if (LexPGN.initial_comments.length() > 0 &&
                    LexPGN.initial_comments.charAt(LexPGN.initial_comments.length() - 1) != '\n')
                fp.println();
        }

        s = "";
        for (i = 0; i <= GameCnt; i++) {
            if ((i % 2) != 0) {
                s = s + (i / 2 + 1) + ". ";
            }
            s = s + Game[i].SANmv + " ";
            if (s.length() > 80) {
                int p = 79;
                while (s.charAt(p--) != ' ') ;
                p++;
                fp.println(s.substring(0, p));
                s = s.substring(p);
            }
            if (Game[i].comments != null) {
                fp.println(s);
                fp.println(Game[i].comments);
                if (Game[i].comments.length() > 0 &&
                        Game[i].comments.charAt(Game[i].comments.length() - 1) != '\n')
                    fp.println();
                s = "";
            }
        }
        fp.println(s);
        fp.println(resultstr);
        fp.println();
        fp.println();
        fp.close();

    }


    /**
     * To read a game from a PGN file.
     */
    static void PGNReadFromFile(String file) {
        Reader fp;

        try {
            fp = new FileReader (file);
        } catch (FileNotFoundException e) {
            e.printStackTrace();
            System.err.println("Cannot open file " + file);
            return;
        }

        LexPGN.yyin = fp;

        Init.InitVars();

        LexPGN.data_dest = LexPGN.DEST_GAME;
        LexPGN.yylex();

        try {
            fp.close();
        } catch (IOException e) {
            e.printStackTrace();
        }
        Output.ShowBoard();
        TTable.TTClear();
    }

/* Only players in the table below are permitted into the opening book
   from the PGN files. Please expand the table as needed. Generally,
   I would recommend only acknowledged GM's and IM's and oneself, but
   because of the self-changing nature of the book, anything inferior
   will eventually be eliminated through automatic play as long as
   you feed the games the program plays back to itself with "book add pgnfile"
*/
/* TODO: Fix this so the list isn't hardcoded. */

    static String[] trusted_players = {
        "Alekhine",
        "Adams",
        "Anand",
        "Anderssen",
        "Andersson",
        "Aronin",
        "Averbakh",
        "Balashov",
        "Beliavsky",
        "Benko",
        "Bernstein",
        "Bird",
        "Bogoljubow",
        "Bolbochan",
        "Boleslavsky",
        "Byrne",
        "Botvinnik",
        "Bronstein",
        "Browne",
        "Capablanca",
        "Chigorin",
        "Christiansen",
        "De Firmian",
        "Deep Blue",
        "Deep Thought",
        "Donner",
        "Dreev",
        "Duras",
        "Euwe",
        "Evans",
        "Filip",
        "Fine",
        "Fischer",
        "Flohr",
        "Furman",
        "Gelfand",
        "Geller",
        "Gereben",
        "Glek",
        "Gligoric",
        "GNU",
        "Golombek",
        "Gruenfeld",
        "Guimard",
        "Hodgson",
        "Ivanchuk",
        "Ivkov",
        "Janowsky",
        "Kamsky",
        "Kan",
        "Karpov",
        "Kasparov",
        "Keres",
        "Korchnoi",
        "Kortschnoj",
        "Kotov",
        "Kramnik",
        "Kupreich",
        "Lasker",
        "Lautier",
        "Letelier",
        "Lilienthal",
        "Ljubojevic",
        "Marshall",
        "Maroczy",
        "Mieses",
        "Miles",
        "Morphy",
        "Mueller",     /* Every other German has this name... */
        "Nimzowitsch",
        "Nunn",
        "Opocensky",
        "Pachman",
        "Petrosian",
        "Piket",
        "Pilnik",
        "Pirc",
        "Polgar",
        "Portisch",
        "Psakhis",
        "Ragozin",
        "Reshevsky",
        "Reti",
        "Romanish",
        "Rubinstein",
        "Saemisch",
        "Seirawan",
        "Shirov",
        "Short",
        "Silman",
        "Smyslov",
        "Sokolsky",
        "Spassky",
        "Sveshnikov",
        "Stahlberg",
        "Steinitz",
        "Tal",
        "Tarjan",
        "Tartakower",
        "Timman",
        "Topalov",
        "Torre",
        "Vidmar"

    };

    static Set trusted_players_set;

    static boolean IsTrustedPlayer(String name)
/* Return 1 if name in trusted_players list, else 0 */ {
        if (trusted_players_set.isEmpty()) {
            // initialize set
            for (int i = 0; i < trusted_players.length; i++) {
                String trusted_player = trusted_players[i];
                trusted_players_set.add(trusted_player);
            }
        }
        return trusted_players_set.contains(name);
    }

    /**
     * To read a game from a PGN file and store out the hash entries to book.
     */
    static void BookPGNReadFromFile(String file) {
        Reader fp;

        try {
            fp = new FileReader (file);
        } catch (FileNotFoundException e) {
            e.printStackTrace();
            System.err.println("Cannot open file " + file);
            return;
        }

        int ngames = 0;
        long t1, t2;
        double et;
        int error;

        et = 0.0;
        t1 = System.currentTimeMillis();

        LexPGN.yyin = fp;

        /* Maybe add some more clever error handling later */
        if (Book.BookBuilderOpen() != BOOK_SUCCESS)
            return;
        newpos = existpos = 0;
        LexPGN.data_dest = LexPGN.DEST_BOOK;

        while (true) {
            Init.InitVars();
            Init.NewPosition();
            ClearFlag(MANUAL);
            ClearFlag(THINK);
            myrating = opprating = 0;

            error = LexPGN.yylex();
            if (error != 0) break;

            ngames++;
            if (ngames % 10 == 0) System.out.println("Games processed: " + ngames);
            System.out.flush();
        }

        try {
            fp.close();
        } catch (IOException e) {
            e.printStackTrace();
        }
        if (Book.BookBuilderClose() != BOOK_SUCCESS) {
            System.err.println("Error writing opening book during BookBuilderClose");  // fixjava perror()
        }

        /* Reset the board otherwise we leave the last position in the book
                            on the board. */
       Init.
        InitVars ();
        Init.NewPosition ();
        ClearFlag(MANUAL);
        ClearFlag(THINK);
       myrating = opprating = 0;

        t2 = System.
        currentTimeMillis();
       et += t2 - t1;

        System.out.println();
        /* Handle divide-by-zero problem */
                if

        (et < 0.5) { et = 1.0; }; System.out. println("Time = "+
        round(et,0)+" seconds"); System.out.println(
        "Games compiled: "+ngames);
       System.out. println(
        "Games per second: "+ngames/et);
         System. out.println(
        "Positions scanned: "+newpos+existpos);
         System.out.println("Positions per second: "+(newpos+existpos)/et);
         System .out.println("New & unique added: "+newpos+" positions");
       System.out.println("Duplicates not added: "+existpos+" positions");
    }

}
