/* This file is part of chessbox_gnu4j, a Java port of GNU Chess (5.0.7).

   Copyright (c) 2005 Xan Gregg (Java port)
   Copyright (c) 1999-2002 Free Software Foundation, Inc.

   chessbox_gnu4j is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 or (at your option)
   any later version.

   chessbox_gnu4j is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GNU Chess; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.

   Contact Info:
     bug-gnu-chess@gnu.org
     xan.gregg@pobox.com
*/
package org.chessbox.gnu4j;

import java.util.Arrays;
import java.util.Comparator;
import java.util.StringTokenizer;
import java.io.*;

@SuppressWarnings("all") 

/**
 * Created by Xan Gregg.
 * Date: May 9, 2005
 */
public class Players extends Common {

    static String PLAYERFILE = "players.dat";

    static int totalplayers = 0;

    static final int MAXPLAYERS = 500;

    static class PlayerEntry {
        String player;
        int wins;
        int losses;
        int draws;
    }

    static PlayerEntry[] playerdb = new PlayerEntry[MAXPLAYERS];
    static String lname;

    static int rscorecompare(Object aa, Object bb) {
        PlayerEntry a = (PlayerEntry) aa;
        PlayerEntry b = (PlayerEntry) bb;
        float ascore, bscore;
        ascore = (a.wins + (a.draws / 2)) / (a.wins + a.draws + a.losses);
        bscore = (b.wins + (b.draws / 2)) / (b.wins + b.draws + b.losses);
        if (ascore > bscore)
            return (-1);
        else if (bscore > ascore)
            return (1);
        else
            return (0);
    }

    static int scorecompare(Object aa, Object bb) {
        PlayerEntry a = (PlayerEntry) aa;
        PlayerEntry b = (PlayerEntry) bb;
        int ascore, bscore;
        ascore = 100 * (a.wins + (a.draws / 2)) / (a.wins + a.draws + a.losses);
        bscore = 100 * (b.wins + (b.draws / 2)) / (b.wins + b.draws + b.losses);
        if (bscore > ascore)
            return (1);
        else if (bscore < ascore)
            return (-1);
        else
            return (0);
    }

    static int namecompare(Object aa, Object bb) {
        PlayerEntry a = (PlayerEntry) aa;
        PlayerEntry b = (PlayerEntry) bb;
        return a.player.compareTo(b.player);
    }

    static void DBSortPlayer(String style) {
        if (style.startsWith("score")) {
            Arrays.sort(playerdb, 0, totalplayers, new Comparator() {
                public int compare(Object a, Object b) {
                    return scorecompare(a, b);
                }
            });
        } else if (style.startsWith("name")) {
            Arrays.sort(playerdb, 0, totalplayers, new Comparator() {
                public int compare(Object a, Object b) {
                    return namecompare(a, b);
                }
            });
        } else if (style.startsWith("reverse")) {
            Arrays.sort(playerdb, 0, totalplayers, new Comparator() {
                public int compare(Object a, Object b) {
                    return rscorecompare(a, b);
                }
            });
        }
    }

    static void DBListPlayer(String style) {
        int i;

        DBReadPlayer();
        DBSortPlayer(style);
        for (i = 0; i < totalplayers; i++) {
            //fixjava printf("%s %2.0f%% %d %d %d\n",
            System.out.println(playerdb[i].player +
                    pad(round(100.0 * (playerdb[i].wins + ((float) playerdb[i].draws / 2)) /
                    (playerdb[i].wins + playerdb[i].draws + playerdb[i].losses), 0), 4) + "% " +
                    playerdb[i].wins + " " +
                    playerdb[i].losses + " " +
                    playerdb[i].draws);
            if ((i + 1) % 10 == 0) {
                System.out.println("[Type a character to continue.]");
                try {
                    System.in.read();
                } catch (IOException e) {
                    e.printStackTrace();
                }
            }
        }
    }

    static void DBWritePlayer() {
        PrintWriter wfp;
        DBSortPlayer("reverse");
        try {
            wfp = new PrintWriter(new FileWriter(PLAYERFILE));
            for (int i = 0; i < totalplayers; i++) {
                wfp.println(playerdb[i].player + " " +
                        playerdb[i].wins + " " +
                        playerdb[i].losses + " " +
                        playerdb[i].draws);
            }
            wfp.close();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    static void DBReadPlayer() {
        totalplayers = 0;
        try {
            BufferedReader rfp = new BufferedReader(new FileReader(PLAYERFILE));
            while (true) {
                try {
                    String line = rfp.readLine();
                    if (line == null)
                        break;  // eof
                    //int n = fscanf(rfp,"%49s %d %d %d\n",  /* 49 MAXNAMESZ-1 */
                    playerdb[totalplayers].player = line.substring(0, 49);
                    line = line.substring(49);  // the rest
                    StringTokenizer st = new StringTokenizer(line);
                    playerdb[totalplayers].wins = Integer.parseInt(st.nextToken());
                    playerdb[totalplayers].losses = Integer.parseInt(st.nextToken());
                    playerdb[totalplayers].draws = Integer.parseInt(st.nextToken());
                    totalplayers++;
                } catch (IOException e) {
                    e.printStackTrace();
                }
            }
            rfp.close();
        } catch (FileNotFoundException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        }

    }

    static int DBSearchPlayer(String player) {
        int index = -1;
        int i;

        for (i = 0; i < totalplayers; i++)
            if (player.startsWith(playerdb[i].player))  // fixc use equality
            {
                index = i;
                break;
            }
        return (index);
    }

    static void DBUpdatePlayer(String player, String resultstr) {
        int index;
        int result = Book.R_NORESULT;

        lname = "";
        // copy the non-spaces
        // fixc this isn't really the last name since we copy the whole string (except for spaces)
        for (StringTokenizer st = new StringTokenizer(player); st.hasMoreTokens();)
            lname += st.nextToken();

        //  playerdb
        DBReadPlayer();
        index = DBSearchPlayer(lname);
        if (index == -1) {
            playerdb[totalplayers].player = lname;
            playerdb[totalplayers].wins = 0;
            playerdb[totalplayers].losses = 0;
            playerdb[totalplayers].draws = 0;
            index = totalplayers;
            totalplayers++;
        }
        if (resultstr.startsWith("1-0"))
            result = Book.R_WHITE_WINS;
        else if (resultstr.startsWith("0-1"))
            result = Book.R_BLACK_WINS;
        else if (resultstr.startsWith("1/2-1/2"))
            result = Book.R_DRAW;

        if ((computerplays == white && result == Book.R_WHITE_WINS) ||
                (computerplays == black && result == Book.R_BLACK_WINS))
            playerdb[index].wins++;
        else if ((computerplays == white && result == Book.R_BLACK_WINS) ||
                (computerplays == black && result == Book.R_WHITE_WINS))
            playerdb[index].losses++;
        else
        /* Shouln't one check for draw here? Broken PGN files surely exist */
            playerdb[index].draws++;
        DBWritePlayer();
    }
}
