/* This file is part of chessbox_gnu4j, a Java port of GNU Chess (5.0.7).

   Copyright (c) 2005 Xan Gregg (Java port)
   Copyright (c) 1999-2002 Free Software Foundation, Inc.

   chessbox_gnu4j is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 or (at your option)
   any later version.

   chessbox_gnu4j is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GNU Chess; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.

   Contact Info:
     bug-gnu-chess@gnu.org
     xan.gregg@pobox.com
*/
package org.chessbox.gnu4j;
@SuppressWarnings("all")
/**
 * Created by Xan Gregg.
 * Date: May 8, 2005
 */
public class Quiesce extends Common {
    /**
     * Our quiescent search.  This quiescent search is able to recognize
     * mates.
     */
    static int Quiesce(int ply, int alpha, int beta) {
        int side, xside;
        int best, delta, score, savealpha;
        int p, pbest;

        if (Eval.EvaluateDraw())
            return (DRAWSCORE);

        side = board.side;
        xside = 1 ^ side;
        InChk[ply] = Atak.SqAtakd(board.king[side], xside);
        best = Eval.Evaluate(alpha, beta);
        if (best >= beta && !InChk[ply])
            return (best);
        TreePtr[ply + 1] = TreePtr[ply];
        if (InChk[ply]) {
            GenMove.GenCheckEscapes(ply);
            if (TreePtr[ply] == TreePtr[ply + 1])
                return (-MATE + ply - 2);
            if (best >= beta)
                return (best);
            Sort.SortMoves(ply);
        } else {
            GenMove.GenCaptures(ply);
            if (TreePtr[ply] == TreePtr[ply + 1])
                return (best);
            Sort.SortCaptures(ply);
        }

        savealpha = alpha;
        pbest = -1;
        alpha = MAX(best, alpha);
        delta = MAX(alpha - 150 - best, 0);

        done:
        for (p = TreePtr[ply]; p < TreePtr[ply + 1]; p++) {
            Sort.pick(p, ply);

            /* We are in check or capture cannot bring score near alpha, give up */
            if (!InChk[ply] && Swap.SwapOff(Tree[p].move) < delta)
                continue;

            /* If capture cannot bring score near alpha, give up */
            if (Tree[p].score == -INFINITY)
                continue;

            Tree[p].move = Move.MakeMove(side, Tree[p].move);
            QuiesCnt++;
            if (Atak.SqAtakd(board.king[side], xside)) {
                Tree[p].move = Move.UnmakeMove(xside, Tree[p].move);
                continue;
            }
            score = -Quiesce(ply + 1, -beta, -alpha);
            Tree[p].move = Move.UnmakeMove(xside, Tree[p].move);
            if (score > best) {
                best = score;
                pbest = p;
                if (best >= beta)
                    break done;
                alpha = MAX(alpha, best);
            }
        }


        if ((flags & USEHASH) != 0 && pbest != -1)
            TTable.TTPut(side, 0, ply, savealpha, beta, best, Tree[pbest].move);

        return (best);
    }

}
