/* This file is part of chessbox_gnu4j, a Java port of GNU Chess (5.0.7).

   Copyright (c) 2005 Xan Gregg (Java port)
   Copyright (c) 1999-2002 Free Software Foundation, Inc.

   chessbox_gnu4j is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 or (at your option)
   any later version.

   chessbox_gnu4j is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GNU Chess; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.

   Contact Info:
     bug-gnu-chess@gnu.org
     xan.gregg@pobox.com
*/
package org.chessbox.gnu4j;
@SuppressWarnings("all")
/**
 * Created by Xan Gregg.
 * Date: May 8, 2005
 */
public class Search extends Common {
    static final int TIMECHECK = 1023;

    static int HISTSCORE(int d) { return d * d;}

    static int FUTSCORE(int side, int fdel) { return MATERIAL(side) + fdel;}

    static int GETNEXTMOVE(int ply) {
        return (InChk[ply] ? Sort.PhasePick1(ply) : Sort.PhasePick(ply));
    }

    static void ShowThinking(int p, int ply) {
        if ((flags & XBOARD) != 0)
            return;
        if ((flags & POST) == 0)
            return;
        if (NodeCnt < 500000 && (flags & SOLVE) != 0)
            return;

        Move.SANMove(Tree[p].move, ply);
        System.out.print("\r" + pad((int) (p - TreePtr[ply] + 1), 2) + ".         " +
                pad((int) (TreePtr[ply + 1] - TreePtr[ply]), 2) + "/" +
                pad(Idepth, 2) + pad(SANmv, 10) + "    ");
        System.out.flush();
    }

    static int ply1score;

    /**
     * This perform searches at ply=1.  For ply>1, it calls the more generic
     * search() routine.  The rationale for splitting these is because at
     * ply==1, things are done slightly differently than from the other plies,
     * e.g. print PVs, not testing null move etc.
     */
    static int SearchRoot(int depth, int alpha, int beta) {
        int best, score, savealpha;
        int side, xside, ply;
        short nodetype;
        int p, pbest;

        ply = 1;
        side = board.side;
        xside = 1 ^ side;
        ChkCnt[2] = ChkCnt[1];
        ThrtCnt[2] = ThrtCnt[1];
        KingThrt[white][ply] = Atak.MateScan(white);
        KingThrt[black][ply] = Atak.MateScan(black);
        InChk[ply] = Atak.SqAtakd(board.king[side], xside);
        if (InChk[ply] && ChkCnt[ply] < 3 * Idepth) {
            ChkExtCnt++;
            ChkCnt[ply + 1]++;
            depth += 1;
        }
        best = -INFINITY;
        savealpha = alpha;
        nodetype = PV;
        pbest = -1;

        done:
        // artifical loop to convert goto to break statement
        while (true) {
            for (p = TreePtr[1]; p < TreePtr[2]; p++) {
                Sort.pick(p, 1);
                ShowThinking(p, ply);
                Tree[p].move = Move.MakeMove(side, Tree[p].move);
                NodeCnt++;

                /*  If first move, search against full alpha-beta window  */
                if (p == TreePtr[1]) {
                    score = -Search(2, depth - 1, -beta, -alpha, nodetype);
                    /*
                  The following occurs when we are re-searching a fail high move
                      and now it has fail low.  This can be disastrous, so immediately
                  adjust alpha and research.
                    */
                    if (beta == INFINITY && score <= alpha) {
                        alpha = -INFINITY;
                        score = -Search(2, depth - 1, -beta, -alpha, nodetype);
                    }
                }

                /*  Else search against zero window  */
                else {
                    nodetype = CUT;
                    alpha = MAX(best, alpha);
                    score = -Search(2, depth - 1, -alpha - 1, -alpha, nodetype);
                    if (score > best) {
                        if (alpha < score && score < beta) {
                            nodetype = PV;
                            score = -Search(2, depth - 1, -beta, -score, nodetype);
                        }
                    }
                }
                Tree[p].move = Move.UnmakeMove(xside, Tree[p].move);

                ply1score = Tree[p].score = score;
                if (score > best) {
                    best = score;
                    pbest = p;
                    if (best > alpha) {
                        rootscore = best;
                        RootPV = Tree[p].move;
                        if (best >= beta)
                            break done;
                        ShowLine(RootPV, best, '&');
                    }
                }

                if ((flags & TIMEOUT) != 0) {
                    /* ply == 1 always at this point, but code
                     * copied from Search
                     */
                    best = (ply & 1) != 0 ? rootscore : -rootscore;
                    return (best);
                }

                if (((flags & PONDER) != 0 || SearchDepth == 0) && (NodeCnt & TIMECHECK) == 0) {
                    if ((flags & PONDER) != 0) {
                        if (Input.input_status != INPUT_NONE)
                            SetFlag(TIMEOUT);
                    } else {
                        ElapsedTime = Iterate.GetElapsed(StartTime);
                        if ((ElapsedTime >= SearchTime && (
                                rootscore == -INFINITY - 1
                                        || ply1score > lastrootscore - 25
                                        || (flags & SOLVE) != 0))
                                || ElapsedTime >= maxtime)
                            SetFlag(TIMEOUT);
                    }
                }

                if (MATE + 1 == best + 1)
                    return (best);
            }

/*  If none of the move is good, we still want to try the same first move */
            if (best <= savealpha)
                Tree[TreePtr[1]].score = savealpha;

            break;
        }
        /*****************************************************************************
         *
         *  Out of main search loop.
         *
         *****************************************************************************/

        /*  Update history  */
        if (best > savealpha)
            history[side][Tree[pbest].move & 0x0FFF] += HISTSCORE(depth);

        rootscore = best;
        return (best);
    }


    /**
     * The basic algorithm for this search routine came from Anthony
     * Marsland.  It is a PVS (Principal Variation Search) algorithm.
     * The fail-soft alpha-beta technique is also used for improved
     * pruning.
     */
    static int Search(int ply, int depth, int alpha, int beta, int nodetype) {
        int best, score, nullscore, savealpha;
        int side, xside;
        int rc, t0, t1;
        boolean firstmove;
        int fdel, savenode;
        boolean fcut, donull, extend;
        int p, pbest;
        int g0, g1;
//        int upperbound;

        /* Check if this position is a known draw */
        if (Eval.EvaluateDraw())
            return (DRAWSCORE);
        if (GameCnt >= Game50 + 3 && Repeat.Repeat() != 0) {
            RepeatCnt++;
            return (DRAWSCORE);
        }

        side = board.side;
        xside = 1 ^ side;
        donull = true;

        /*************************************************************************
         *
         *  Perform some basic search extensions.
         *  1.  One reply extensions.
         *  2.  If in check, extend (maximum of Idepth-1).
         *  3.  If there is a threat to the King, extend (not beyond 2*Idepth)
         *  4.  If recapture to same square and not beyond Idepth+2
         *  5.  If pawn move to 7th rank at the leaf node, extend.
         *
         *************************************************************************/
        extend = false;
        InChk[ply] = Atak.SqAtakd(board.king[side], xside);
        if (InChk[ply]) {
            TreePtr[ply + 1] = TreePtr[ply];
            GenMove.GenCheckEscapes(ply);
            if (TreePtr[ply] == TreePtr[ply + 1])
                return (-MATE + ply - 2);
            if (TreePtr[ply] + 1 == TreePtr[ply + 1]) {
                depth += 1;
                extend = true;
                OneRepCnt++;
            }
        }

/*
   We've already found a mate at the next ply.  If we aren't being mated by
   a shorter line, so just return the current material value.
*/
        if (rootscore + ply >= MATE)
            return (MATERIAL(side));

        g0 = Game[GameCnt].move;
        g1 = GameCnt > 0 ? Game[GameCnt - 1].move : 0;
        t0 = TOSQ(g0);
        t1 = TOSQ(g1);
        ChkCnt[ply + 1] = ChkCnt[ply];
        ThrtCnt[ply + 1] = ThrtCnt[ply];
        KingThrt[white][ply] = Atak.MateScan(white);
        KingThrt[black][ply] =
                Atak.MateScan(black);
        if (InChk[ply] && /* ChkCnt[ply] < Idepth-1*/ ply <= 2 * Idepth) {
            ChkExtCnt++;
            ChkCnt[ply + 1]++;
            depth += 1;
            extend = true;
        } else if (!KingThrt[side][ply - 1] && KingThrt[side][ply] && ply <= 2 * Idepth) {
            KingExtCnt++;
            extend = true;
            depth += 1;
            extend = true;
            donull = false;
        }
        /* Promotion extension */
        else if ((g0 & PROMOTION) != 0) {
            PawnExtCnt++;
            depth += 1; /* Not reached, but why?! */
            extend = true;
        }
        /* Recapture extension */
        else if ((g0 & CAPTURE) != 0 && (board.material[computer] -
                board.material[1 ^ computer] == RootMaterial)) {
            RcpExtCnt++;
            depth += 1;
            extend = true;
        }
        /* 6th or 7th rank extension */
        else if (depth <= 1 && cboard[t0] == pawn && (RANK(t0) == rank7[xside] || RANK(t0) == rank6[xside])) {
            PawnExtCnt++;
            depth += 1;
            extend = true;
        }

        /****************************************************************************
         *
         *  The following extension is to handle cases when the opposing side is
         *  delaying the mate by useless interposing moves.
         *
         ****************************************************************************/
        if (ply > 2 && InChk[ply - 1] && cboard[t0] != king && t0 != t1 &&
                !Atak.SqAtakd(t0, xside)) {
            HorzExtCnt++;
            depth += 1;
            extend = true;
        }

        /***************************************************************************
         *
         *  This is a new code to perform search reductiion.  We introduce some
         *  form of selectivity here.
         *
         **************************************************************************/

        if (depth <= 0)
            return (Quiesce.Quiesce(ply, alpha, beta));

        /****************************************************************************
         *
         *  Probe the transposition table for a score and a move.
         *  If the score is an upperbound, then we can use it to improve the value
         *  of beta.  If a lowerbound, we improve alpha.  If it is an exact score,
         *  if we now get a cut-off due to the new alpha/beta, return the score.
         *
         ***************************************************************************/
        Hashmv[ply] = 0;
        if ((flags & USEHASH) != 0) {
            int[] scoreMove = new int[]{0, 0}; // simulate pass by reference
            rc = TTable.TTGet(side, depth, ply, scoreMove);
            score = scoreMove[0];
            if (rc != 0) {
                Hashmv[ply] = scoreMove[1] & MOVEMASK;
                switch (rc) {
                    case POORDRAFT:  /* Not reached */
                        break;
                    case EXACTSCORE:  /* Not reached */
                        return (score);
                    case UPPERBOUND:
                        beta = MIN(beta, score);
                        donull = false;
                        break;
                    case LOWERBOUND:  /*alpha = MAX (alpha, score);*/
                        alpha = score;
                        break;
                    case QUIESCENT:
                        Hashmv[ply] = 0;
                        break;
                    default :
                        break;
                }
                if (alpha >= beta)
                    return (score);
            }
        }

        /*****************************************************************************
         *
         *  Perform the null move here.  There are certain cases when null move
         *  is not done.
         *  1.  When the previous move is a null move.
         *
         2.  At the frontier (depth == 1)
         *  3.  At a PV node.
         *  4.  If side to move is in check.
         *  5.  If the material score + pawn value is still below beta.
         *  6.  If we are being mated at next ply.
         *  7.  If hash table indicate the real score is below beta (UPPERBOUND).
         *  8.  If side to move has less than or equal to a bishop in value.
         *  9.  If Idepth <= 3.  This allows us to find mate-in 2 problems quickly.
         *  10. We are looking for a null threat.
         *
         *****************************************************************************/
        if (ply > 4 && InChk[ply - 2] && InChk[ply - 4])
            donull = false;
        if ((flags & USENULL) != 0 && g0 != NULLMOVE && depth > 1 && nodetype != PV &&
                !InChk[ply] &&
                MATERIAL(side) + ValueP > beta && beta > -MATE + ply &&
                donull &&
                board.pmaterial[side] > ValueB) {
            TreePtr[ply + 1] = TreePtr[ply];
            Null.MakeNullMove(side);
            nullscore = -Search(ply + 1, depth - 3, -beta, -beta + 1, nodetype);
            Null.UnmakeNullMove(xside);
            if (nullscore >= beta) {
                NullCutCnt++;
                return (nullscore);
            }
            if (depth - 3 >= 1 && MATERIAL(side) > beta && nullscore <= -MATE + 256) {
                depth += 1;
                extend = true;
            }
        }

        if (InChk[ply] && TreePtr[ply] + 1 < TreePtr[ply + 1])
            Sort.SortMoves(ply);

        pickphase[ply] = PICKHASH;
        p = GETNEXTMOVE(ply);

        /*************************************************************************
         *
         *  Razoring + Futility.
         *  At depth 3, if there is no extensions and we are really bad, decrease
         *  the search depth by 1.
         *  At depth 2, if there is no extensions and we are quite bad, then we
         *  prune all non checking moves and capturing moves that don't bring us up
         *  back to alpha.
         *  Caveat: Skip all this if we are in the ending.
         **************************************************************************/
        fcut = false;
        fdel = MAX(ValueQ, maxposnscore[side]);
        if (!extend && nodetype != PV
                && depth == 3 && FUTSCORE(side, fdel) <= alpha) {
            depth = 2;
            RazrCutCnt++;
        }
        fdel = MAX(ValueR, maxposnscore[side]);
        fcut = (!extend && nodetype != PV && depth == 2 && FUTSCORE(side, fdel) <= alpha);
        if (!fcut) {
            fdel = MAX(3 * ValueP, maxposnscore[side]);
            fcut = (nodetype != PV && depth == 1 && FUTSCORE(side, fdel) <= alpha);
        }

        Tree[p].move = Move.MakeMove(side, Tree[p].move);
        NodeCnt++;
        g1 = 0;
        boolean attacked;
        while ((attacked = Atak.SqAtakd(board.king[side], xside)) ||
                (fcut && FUTSCORE(side, fdel) < alpha && !Atak.SqAtakd(board.king[xside], side) &&
                        !Atak.MateScan(xside))) {
            if (!attacked) g1++;
            Tree[p].move = Move.UnmakeMove(xside, Tree[p].move);
            p = GETNEXTMOVE(ply);
            if (p < 0)
                return (g1 != 0 ? Eval.Evaluate(alpha, beta) : DRAWSCORE);
            Tree[p].move = Move.MakeMove(side, Tree[p].move);
            NodeCnt++;
        }
        firstmove = true;
        pbest = p;
        best = -INFINITY;
        savealpha = alpha;
        nullscore = INFINITY;
        savenode = nodetype;
        if (nodetype != PV)
            nodetype = (nodetype == CUT) ? ALL : CUT;

        done:
        while (true) {
            /* We have already made the move before the loop. */
            if (firstmove) {
                firstmove = false;
                score = -Search(ply + 1, depth - 1, -beta, -alpha, nodetype);
            }

            /* Zero window search for rest of moves */
            else {
                p = GETNEXTMOVE(ply);
                if (p < 0)
                    break;

                Tree[p].move = Move.MakeMove(side, Tree[p].move);
                NodeCnt++;
                if (Atak.SqAtakd(board.king[side], xside)) {
                    Tree[p].move = Move.UnmakeMove(xside, Tree[p].move);
                    continue;
                }

                /*****************************************************************************
                 *
                 *  Futility pruning.  The idea is that at the frontier node (depth == 1),
                 *  if the side on the move is materially bad, then if the move doesn't win
                 *  back material or the move isn't a check or doesn't threatened the king,
                 *  then there is no point in searching this move.  So skip it.
                 *  Caveat:  However if the node is a PV, we skip this test.
                 *
                 *****************************************************************************/
                if (fcut && FUTSCORE(side, fdel) <= alpha && !Atak.SqAtakd(board.king[xside], side) &&
                        !Atak.MateScan(xside)) {
                    Tree[p].move = Move.UnmakeMove(xside, Tree[p].move);
                    FutlCutCnt++;
                    NodeCnt--;
                    continue;
                }
                NodeCnt++;

                if (nodetype == PV)
                    nodetype = CUT;
                alpha = MAX(best, alpha);                /* fail-soft condition */
                score = -Search(ply + 1, depth - 1, -alpha - 1, -alpha, nodetype);
                if (score > best) {
                    if (savenode == PV)
                        nodetype = PV;
                    if (alpha < score && score < beta) {
                        score = -Search(ply + 1, depth - 1, -beta, -score, nodetype);
                    }
                    if (nodetype == PV && score <= alpha &&
                            Game[GameCnt + 1].move == NULLMOVE) {
                        score = -Search(ply + 1, depth - 1, -alpha, INFINITY, nodetype);
                    }
                }
            }

            Tree[p].move = Move.UnmakeMove(xside, Tree[p].move);

            if (score > best) {
                best = score;
                pbest = p;
                if (best >= beta)
                    break done;
            }

            if ((flags & TIMEOUT) != 0) {
                best = (ply & 1) != 0 ? rootscore : -rootscore;
                return (best);
            }

            if (((flags & PONDER) != 0 || SearchDepth == 0) && (NodeCnt & TIMECHECK) == 0) {
                if ((flags & PONDER) != 0) {
                    if (Input.input_status != INPUT_NONE)
                        SetFlag(TIMEOUT);
                } else {
                    ElapsedTime = Iterate.GetElapsed(StartTime);
                    if ((ElapsedTime >= SearchTime &&
                            (rootscore == -INFINITY - 1 ||
                                    ply1score > lastrootscore - 25 || (flags & SOLVE) != 0)) ||
                            ElapsedTime >= maxtime)
                        SetFlag(TIMEOUT);
                }
            }

/*  The following line should be explained as I occasionally forget too :) */
/*  This code means that if at this ply, a mating move has been found,     */
/*  then we can skip the rest of the moves!  				   */
            if (MATE + 1 == best + ply)
                break done;
        }

        /*****************************************************************************
         *
         *  Out of main search loop.
         *
         *****************************************************************************/
        // done:

/*
   if (upperbound < best)
      System.out.println ("Inconsistencies %d %d\n", upperbound, best);
*/

        /*  Save the best move inside the transposition table  */
        if ((flags & USEHASH) != 0) {
/*
 * Nasty temporary hack to try and work around timeout problem
 * If we are pondering and timeout don't save incomplete answers
 * Must look at failure of TIMEOUT condition more carefully!
 */
            if ((flags & TIMEOUT) == 0)
                TTable.TTPut(side, depth, ply, savealpha, beta, best, Tree[pbest].move);
        }

        /*  Update history  */
        if (best > savealpha)
            history[side][Tree[pbest].move & 0x0FFF] += HISTSCORE(depth);

        /*  Don't store captures as killers as they are tried before killers */
        if ((Tree[pbest].move & (CAPTURE | PROMOTION)) == 0 && best > savealpha) {
            if (killer1[ply] == 0)
                killer1[ply] = Tree[pbest].move & MOVEMASK;
            else if ((Tree[pbest].move & MOVEMASK) != killer1[ply])
                killer2[ply] = Tree[pbest].move & MOVEMASK;
        }

        return (best);
    }


    /**
     * Print out the latest PV found during the search.
     * The only move we know is the root move.  The rest of the PV is taken
     * from the hash table.  This strategy avoids all the headaches associated
     * with returning the PV up from the leaf to the root.
     */
    static void ShowLine(int move, int score, char c) {
        int i, len;
        int[] pvar = new int[MAXPLYDEPTH];

        /* SMC */
        if ((flags & POST) == 0)
            return;
        if (NodeCnt < 500000 && (flags & SOLVE) != 0) {
            /* System.out.println("NodeCnt = %d\n",NodeCnt); getchar(); */
            return;
        }
        if (Idepth == 1 && c == '&')
            return;
        if ((flags & XBOARD) != 0 && c == '&')
            return;
        if (rootscore == -INFINITY - 1)
            return;
        ElapsedTime = Iterate.GetElapsed(StartTime);

        /*
         * The different output formats for Xboard and GNU Chess are documented
         * in the engine protocol guide.
         *
         * In particular if the character after ply is not a space, Xboard
         * assume it is talking to a GNU Chess compatible engine and
         * uses time in seconds, not centiseconds.
         *
         * This code should be simplified!
         *
         */
        if ((flags & XBOARD) != 0) {
            if (score > MATE - 255) {
                System.out.print(Idepth + c +
                        " Mat" + (MATE + 2 - Math.abs(score)) / 2 + " " + (NodeCnt + QuiesCnt) + "\t");
                if (!ofpIsStdout)
                    ofp.print(pad(Idepth, 2) + c + pad(round(ElapsedTime, 2), 10) +
                            "  Mat" + pad0((MATE + 2 - Math.abs(score)) / 2, 2) + pad(NodeCnt + QuiesCnt, 10) + "\t");
            } else if (score < -MATE + 255) {
                System.out.print(Idepth + c + " -Mat"
                        + pad((int) (MATE + 2 - Math.abs(score)) / 2, 2) + " "
                        + (int) (ElapsedTime) + " " + (NodeCnt + QuiesCnt) + "\t");
                if (!ofpIsStdout)
                    ofp.print(pad(Idepth, 2) + c + pad(round(ElapsedTime, 2), 10) +
                            " -Mat" + pad0((MATE + 2 - Math.abs(score)) / 2, 2) + pad(NodeCnt + QuiesCnt, 10) + "\t");
            } else {
                System.out.print(Idepth + c + " " + (int) score + " " +
                        (int) (ElapsedTime) + " " + (NodeCnt + QuiesCnt) + "\t");
                if (!ofpIsStdout)
                    ofp.print(pad(Idepth, 2) + c + pad(round(ElapsedTime, 2), 10)
                            + pad(score, 7) + pad(NodeCnt + QuiesCnt, 10) + "\t");
            }
        } else {              /* Not XBOARD */
            if (score > MATE - 255) {
                System.out.print("\r" + pad(Idepth, 2) + c + pad(round(ElapsedTime, 2), 10) +
                        "  Mat" + pad0((MATE + 2 - Math.abs(score)) / 2, 2) + pad(NodeCnt + QuiesCnt, 10) + "\t");
                if (!ofpIsStdout)
                    ofp.print("\r" + pad(Idepth, 2) + c + pad(round(ElapsedTime, 2), 10) +
                            "  Mat" + pad0((MATE + 2 - Math.abs(score)) / 2, 2) + pad(NodeCnt + QuiesCnt, 10) + "\t");
            } else if (score < -MATE + 255) {
                System.out.print("\r" + pad(Idepth, 2) + c + pad(round(ElapsedTime, 2), 10) +
                        " -Mat" + pad0((MATE + 2 - Math.abs(score)) / 2, 2) + pad(NodeCnt + QuiesCnt, 10) + "\t");
                if (!ofpIsStdout)
                    ofp.print("\r" + pad(Idepth, 2) + c + pad(round(ElapsedTime, 2), 10) +
                            " -Mat" + pad0((MATE + 2 - Math.abs(score)) / 2, 2) + pad(NodeCnt + QuiesCnt, 10) + "\t");
            } else {
                System.out.print("\r" + pad(Idepth, 2) + c + pad(round(ElapsedTime, 2), 10)
                        + pad(score, 7) + pad(NodeCnt + QuiesCnt, 10) + "\t");
                if (!ofpIsStdout)
                    ofp.print("\r" + pad(Idepth, 2) + c + pad(round(ElapsedTime, 2), 10)
                            + pad(score, 7) + pad(NodeCnt + QuiesCnt, 10) + "\t");
            }
        }

        if (c == '-') {
            System.out.println();
            if (!ofpIsStdout) ofp.println();
            return;
        } else if (c == '+') {
            Move.SANMove(RootPV, 1);
            System.out.println(" " + SANmv);
            if (!ofpIsStdout) ofp.println(" " + SANmv);
            return;
        }

        Move.SANMove(RootPV, 1);
        System.out.print(" " + SANmv);
        if (!ofpIsStdout) ofp.println(" " + SANmv);
        RootPV = Move.MakeMove(board.side, RootPV);
        TreePtr[3] = TreePtr[2];
        GenMove.GenMoves(2);
        len = SANmv.length();
        i = 2;
        pvar[1] = RootPV;

        /*  We fill the rest of the PV with moves from the hash table */
        if ((flags & USEHASH) != 0) {
            while (true) {
                pvar[i] = TTable.TTGetPV(board.side, i, rootscore);
                if (pvar[i] == -1)
                    break;

                if ((MATESCORE(score) != 0 && Math.abs(score) == MATE + 2 - i) || Repeat.Repeat() != 0)
                    break;

                if (len >= 32) {
                    System.out.print("\n\t\t\t\t");
                    if (!ofpIsStdout) ofp.print("\n\t\t\t\t");
                    len = 0;
                }
                Move.SANMove(pvar[i], i);
                System.out.print(" " + SANmv);
                if (!ofpIsStdout) ofp.print(" " + SANmv);
                pvar[i] = Move.MakeMove(board.side, pvar[i]);
                TreePtr[i + 2] = TreePtr[i + 1];
                GenMove.GenMoves(++i);
                len += SANmv.length();
            }
        }

        System.out.println();
        if (!ofpIsStdout) ofp.println();
        for (--i; i != 0; i--)
            pvar[i] = Move.UnmakeMove(board.side, pvar[i]);
        System.out.flush();
        if (!ofpIsStdout) ofp.flush();
    }
}
