/* This file is part of chessbox_gnu4j, a Java port of GNU Chess (5.0.7).

   Copyright (c) 2005 Xan Gregg (Java port)
   Copyright (c) 1999-2002 Free Software Foundation, Inc.

   chessbox_gnu4j is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 or (at your option)
   any later version.

   chessbox_gnu4j is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GNU Chess; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.

   Contact Info:
     bug-gnu-chess@gnu.org
     xan.gregg@pobox.com
*/
package org.chessbox.gnu4j;

/**
 * Created by Xan Gregg.
 * Date: May 7, 2005
 */
public class Sort extends Common {
    static final int WEIGHT = 12;
    static final int HASHSORTSCORE = INFINITY;
    static final int KILLERSORTSCORE = 1000;
    static final int CASTLINGSCORE = 500;

    /**
     * Actually no sorting is done.  Just scores are assigned to the captures.
     * When we need a move, we will select the highest score move from the
     * list, place it at the top and try it.  This move might give us a cut
     * in which case, there is no reason to sort.
     */
    static void SortCaptures(int ply) {
        int p;
        int temp, f, t;

        for (p = TreePtr[ply]; p < TreePtr[ply + 1]; p++) {
            f = Value[cboard[FROMSQ(Tree[p].move)]];
            t = Value[cboard[TOSQ(Tree[p].move)]];
            if (f < t)
                Tree[p].score = t - f;
            else {
                temp = Swap.SwapOff(Tree[p].move);
                Tree[p].score = (temp < 0 ? -INFINITY : temp);
            }

        }
    }


    /**
     * Sort criteria is as follows.
     * 1.  The move from the hash table
     * 2.  Captures as above.
     * 3.  Killers.
     * 4.  History.
     * 5.  Moves to the centre.
     */
    static void SortMoves(int ply) {
        int p;
        int f, t, m, tovalue;
        int side, xside;
        /*BitBoard*/long enemyP;

        side = board.side;
        xside = 1 ^ side;
        enemyP = board.b[xside][pawn];

        for (p = TreePtr[ply]; p < TreePtr[ply + 1]; p++) {
            Tree[p].score = -INFINITY;
            f = FROMSQ(Tree[p].move);
            t = TOSQ(Tree[p].move);
            m = Tree[p].move & MOVEMASK;

            /* Hash table move (highest score) */
            if (m == Hashmv[ply])
                Tree[p].score += HASHSORTSCORE;

            else if (cboard[t] != 0 || (Tree[p].move & PROMOTION) != 0) {

                /* ***** SRW My Interpretation of this code *************
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                       * Captures normally in other places but.....         *
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                       *                                                    *
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                       * On capture we generally prefer to capture with the *
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   * with the lowest value piece so to chose between    *
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                       * pieces we should subtract the piece value .... but *
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                       *                                                    *
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                       * The original code was looking at some captures     *
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                       * last, especially where the piece was worth more    *
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                       * than the piece captured - KP v K in endgame.epd    *
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                       *                                                    *
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                       * So code modified to prefer any capture by adding   *
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                       * ValueK                                             *
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                       ****************************************************** */

                tovalue = (Value[cboard[t]] + Value[PROMOTEPIECE(Tree[p].move)]);
                Tree[p].score += tovalue + ValueK - Value[cboard[f]];


            }
            /* Killers */
                    else if (m == killer1[ply] || m == killer2[ply])
                Tree[p].score += KILLERSORTSCORE;
            else if (ply > 2 && (m == killer1[ply - 2] || m == killer2[ply - 2]))
                Tree[p].score += KILLERSORTSCORE;

            Tree[p].score += history[side][(Tree[p].move & 0x0FFF)] + taxicab[f][D5] - taxicab[t][E4];

            if (cboard[f] == pawn) {
                /* Look at pushing Passed pawns first */
                if ((enemyP & PassedPawnMask[side][t]) == NULLBITBOARD)
                    Tree[p].score += 50;
            }
        }
    }


    /**
     * Sort the moves at the root.  The heuristic is simple.  Try captures/
     * promotions first.  Other moves are ordered based on their swapoff values.
     */
    static void SortRoot() {
        int p;
        int f, t ;
        int side, xside;
        /*BitBoard*/long enemyP;

        side = board.side;
        xside = 1 ^ side;
        enemyP = board.b[xside][pawn];

        for (p = TreePtr[1]; p < TreePtr[2]; p++) {
            f = Value[cboard[FROMSQ(Tree[p].move)]];
            if (cboard[TOSQ(Tree[p].move)] != 0 || (Tree[p].move & PROMOTION) != 0) {
                t = Value[cboard[TOSQ(Tree[p].move)]];
                if (f < t)
                    Tree[p].score = -1000 + t - f;
                else
                    Tree[p].score = -1000 + Swap.SwapOff(Tree[p].move);
            } else
                Tree[p].score = -3000 + Swap.SwapOff(Tree[p].move);

            Tree[p].score += taxicab[FROMSQ(Tree[p].move)][D5] - taxicab[TOSQ(Tree[p].move)][E4];

            if (f == ValueP) {
                /* Look at pushing Passed pawns first */
                if ((enemyP & PassedPawnMask[side][TOSQ(Tree[p].move)]) == NULLBITBOARD)
                    Tree[p].score += 50;
            }
        }
    }


    /**
     * This pick routine searches the movelist and swap the high score entry
     * with the one currently at the head.
     */
    static void pick(/*leaf* */int head, int ply) {
        int best;
        int p, pbest;

        best = Tree[head].score;
        pbest = head;
        for (p = head + 1; p < TreePtr[ply + 1]; p++) {
            if (Tree[p].score > best) {
                pbest = p;
                best = Tree[p].score;
            }
        }

        /*  Swap pbest with the head  */
        Tree[head].swap(Tree[pbest]);
    }

    static int[] PhasePick_p = new int[MAXPLYDEPTH];    // was a static local

    /**
     * A phase style routine which returns the next move to the search.
     * Hash move is first returned.  If it doesn't fail high, captures are
     * generated, sorted and tried.  If no fail high still occur, the rest of
     * the moves are generated and tried.
     * The idea behind all this is to save time generating moves which might
     * not be needed.
     * CAVEAT: To implement this, the way that genmoves & friends are called
     * have to be modified.  In particular, TreePtr[ply+1] = TreePtr[ply] must
     * be set before the calls can be made.
     * If the board ever gets corrupted during the search, then there is a bug
     * in IsLegalMove() which has to be fixed.
     */
    static /*leaf*/int PhasePick(int ply)  // return -1 for no pick
    {
        int p1; // result
        int p2;  // leaf *
        int mv;
        int side;

        side = board.side;
        switch (pickphase[ply]) {
            case PICKHASH:
                TreePtr[ply + 1] = TreePtr[ply];
                pickphase[ply] = PICKGEN1;
                if (Hashmv[ply] != 0 && Move.IsLegalMove(Hashmv[ply])) {
                    Tree[TreePtr[ply + 1]].move = Hashmv[ply];
                    p1 = TreePtr[ply + 1]++;
                    return p1;
                }

            case PICKGEN1:
                pickphase[ply] = PICKCAPT;
                PhasePick_p[ply] = TreePtr[ply + 1];
                GenMove.GenCaptures(ply);
                for (p2 = PhasePick_p[ply]; p2 < TreePtr[ply + 1]; p2++)
                    Tree[p2].score = Swap.SwapOff(Tree[p2].move) * WEIGHT +
                            Value[cboard[TOSQ(Tree[p2].move)]];

            case PICKCAPT:
                while (PhasePick_p[ply] < TreePtr[ply + 1]) {
                    pick(PhasePick_p[ply], ply);
                    if ((Tree[PhasePick_p[ply]].move & MOVEMASK) == Hashmv[ply]) {
                        PhasePick_p[ply]++;
                        continue;
                    }
                    p1 = PhasePick_p[ply]++;
                    return p1;
                }

            case PICKKILL1:
                pickphase[ply] = PICKKILL2;
                if (killer1[ply] != 0 && killer1[ply] != Hashmv[ply] &&
                        Move.IsLegalMove(killer1[ply])) {
                    Tree[TreePtr[ply + 1]].move = killer1[ply];
                    p1 = TreePtr[ply + 1];
                    TreePtr[ply + 1]++;
                    return p1;
                }

            case PICKKILL2:
                pickphase[ply] = PICKGEN2;
                if (killer2[ply] != 0 && killer2[ply] != Hashmv[ply] &&
                        Move.IsLegalMove(killer2[ply])) {
                    Tree[TreePtr[ply + 1]].move = killer2[ply];
                    p1 = TreePtr[ply + 1];
                    TreePtr[ply + 1]++;
                    return p1;
                }

            case PICKGEN2:
                pickphase[ply] = PICKREST;
                PhasePick_p[ply] = TreePtr[ply + 1];
                GenMove.GenNonCaptures(ply);
                for (p2 = PhasePick_p[ply]; p2 < TreePtr[ply + 1]; p2++) {
                    Tree[p2].score = (int) (history[side][(Tree[p2].move & 0x0FFF)] +
                            taxicab[FROMSQ(Tree[p2].move)][D5] - taxicab[TOSQ(Tree[p2].move)][E4]);
                    if ((Tree[p2].move & CASTLING) != 0)
                        Tree[p2].score += CASTLINGSCORE;
                }

            case PICKREST:
                while (PhasePick_p[ply] < TreePtr[ply + 1]) {
                    pick(PhasePick_p[ply], ply);
                    mv = Tree[PhasePick_p[ply]].move & MOVEMASK;
                    if (mv == Hashmv[ply] || mv == killer1[ply] ||
                            mv == killer2[ply]) {
                        PhasePick_p[ply]++;
                        continue;
                    }
                    p1 = PhasePick_p[ply]++;
                    return p1;
                }
        }
        return -1;
    }

    static int[] PhasePick1_p = new int[MAXPLYDEPTH];   // was a static local

    /**
     * Similar to phase pick, but only used when the King is in check.
     */
    static /*leaf*/int PhasePick1(int ply)  // return -1 for no pick
    {
        int p1;

        switch (pickphase[ply]) {
            case PICKHASH:
                pickphase[ply] = PICKREST;
                PhasePick1_p[ply] = TreePtr[ply];

            case PICKREST:
                while (PhasePick1_p[ply] < TreePtr[ply + 1]) {
                    pick(PhasePick1_p[ply], ply);
                    p1 = PhasePick1_p[ply]++;
                    return p1;
                }
        }
        return -1;
    }
}
